/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: addkeys.cpp 155 2014-01-22 15:36:31Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fits.h>
using namespace ltl;

#include <iostream>
#include <string>
#include <list>
#include <cstdlib>
using namespace std;

// structure for flag values
class flagValues
{
   public:
      string prefix;
      string history;
      string comment;
      bool cah;
      bool cac;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   list<string> keys, values, comments, pure_comments;
   
   while(!cin.eof())
   {
      string line_i;
      getline(cin, line_i);
      const string::size_type keywordstart = line_i.find_first_not_of(' ');
      const string::size_type keywordend = line_i.find_first_of('=');
      const string::size_type valueend = line_i.find_first_of('#');
      if(keywordend != string::npos)
      {
         keys.push_back(line_i.substr(keywordstart, keywordend-keywordstart));
         values.push_back(line_i.substr(keywordend+1,valueend));
         const string::size_type commentstart = (valueend != line_i.npos) ?
            line_i.find_first_not_of(' ', valueend+1) :
            line_i.npos;
         if(commentstart != line_i.npos)
            comments.push_back(line_i.substr(commentstart, line_i.npos));
         else
            comments.push_back(string(""));
         if(flagVal.verbose > 1)
            cerr << line_i.substr(keywordstart,keywordend)
                 << line_i.substr(keywordend+1,valueend)
                 << ((commentstart != line_i.npos) ?
                     line_i.substr(commentstart, line_i.npos) : string(""))
                 << endl;
      }
      else
      {
         if(valueend == string::size_type(0))
            if(flagVal.cah || flagVal.cac)
            {
               const string::size_type commentstart =
               (valueend != line_i.npos) ?
               line_i.find_first_not_of(' ', valueend+1) :
               line_i.npos;
               if(commentstart != line_i.npos)
                  pure_comments.push_back(line_i.substr(commentstart,
                                                        line_i.npos));
            }
      }
   }   

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         const string outname = prepend_filename(*file_i, flagVal.prefix);
         FitsOut outfile(outname,
                         infile, flagVal.quiet_please);
         if(flagVal.history.size()>0)
            outfile.addHistory(flagVal.history);

         list<string>::const_iterator k_i = keys.begin();
         list<string>::const_iterator v_i = values.begin();
         list<string>::const_iterator c_i = comments.begin();
         while(k_i != keys.end())
         {
            if((*v_i)[0] == '\'')
            {
               const string::size_type stringend =
                  (*v_i).find_first_of('\'',1);
               const string value_i = (*v_i).substr(1, stringend-1);
               if(flagVal.verbose > 1)
                  cerr << "string type key:"
                       << *k_i << '=' << *v_i << " # " << *c_i << endl
                       << outfile.adjustKeyword(*k_i) << "= "
                       << '\'' << value_i << '\'' 
                       << " / " << *c_i << endl;
               outfile.addValueCard(outfile.adjustKeyword(*k_i),
                                    value_i,
                                    *c_i);
            }
            else
            {
               const long vint = long(atoi((*v_i).c_str()));
               const double vdbl = atof((*v_i).c_str());
               if(flagVal.verbose > 1)
               {
                  if(double(vint) == vdbl)
                     cerr << "int type key: "
                          << *k_i << '=' << *v_i << " # " << *c_i << endl
                          << outfile.adjustKeyword(*k_i) << "= "
                          << vint 
                          << " / " << *c_i << endl;
                  else
                     cerr << "int type key:"
                          << *k_i << '=' << *v_i << " # " << *c_i << endl
                          << outfile.adjustKeyword(*k_i) << "= "
                          << vdbl 
                          << " / " << *c_i << endl;
               }
               if(double(vint) == vdbl)
                  outfile.addValueCard(outfile.adjustKeyword(*k_i),
                                       vint,
                                       *c_i);
               else
                  outfile.addValueCard(outfile.adjustKeyword(*k_i),
                                       vdbl,
                                       *c_i);
            }
            ++k_i;
            ++v_i;
            ++c_i;
         }
         
         if(flagVal.cah)
            for(list<string>::const_iterator c_i = pure_comments.begin();
                c_i != pure_comments.end();\
                ++c_i)
               outfile.addHistory(*c_i);
         if(flagVal.cac)
            for(list<string>::const_iterator c_i = pure_comments.begin();
                c_i != pure_comments.end();\
                ++c_i)
               outfile.addComment(*c_i);
         if(flagVal.history.size()>0)
            outfile.addHistory(flagVal.history);
         if(flagVal.comment.size()>0)
            outfile.addComment(flagVal.comment);

         if(flagVal.verbose > 0)
            cerr << "Saving " << outname << endl;
         outfile << infile;

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: addkeys [-FLAGS [option]] FILES\n") +
      string("Add FITS keys read from stdin in syntax:\n") +
      string("KEYWORD=value # optional comment\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new StringOption("history", "", "history string added to output files",
                                       'h', &flagVal.history) );
      flags.addOption(new StringOption("comment", "", "comment string added to output files",
                                       'c', &flagVal.comment) );
      flags.addOption(new BoolOption("comment_as_history", "FALSE", "add commentary lines to FITS history",
                                       0, &flagVal.cah) );
      flags.addOption(new BoolOption("comment_as_comment", "FALSE", "add commentary lines to FITS comments",
                                       0, &flagVal.cac) );
      flags.addOption(new StringOption("prefix", "k", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

