/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: addphotonnoise.cpp 93 2005-02-19 03:57:36Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe of any but output will be of type FT
#define FT float

#include "fitstools.h"


// class for flag values
class flagValues
{
   public:
      flagValues(int dimension) : region(dimension), verbose(1) { };
      FT gain;
      FT gain_error;
      bool nan;
      FT nan_value;
      Region region; // region class defined in util/option.h
      string prefix;
      string gainkey;
      string errorkey;
      string writekey;
      bool zero;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   // for all FILES
   for( list<string>::const_iterator file_i(files.begin());
        file_i != files.end();
        ++file_i )
   {
      try{
         char buf[128];
         // gain for this run
         FT g = flagVal.gain;
         FT e_g = flagVal.gain_error;
         e_g *= e_g / (g * g);

         // open IO-file(s)
         if(flagVal.verbose > 0)
            cerr  << "Checking " << *file_i << endl;
         FitsIn infile(*file_i, flagVal.quiet_please, flagVal.ignore_header);
         {
            string to_history = string("added photon noise to error frame");
            if(flagVal.region.getStart(1) > 0)
               to_history += string(" for region ") + flagVal.region.toString();
            to_history +=string(":");
            if(flagVal.verbose > 0)
               cerr << to_history << endl;
            infile.addHistory(to_history);
         }

         // get gain

         // get value of FITS key
         if( flagVal.gainkey.length() > 0 )
         {
            const double holder =
               infile.getFloat( infile.adjustKeyword(flagVal.gainkey) );
            g *= FT(holder);
            sprintf(buf, "GAIN of FITS key %s", flagVal.gainkey.c_str());
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            infile.addHistory(buf);
         }

         if( flagVal.errorkey.length() > 0 )
         {
            const double holder =
               infile.getFloat( infile.adjustKeyword(flagVal.errorkey) );
            e_g = FT(holder * holder) / (g * g);
            sprintf(buf, "GAIN ERROR of FITS key %s", flagVal.errorkey.c_str());
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            infile.addHistory(buf);
         }
         
         {
            if( g != FT(1.0)) // get constant
               sprintf(buf, "photon noise = SQRT( signal / %g )", g);
            else
               strcpy(buf, "photon noise = SQRT( signal )");
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            infile.addHistory(buf);
            if( e_g != FT(0.0)) // get constant
               sprintf(buf, "using gain error = %g", sqrt(e_g));
         }
         
         // if requested write gain constant to key
         if( flagVal.writekey.length() > 0 )
            infile.addValueCard( infile.adjustKeyword(flagVal.writekey), g,
                                 "gain used to calculate photon noise");

         // save file
         const string outfilename = ft.prependFilename(*file_i, flagVal.prefix);
         {
            FitsOut outfile(outfilename, infile, flagVal.quiet_please);
            if(flagVal.verbose > 0)
            cerr << "Copying " << *file_i
                 << " to " << outfilename << endl;
            outfile << infile;
         }
         
         // arrays for input and output
         MArray<FT, dimension> a, e_a, ra, e_ra;

         // load image and (if present) error
         const bool have_error = ft.loadImageError2(infile, a, e_a);

         // if no input error start with zero
         if( !have_error )
         {
            e_a.makeReference( MArray<FT, dimension>( a.shape()) );
            e_a = FT(0);
         }
         // get region to operate on, default is whole array
         ft.referenceRegion(e_ra, e_a, flagVal.region );
         ft.referenceRegion(ra, a, flagVal.region);
         
         // calculate error
         if(flagVal.verbose > 0)
            cerr << "Calculating error" << endl;
         if( (!have_error) || flagVal.zero ) // no initial error in file -> zero not dominant
            ft.addZeroPhotonNoise2(e_ra, ra, g, flagVal.nan, flagVal.nan_value);
         else
            ft.addPhotonNoise2(e_ra, ra, g, flagVal.nan, flagVal.nan_value);
         if(e_g != FT(0.0))
         {
            ft.convert2RelativeError2(e_ra, ra, flagVal.nan, flagVal.nan_value);
            e_ra += e_g;
            ft.convert2AbsoluteError2(e_ra, ra, flagVal.nan, flagVal.nan_value);
         }
         ra.free(); a.free();
         
         {
            FitsOut outfile(outfilename, infile, flagVal.quiet_please);
            ft.saveError2(e_a, outfile);
         }
      } // this file should have worked, otherwise ->
      catch (exception &e)
      {
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: addphotonnoise [-FLAGS [option]] FILES\n") +
      string("add photon noise to error frame using SQRT(signal/gain)\n") +
      string("with given gain (default 1)\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new FloatOption("gain", "1.0", "gain",
                                      'g', &flagVal.gain ) );
      flags.addOption(new FloatOption("gain_error", "0.0", "gain",
                                      0, &flagVal.gain_error ) );
      flags.addOption(new StringOption("gain_key", "", "read gain of FITS keyword",
                                       0, &flagVal.gainkey) );
      flags.addOption(new StringOption("error_key", "", "read gain error of FITS keyword",
                                       0, &flagVal.errorkey) );
      flags.addOption(new StringOption("write_key", "", "write gain to FITS key",
                                       0, &flagVal.writekey) );
      flags.addOption(new BoolOption("zero_error_not_dominant", "FALSE",
                                     "do NOT keep zero error dominant (only sometimes makes sense)",
                                     'z', &flagVal.zero ) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform only on region, leave rest of frame(s) untouched",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("prefix", "p", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

