/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: anticosmicfits.cpp 115 2007-08-29 16:07:55Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT float

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#ifdef UTIL_USING_NAMESPACE
using namespace util;
#endif

#include <ltl/fitsio.h>
#include <ltl/statistics.h>

#ifdef LTL_USING_NAMESPACE
using namespace ltl;
#endif

#include <string>
#include <iostream>
#include <list>

using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) //: region(dimension)
      { };
      float thresh;
      bool new_error;
      bool sigma;
      bool nan;
      FT nan_value;
      //Region region; // region class defined in util/option.h
      string exptime;
      string juldate;
      string prefix;
      bool zero;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

inline string prependFilename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // class for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 2) print_usage(flags);
   
   try{
      const int n_files = files.size();

      string history = "";

      if(flagVal.verbose > 0)
         cerr << "Building " << history << endl;

      // open input files
      list<FitsIn> infiles;
      list<FitsIn> inerrors;
      bool have_errors = true;
      if(flagVal.new_error) have_errors = false;
      list<string>::const_iterator file_i = files.begin();
      while( file_i != files.end() )
      {
         const string filename = *file_i;
         if(flagVal.verbose > 0)
            cerr << filename << endl;
         infiles.push_back( FitsIn(filename,
                                   flagVal.quiet_please,
                                   flagVal.ignore_header) );
         if(have_errors)
         {
            const string errorname = prependFilename(filename, "e.");
            try
            {
               inerrors.push_back( FitsIn(errorname,
                                          flagVal.quiet_please,
                                          true) );
            }
            catch (exception &e){
               cerr << e.what() << endl;
               have_errors = false;
            }
         }
         history += string(" ") + filename;
         ++file_i;
      }

      const list<FitsIn>::iterator first_error = inerrors.begin();
      const list<FitsIn>::iterator first_file = infiles.begin();
      const int xlength = (*(first_file)).getNaxis(1);
      const int ylength = (*(first_file)).getNaxis(2);

      MArray<FT, 2> files_rows(n_files, xlength);
      MArray<FT, 2> errors_rows(n_files, xlength);
      MArray<FT, 2> outarray(xlength, ylength);
      MArray<FT, 2> outerror(xlength, ylength);
      MArray<FT, 2>::iterator out_i = outarray.begin();
      MArray<FT, 2>::iterator out_e = outerror.begin();

      MArray<FT, 1> refholder;
      MArray<FT, 1> e_refholder;
      // loop over alls rows of all files
      for(int y = 1; y <= ylength; ++y)
      {
         if(flagVal.verbose > 1)
            if(!(y%10))
               cerr << "lines up to " << y << endl;
         list<FitsIn>::iterator infile_i = first_file;
         list<FitsIn>::iterator inerror_i = first_error;

         // read y-th row of all files
         for(int z = 1; z <= n_files; ++z)
         {
            refholder.makeReference( files_rows(z, Range::all()) );
            (*infile_i) >> refholder;
            ++infile_i;
            if(have_errors)
            {
               e_refholder.makeReference( errors_rows(z, Range::all()) );
               (*inerror_i) >> e_refholder;
               ++inerror_i;
               e_refholder = merge(e_refholder > FT(0.0),
                                   pow2(e_refholder),
                                   e_refholder);
            }
         }

         // evaluate mean for every pixel of a row in all files
         for(int x = 1; x <= xlength; ++x)
         {
            refholder.makeReference( files_rows(Range::all(), x) );
	    if(have_errors)  e_refholder.makeReference( errors_rows(Range::all(), x) );

	   
            double mean, sigma, asum;

		  if(flagVal.nan) // with nan
                     sigma = stddev( refholder, flagVal.nan_value, &mean );
                  else sigma = stddev( refholder, &mean ); // without nan

		 for(int i = 1; i<= refholder.nelements(); i++)
		 {
			if(refholder(i) > mean*(1+flagVal.thresh/100))
			{
				refholder(i) = min(refholder);
				if(have_errors) e_refholder(i) = pow2(refholder(i));
			}
		 }
		asum = sum(refholder);
                  if(flagVal.nan) // with nan
                     sigma = stddev( refholder, flagVal.nan_value, &mean );
                  else sigma = stddev( refholder, &mean ); // without nan

            if( !flagVal.sigma ) sigma /= sqrt(double(refholder.nelements())); // use SEM

            if(have_errors) // use error from files
            {
               if(flagVal.nan)
               {
                  const FT e_min = min(e_refholder);
                  if(e_min < FT(0.0)) // dominant error
                     *out_e = e_min;
                  else // error of mean ignoring NaNs
                  {
                     const double ssum = sqrt(sum( e_refholder ));
                     const int n_valid = count(e_refholder > FT(0.0));
                     *out_e = FT(ssum) * FT(sqrt(n_valid));
                  }
               }
               else // simple error of mean
                  *out_e = FT(sqrt(sum(e_refholder))) /
                     FT(sqrt(e_refholder.nelements()));
            }
            else // no error (at least one file missing or preset) -> use sem
               *out_e = FT( sigma ) / FT(sqrt(e_refholder.nelements()));;
            ++out_e;
            *out_i = FT( asum );
            ++out_i;
         }
      }

      // free some memory
      refholder.free();
      e_refholder.free();
      
      files_rows.free();

      const string outname = prependFilename( (*first_file).getFilename(), 
                                              flagVal.prefix);
      FitsOut outfile(outname,
                            *first_file, flagVal.quiet_please);

      if(flagVal.juldate.size() > 0)
      {
         double juldate = 0.0;
         list<FitsIn>::const_iterator infile_i = first_file;
         const string jdkey = (*first_file).adjustKeyword(flagVal.juldate);
         const list<FitsIn>::const_iterator infile_end = infiles.end();
         size_t n_jd = size_t(0);
         while(infile_i != infile_end)
         {
            try{
               juldate += (*infile_i).getFloat(jdkey);
               ++n_jd;
            }
            catch(FitsException e){
               if(flagVal.verbose > 1)
                  cerr << e.what() << endl;
            }
            ++infile_i;
         }
         juldate /= double(n_jd);
         outfile.addValueCard(jdkey, juldate,
                              string("average Julian Date"));
      }

      if(flagVal.exptime.size() > 0)
      {
         double exptime = 0.0;
         list<FitsIn>::const_iterator infile_i = first_file;
         const string exptkey = (*first_file).adjustKeyword(flagVal.exptime);
         const list<FitsIn>::const_iterator infile_end = infiles.end();
         while(infile_i != infile_end)
         {
            try{
               exptime += (*infile_i).getFloat(exptkey);
            }
            catch(FitsException e){
               if(flagVal.verbose > 1)
                  cerr << e.what() << endl;
            }
            ++infile_i;
         }
         outfile.addValueCard(exptkey, exptime,
                              string("summed exposure time"));
      }
      
      infiles.clear();

      outfile.addHistory(history);
      if(flagVal.verbose > 0)
         cerr << "Saving " << outname << endl;
      outfile << outarray;
      const string outerrorname = prependFilename( outfile.getFilename(), "e." );
      outfile.setFilename(outerrorname);
      if(flagVal.verbose > 0)
         cerr << "Saving " << outerrorname << endl;
      outfile << outerror;
   }
   catch (exception &e){
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
   }
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: anticosmicfits [-FLAGS [option]] FILES\n") +
      string("build a sum of cosmic corrected FILES, keywords and filename taken from 1st file\n\n");

   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new FloatOption("threshold", "50",
                                     "skip values that are threshold% higher than mean",
                                     't', &flagVal.thresh ) );
      flags.addOption(new BoolOption("new_error", "FALSE",
                                     "ignore input error files and use sem/sigma for new error which is default if at least one error file is missing",
                                     0, &flagVal.new_error ) );
      flags.addOption(new BoolOption("sigma", "FALSE",
                                     "add (kappa corrected) sigma to error; default is using just (kappa corrected) standard error of average/median",
                                     's', &flagVal.sigma ) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
//      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
//                                            "x_start:x_end,y_start:y_end perform only on region, leave rest of frame(s) untouched",
//                                            'r', 2, &flagVal.region) );
      flags.addOption(new StringOption("exposuretime", "",
                                       "also sum up value of FITS keyword exposuretime",
                                       'e', &flagVal.exptime) );
      flags.addOption(new StringOption("juldate", "",
                                       "calculate mean observation julian date using FITS keyword matching juldate",
                                       'j', &flagVal.juldate) );
      flags.addOption(new StringOption("prefix", "m", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

