/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: templatefits.cpp 94 2005-02-21 12:18:38Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

//#define LTL_RANGE_CHECKING

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#include <ltl/fitsio.h>

using namespace ltl;
using namespace util;
using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues() { };
      int xbin;
      int ybin;
      
      int binning[2];
      float nan_value;
      bool nan;
      string prefix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
      bool no_clean_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

string crpix(const int i)
{
   char buf[16];
   sprintf(buf, "CRPIX%-3d", i);
   return string(buf);
}

string cdi_j(const int i, const int j)
{
   char buf[16];
   sprintf(buf, "CD%d_%-*d", i, 4 - int(log10f(float(i))) , j);
   return string(buf);
}

void clean_header(FitsHeader& header)
{
   header.eraseCard("BIASSEC ");
   header.eraseCard("TRIMSEC ");
   header.eraseCard("DATASEC ");
   header.eraseCard("CCDSEC  ");
   header.eraseCard("ORIGSEC ");
   header.eraseCard("CDELT1  ");
   header.eraseCard("CDELT2  ");
   header.eraseCard("PC001001");
   header.eraseCard("PC001002");
   header.eraseCard("PC002001");
   header.eraseCard("PC002002");
   header.eraseCard("CROTA1  ");
   header.eraseCard("CROTA2  ");
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   flagVal.xbin = flagVal.binning[0];
   flagVal.ybin = flagVal.binning[1];

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1 || flagVal.xbin < 1 || flagVal.ybin < 1) print_usage(flags);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<float, 2> a;

         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
         infile >> a; infile.freeData();

         if(flagVal.xbin > 1)
         {
            const int binlength = (a.length(1) / flagVal.xbin) - 1;
            const Range xrange(1, (binlength * flagVal.xbin)+1, flagVal.xbin);
            MArray<float, 2> b( binlength+1, a.length(2));
            if(flagVal.nan){
               b = merge( (a( xrange, Range::all() ) == flagVal.nan_value) ||
                          (a( xrange+1, Range::all() ) == flagVal.nan_value),
                          flagVal.nan_value,
                          a( xrange, Range::all() ) + a( xrange+1, Range::all() ) );
               for(int i = 2; i < flagVal.xbin; ++i)
                  b = merge( (a( xrange + i, Range::all() ) == flagVal.nan_value) ||
                             (b == flagVal.nan_value),
                             flagVal.nan_value,
                             b + a( xrange+i, Range::all() ) );
            }
            else{
               b = a( xrange, Range::all() ) + a( xrange+1, Range::all() );
               for(int i = 2; i < flagVal.xbin; ++i)
                  b += a( xrange + i, Range::all() );
            }
            a.makeReference(b);
         }

         if(flagVal.ybin > 1)
         {
            const int binlength = (a.length(2) / flagVal.ybin) - 1;
            const Range yrange(1, (binlength * flagVal.ybin)+1, flagVal.ybin);
            MArray<float, 2> b( a.length(1), binlength+1);
            if(flagVal.nan){
               b = merge( (a( Range::all(), yrange ) == flagVal.nan_value) ||
                          (a( Range::all(), yrange+1 ) == flagVal.nan_value),
                          flagVal.nan_value,
                          a( Range::all(), yrange ) + a( Range::all(), yrange+1 ) );
               for(int i = 2; i < flagVal.ybin; ++i)
                  b = merge( (a( Range::all(), yrange+i ) == flagVal.nan_value) ||
                             (b == flagVal.nan_value),
                             flagVal.nan_value,
                             b + a( Range::all(), yrange+i ) );
            }
            else{
               b = a( Range::all(), yrange ) + a( Range::all(), yrange+1 );
               for(int i = 2; i < flagVal.ybin; ++i)
                  b += a( Range::all(), yrange + i );
            }
            a.makeReference(b);
         }

         const string outname = prepend_filename(*file_i, flagVal.prefix);
         FitsOut outfile(outname,
                         infile, flagVal.quiet_please);
         if(!flagVal.no_clean_header){
            clean_header(outfile);
	    // correct WCS, i.e. modify CRPIX if applicable
	    try{
	       for(int i = 1; i<=2; ++i){
		  const double crpix_out =
		     outfile.getFloat(crpix(i)) / (double( flagVal.binning[i-1] ));
		  outfile.addValueCard(crpix(i), crpix_out);
	       }
	    }
	    catch (exception &e){
	       if(flagVal.verbose > 1)
		  cerr << "No WCS\n";
	    }
	    try{
	       for(int i = 1; i<=2; ++i)
		  for(int j = 1; j<=2; ++j){
		     const double cdi_j_out =
			outfile.getFloat(cdi_j(i, j)) * (double( flagVal.binning[i-1] ));
		     outfile.addValueCard(cdi_j(i, j), cdi_j_out);
		  }
	    }
	    catch (exception &e){
	       if(flagVal.verbose > 1)
		  cerr << "No WCS CD matrix\n";
	    }
	 }
	 
         char buf[160];
         sprintf(buf, "Applied binning of %d,%d to ", flagVal.xbin, flagVal.ybin);
         outfile.addHistory(string(buf) + *file_i);
         if(flagVal.verbose > 0)
            cerr << "Saving " << outname << endl;
         outfile << a; a.free();

         try{
            const string errorname = prepend_filename(*file_i, "e.");
            FitsIn inerror(errorname,
                           flagVal.quiet_please, true);
            if(flagVal.verbose > 0)
               cerr << "Reading " << errorname << endl;
            inerror >> a; inerror.freeData();
            if(flagVal.nan)
               a = merge(a > 0.0f, a*a, a);
            else
               a = merge(a > 0.0f, a*a, 0.0f);

            if(flagVal.xbin > 1)
            {
               const int binlength = a.length(1) / flagVal.xbin;
               const Range xrange(1, binlength * flagVal.xbin, flagVal.xbin);
               MArray<float, 2> b( binlength, a.length(2));
               if(flagVal.nan){
                  b = merge( (a( xrange, Range::all() ) <= 0.0f) ||
                             (a( xrange+1, Range::all() ) <= 0.0f),
                             merge(a( xrange, Range::all() ) < a( xrange+1, Range::all() ),
                                   a( xrange, Range::all() ), a( xrange+1, Range::all() )),
                             a( xrange, Range::all() ) + a( xrange+1, Range::all() ) );
                  for(int i = 2; i < flagVal.xbin; ++i)
                     b = merge( (a( xrange + i, Range::all() ) <= 0.0f) ||
                                (b <= 0.0f),
                                merge(a( xrange + i, Range::all() ) < b, 
                                      a( xrange + i, Range::all() ), b),
                                b + a( xrange+i, Range::all() ) );
               }
               else{
                  b = a( xrange, Range::all() ) + a( xrange+1, Range::all() );
                  for(int i = 2; i < flagVal.xbin; ++i)
                     b += a( xrange + i, Range::all() );
               }
               a.makeReference(b);
            }
            
            if(flagVal.ybin > 1)
            {
               const int binlength = a.length(2) / flagVal.ybin;
               const Range yrange(1, binlength * flagVal.ybin, flagVal.ybin);
               MArray<float, 2> b( a.length(1), binlength);
               if(flagVal.nan){
                  b = merge( (a( Range::all(), yrange ) <= 0.0f) ||
                             (a( Range::all(), yrange+1 ) <= 0.0f),
                             merge(a( Range::all(), yrange ) < a( Range::all(), yrange+1 ),
                                   a( Range::all(), yrange ), a( Range::all(), yrange+1 )),
                             a( Range::all(), yrange ) + a( Range::all(), yrange+1 ) );
                  for(int i = 2; i < flagVal.ybin; ++i)
                     b = merge( (a( Range::all(), yrange+i ) <= 0.0f) ||
                                (b <= 0.0f),
                                merge(a( Range::all(), yrange+i ) < b,
                                      a( Range::all(), yrange+i ), b),
                                b + a( Range::all(), yrange+i ) );
               }
               else{
                  b = a( Range::all(), yrange ) + a( Range::all(), yrange+1 );
                  for(int i = 2; i < flagVal.ybin; ++i)
                     b += a( Range::all(), yrange + i );
               }
               a.makeReference(b);
               a = merge(a<=0.0f, a, sqrt(a));
            }
            a = merge(a<=0.0f, a, sqrt(a));
         
            const string outerrorname = prepend_filename(outname, "e.");
            outfile.setFilename(outerrorname);
            if(flagVal.verbose > 0)
               cerr << "Saving " << outerrorname << endl;
            outfile << a;
         }
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << e.what() << endl;
         }

         

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: binfits [-FLAGS [option]] FILES\n") +
      string("bin pixels along x,y axes.\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2008 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new IntArrayOption("binning", "1,1",
                                         "set binning to x,y",
                                         'b', 2, flagVal.binning ) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan_value in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("prefix", "b", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new BoolOption("dont_clean_header", "FALSE",
                                     "do not erase the FITS cards BIASSEC, TRIMSEC, DATASEC, CCDSEC and ORIGSEC, and do not alter any WCS keys",
                                     0, &flagVal.no_clean_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

