/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------
 *
 * Copyright (C) Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#include <ltl/fitsio.h>

using namespace ltl;
using namespace util;
using namespace std;

// structure for flag values
class flagValues
{
public:
   flagValues() { };

   int extno;
   int colno;
   string extname;
   string ttype;
   
   int verbose;
   bool quiet_please;
   bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{
  // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if( (files.size() < 1) ||
       (flagVal.extno == 0 && flagVal.extname.size() == 0) ||
       (flagVal.colno == 0 && flagVal.ttype.size() == 0) )
      print_usage(flags);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{

         FitsExtensionIn infile(*file_i,
                                flagVal.quiet_please, flagVal.ignore_header);
         FitsBinTableIn intable( flagVal.extname.size() > 0 ? // by name or number?
                                 infile.getExtension("BINTABLE", flagVal.extname) :
                                 infile.getExtension(flagVal.extno) );
         if(flagVal.verbose > 0)
            cerr << "Reading data from " << *file_i << endl;
         MArray<double, 2> a;
         a.makeReference( flagVal.ttype.size() > 0 ? // by name or number?
                          intable.readColumn<double, 2>(flagVal.ttype) :
                          intable.readColumn<double, 2>(flagVal.colno) );
         for(int y = 1; y <= a.length(2); ++y)
         {
            for(int x = 1; x <= a.length(1); ++x)
            {
               cout << a(x, y);
               if(a.length(2) == 1) cout << endl;
               else if(x < a.length(1)) cout << '\t';
            }
            if(a.length(2) > 1) cout << endl;
         }
         a.free();

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: bintablefits [-FLAGS [option]] FILES\n") +
      string("Extract (for now just) a column of a FITS BINTABLE extension.\n") +
      string("You have to either give number or name for the extension and the column.\n\n");
   const string copyright = 
      string("Copyright (C) Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      
      flags.addOption(new IntOption("extno", "0",
                                    "extract extension by number",
                                    'x', &flagVal.extno ) );
      flags.addOption(new IntOption("colno", "0",
                                    "extract column by number",
                                    'n', &flagVal.colno ) );
      flags.addOption(new StringOption("extname", "",
                                       "extract extension by name",
                                       'e', &flagVal.extname) );
      flags.addOption(new StringOption("ttype", "",
                                       "extract column by ttype name",
                                       't', &flagVal.ttype) );

      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

