/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: clipfits.cpp 108 2006-07-03 17:25:52Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe of any but output will be of type FT
#define FT float

#include "fitstools.h"


// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) :
         region(dimension), verbose(1) { };
      bool less;
      bool equal;
      bool nan;
      int left, right, up, down;
      FT nan_value;
      FT image_value;
      FT error_value;
      FT limit;
      float radius;
      Region region;
      string prefix;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // class for flag values
   flagValues flagVal(DIMENSION);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // array for input and output
         MArray<FT, DIMENSION> a, ra, e_a, e_ra;

         // open file, read data from file
         FitsIn infile(*file_i,
                             flagVal.quiet_please, flagVal.ignore_header);
         ft.loadImage(infile, a);
         try{
            FitsIn inerror(ft.prependFilename(*file_i, "e."),
                                 flagVal.quiet_please, true);
            ft.loadImage(inerror, e_a);
         }
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << "No errorfile, creating new one = 0.0\n";
            e_a.makeReference(MArray<FT, DIMENSION>(a.shape()));
            e_a = FT(0.0);
         }
         // get region to operate on, default is whole array
         ft.referenceRegion(e_ra, e_a, flagVal.region);
         ft.referenceRegion(ra, a, flagVal.region);

         const IndexList<DIMENSION> pixellist =
            ( flagVal.nan ?
              ( flagVal.less ?
                (flagVal.equal ?
                 where((e_ra >= flagVal.error_value) && 
                       (ra != flagVal.nan_value) && (ra <= flagVal.limit)) :
                 where((e_ra >= flagVal.error_value) && 
                       (ra != flagVal.nan_value) && (ra < flagVal.limit))) :
                (flagVal.equal ?
                 where((e_ra >= flagVal.error_value) && 
                       (ra != flagVal.nan_value) && (ra >= flagVal.limit)) :
                 where((e_ra >= flagVal.error_value) && 
                       (ra != flagVal.nan_value) && (ra > flagVal.limit))) ) :
              ( flagVal.less ?
                (flagVal.equal ?
                 where(ra <= flagVal.limit) :
                 where(ra < flagVal.limit)) :
                (flagVal.equal ?
                 where(ra >= flagVal.limit) :
                 where(ra > flagVal.limit)) ) );

         list<FixedVector<int, DIMENSION> >::const_iterator il_i =
            pixellist.begin();
         while(il_i != pixellist.end())
         {
            if(flagVal.radius != 0.0f)
            {
               const float x = float((*il_i)(1));
               const float y = float((*il_i)(2));
               const float radius = flagVal.radius;
               const float rp1 = radius + 1.0;
               const float radius2 = pow2(radius);
               const float xdstart = x - rp1;
               const float ydstart = y - rp1;
               const float dnaxis1 = float(ra.maxIndex(1));
               const float dnaxis2 = float(ra.maxIndex(2));
               if((xdstart < dnaxis1) && (ydstart < dnaxis2))
               {
                  const float xdend = x + rp1;
                  const float ydend = y + rp1;
                  const int xstart = xdstart < 1.0 ? 1 : int(xdstart);
                  const int xend = xdend > dnaxis1 ? ra.maxIndex(1) : int(xdend);
                  const int ystart = ydstart < 1.0 ? 1 : int(ydstart);
                  const int yend = ydend > dnaxis2 ? ra.maxIndex(2) : int(ydend);
                  MArray<float, 2> rra( ra( Range(xstart, xend),
                                          Range(ystart, yend) ) );
                  const float x_off = float(xstart - 1);
                  const float y_off = float(ystart - 1);
                  rra = merge( ( pow2( (indexPosFlt(rra,1) + x_off) - x) +
                                 pow2( (indexPosFlt(rra,2) + y_off) - y) )
                               <= radius2, flagVal.image_value, rra);
               }
//                const IndexList<DIMENSION> radiuslist =
//                   where( pow2(x - indexPosFlt(ra,1)) +
//                          pow2(y - indexPosFlt(ra,2)) <= horizon);
//                ra[radiuslist] = flagVal.image_value;
//                e_ra[radiuslist] = flagVal.error_value;
            }
            const int xmin = (*il_i)(1) - flagVal.left;
            const int xmax = (*il_i)(1) + flagVal.right;
            const Range xRange( ((xmin > ra.minIndex(1)) ?
                                       xmin : ra.minIndex(1)),
                                      ((xmax < ra.maxIndex(1)) ?
                                       xmax : ra.maxIndex(1)) );
            ra(xRange, (*il_i)(2)) = flagVal.image_value;
            e_ra(xRange, (*il_i)(2)) = flagVal.error_value;
         
            const int ymin = (*il_i)(2) - flagVal.down;
            const int ymax = (*il_i)(2) + flagVal.up;
            const Range yRange( ((ymin > ra.minIndex(2)) ?
                                       ymin : ra.minIndex(2)),
                                      ((ymax < ra.maxIndex(2)) ?
                                       ymax : ra.maxIndex(2)) );
            ra((*il_i)(1), yRange) = flagVal.image_value;
            e_ra((*il_i)(1), yRange) = flagVal.error_value;
            ++il_i;
         }

//          ra[pixellist] = flagVal.image_value;
         ra.free();
//          e_ra[pixellist] = flagVal.error_value;
         e_ra.free();

         FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                               infile, flagVal.quiet_please);
         const string hist_condition =
            flagVal.less ?
            (flagVal.equal ? string("<=") : string("<")) :
            (flagVal.equal ? string(">=") : string(">"));
         char buf[256];
         sprintf(buf," %g to %g (error to %g)",
                 flagVal.limit, flagVal.image_value, flagVal.error_value);
         const string to_history =
            string("clipped pixels ") + hist_condition + string(buf);
         
         if(flagVal.verbose > 0)
            cerr << to_history << endl;
         outfile.addHistory(to_history);

         // save file
         ft.saveImage(a, outfile);
         a.free();
         if(anyof(e_a))
         {
            outfile.setFilename( ft.prependFilename(outfile.getFilename(),
                                                    "e.") );
            ft.saveImage(e_a, outfile);
         }
         e_a.free();
      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: clipfits [-FLAGS [option]] FILES\n") +
      string("test on flag dependent condition, set image to image_value\n") +
      string("and error to error_value, where condition is true.\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("less", "FALSE",
                                     "use < (less) not > (greater == DEFAULT) for condition",
                                     'l', &flagVal.less ) );
      flags.addOption(new BoolOption("equal", "FALSE",
                                     "use <= / >= (less / greater EQUAL not THAN) for condition",
                                     'q', &flagVal.equal ) );
      flags.addOption(new FloatOption("limit", "0.0",
                                      "use as rhs argument for condition (replace where: pixel CONDITION limit is true)",
                                      0, &flagVal.limit) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new FloatOption("image_value", "0.0",
                                      "replace in image with image_value",
                                      0, &flagVal.image_value) );
      flags.addOption(new FloatOption("error_value", "0.0",
                                      "replace in error image with error_value",
                                      0, &flagVal.error_value) );
      flags.addOption(new IntOption("left", "0",
                                    "also replace LEFT count pixels left to the pixels where condition is true",
                                    0, &flagVal.left ) );
      flags.addOption(new IntOption("right", "0",
                                    "also replace RIGHT count pixels right to the pixels where condition is true",
                                    0, &flagVal.right ) );
      flags.addOption(new IntOption("up", "0",
                                    "also replace UP count pixels above the pixels where condition is true",
                                    0, &flagVal.up ) );
      flags.addOption(new IntOption("down", "0",
                                    "also replace DOWN count pixels below the pixels where condition is true",
                                    0, &flagVal.down ) );
      flags.addOption(new FloatOption("radius", "0.0",
                                      "also replace all pixels within radius of those pixels where condition is true",
                                      0, &flagVal.radius) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform only on region, leave rest of frame(s) untouched",
                                            'r', DIMENSION, &flagVal.region) );
      flags.addOption(new StringOption("prefix", "y", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

