/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: comparefits.cpp 146 2010-11-18 16:58:45Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Arno Riffeser <arri@usm.uni-muenchen.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

//#define LTL_TEMPLATE_LOOP_LIMIT 0

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fitsio.h>
#include <ltl/statistics.h>

#include <ltl/linlsqfit.h>

using namespace ltl;

#include<iostream>
using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues() : region(2) { };

      int clip_limit;
      int length;
      int interpol;
      float kappa;
      double norm_kappa;
      double ref_limit;
      string ref_file;
      string ref_key;
      
      Region region; // region class defined in util/region.h

      float nan_value;
//       bool nan;
      string prefix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   list<FitsIn> infiles;
   list<FitsIn> inerrors;

   MArray<float, 2> refframe, referror2;

   try {
      // open input files
      list<string>::const_iterator file_i = files.begin();
      while( file_i != files.end() )
      {
         const string filename = *file_i;
         if(flagVal.verbose > 0)
            cerr << "Checking " << filename << endl;
         infiles.push_back( FitsIn(filename,
                                   flagVal.quiet_please,
                                   flagVal.ignore_header) );
         const string errorname = prepend_filename(filename, "e.");
         if(flagVal.verbose > 0)
            cerr << "Checking " << errorname << endl;
         try
         {
            inerrors.push_back( FitsIn(errorname,
                                       flagVal.quiet_please,
                                       true) );
         }
         catch( exception& e ) {
            if(flagVal.verbose > 0)
               cerr << e.what() << endl
                    << "Disregarding " << filename << endl;
            infiles.pop_back();
         }
         ++file_i;
      }
      // find reference frame
      // if set in options,  default is using 1st of file list
      if(flagVal.ref_file.size() > 0)
      {
         const string filename = flagVal.ref_file;
         if(flagVal.verbose > 0)
            cerr << "Checking " << filename << endl;
         infiles.push_front( FitsIn(filename,
                                    flagVal.quiet_please,
                                    flagVal.ignore_header) );
         const string errorname = prepend_filename(filename, "e.");
         if(flagVal.verbose > 0)
            cerr << "Checking " << errorname << endl;
         inerrors.push_front( FitsIn(errorname,
                                     flagVal.quiet_please,
                                     true) );
      }
      // now use 1st on list or evaluate list
      list<FitsIn>::iterator infile_i = infiles.begin();
      list<FitsIn>::iterator inerror_i = inerrors.begin();
      if(flagVal.ref_key.size() > 0)
      {
         // evaluate list:
         // find file which has the smallest value of FITS header
         // >= a limit or, if none is bigger than the limit,
         // the biggest
         double minmax = 0.0;
         list<FitsIn>::iterator if_i = infile_i;
         list<FitsIn>::iterator ie_i = inerror_i;
         flagVal.ref_key = (*if_i).adjustKeyword(flagVal.ref_key);
         while(if_i != infiles.end()) {
            try {
               const double value = (*if_i).getFloat(flagVal.ref_key);
               if(value >= flagVal.ref_limit) {
                  if( (value < minmax) || (minmax < flagVal.ref_limit) )
                  {
                     minmax = value;
                     infile_i = if_i;
                     inerror_i = ie_i;
                  }
               }
               else {
                  if(value > minmax)
                  {
                     minmax = value;
                     infile_i = if_i;
                     inerror_i = ie_i;
                  }
               }
            }
            catch( exception& e ) {
               if(flagVal.verbose > 0)
                  cerr << e.what() << endl;
            }
            ++if_i;
            ++ie_i;
         }
      }
      const string reffilename = (*infile_i).getFilename();
      if(flagVal.verbose > 0)
         cerr << "Loading and saving reference file (and error) "
              << reffilename << endl;
      FitsOut outfile( prepend_filename(reffilename, flagVal.prefix),
                       *infile_i, flagVal.quiet_please);
      outfile.addHistory("Used for reference in comparefits");
      (*infile_i) >> outfile;
      (*infile_i) >> refframe;
      infiles.erase(infile_i);
      outfile.setFilename( prepend_filename(outfile.getFilename(), "e.") );
      (*inerror_i) >> outfile;
      (*inerror_i) >> referror2;
      // relative error^2 of reference
      referror2 = merge( (referror2 > 0.0f) &&
                         (refframe != 0.0f) &&
                         (refframe != flagVal.nan_value),
                         pow2(referror2 / refframe),
                         0.0f );
      inerrors.erase(inerror_i);
   }
   catch( exception& e ) {
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
      exit(1);
   }

   // norm ranges
   const int xmin = refframe.minIndex(1);
   const int xmax = refframe.maxIndex(1);
   const int ymin = refframe.minIndex(2);
   const int ymax = refframe.maxIndex(2);
   const Range norm_xrange( flagVal.region.getStart(1) != 0 ?
                            flagVal.region.getStart(1) :
                            xmin + (xmax + 1 - xmin) / 4,
                            flagVal.region.getEnd(1) != 0 ?
                            flagVal.region.getEnd(1) :
                            xmax - (xmax + 1 - xmin) / 4 );
   const Range norm_yrange( flagVal.region.getStart(2) != 0 ?
                            flagVal.region.getStart(2) :
                            ymin + (ymax + 1 - ymin) / 4,
                            flagVal.region.getEnd(2) != 0 ?
                            flagVal.region.getEnd(2) :
                            ymax - (ymax + 1 - ymin) / 4 );
   
   // circle stencil
   MArray<float, 2> circ( Range(-flagVal.length, flagVal.length),
                          Range(-flagVal.length, flagVal.length) );
   circ = cast<float>()( (pow2(indexPosInt(circ, 1)) + pow2(indexPosInt(circ, 2))) <
                         (flagVal.length * (flagVal.length + 1)) );

   // place holder for reference / file with additional length border
   MArray<float, 2> rfull( Range(xmin- flagVal.length,
                                 xmax + flagVal.length),
                           Range(ymin - flagVal.length,
                                 ymax + flagVal.length));
   rfull( Range::all(),
          Range(ymin - flagVal.length, ymin - 1) ) = 0.0f;
   rfull( Range(xmin - flagVal.length, xmin - 1),
          Range(ymin, ymax) ) = 0.0f;
   rfull( Range(xmax + 1, xmax + flagVal.length),
          Range(ymin, ymax) ) = 0.0f;
   rfull( Range::all(),
          Range(ymax + 1, ymax + flagVal.length) ) = 0.0f;
   MArray<float, 2> r( rfull( Range(xmin, xmax),
                              Range(ymin, ymax) ) );

   MArray<float, 2> e_rfull( rfull.shape() );
   e_rfull( Range::all(),
            Range(ymin - flagVal.length, ymin - 1) ) = 0.0f;
   e_rfull( Range(xmin - flagVal.length, xmin - 1),
            Range(ymin, ymax) ) = 0.0f;
   e_rfull( Range(xmax + 1, xmax + flagVal.length),
            Range(ymin, ymax) ) = 0.0f;
   e_rfull( Range::all(),
            Range(ymax + 1, ymax + flagVal.length) ) = 0.0f;
   MArray<float, 2> e_r( e_rfull( Range(xmin, xmax),
                                  Range(ymin, ymax) ) );

   // for all other FILES
   list<FitsIn>::iterator infile_i = infiles.begin();
   list<FitsIn>::iterator inerror_i = inerrors.begin();
   while(infile_i != infiles.end())
   {
      try{
         // arrays for input and output
         MArray<float, 2> a, e_a;

         const string filename = (*infile_i).getFilename();
         if(flagVal.verbose > 0)
            cerr << "Reading " << filename << endl;
         (*infile_i) >> a;
         if(!a.isConformable(refframe))
            throw(FitsException("NAXIS don't match reference frame"));
         (*infile_i).freeData();

         const string errorname = (*inerror_i).getFilename();
         if(flagVal.verbose > 0)
            cerr << "Reading " << errorname << endl;
         (*inerror_i) >> e_a;
         (*inerror_i).freeData();

         // ratio reference / image
         if(flagVal.verbose > 1)
            cerr << "Calculating ratio\n";
         r = merge( (refframe != flagVal.nan_value) &&
                    (a != flagVal.nan_value) &&
                    (a != 0.0f),
                    refframe / a,
                    0.0f);

         // normalise ratio in region
         double norm;
         if(flagVal.norm_kappa)
            kappa_sigma_median(r(norm_xrange, norm_yrange),
                               flagVal.norm_kappa, 0.0f, &norm);
         else
            norm = median_exact(r(norm_xrange, norm_yrange), 0.0f);
         r /= norm;
         if(flagVal.verbose > 1)
            cerr << "Normalising ratio with " << norm << endl;

         // absolute error of ratio
         if(flagVal.verbose > 1)
            cerr << "Propagating error\n";
         e_r = merge( (r != 0.0f) &&
                      (e_a > 0.0f) &&
                      (referror2 > 0.0f),
                      r * sqrt(pow2(e_a / a) + referror2),
                      0.0f);

         // rigorous clip bad pixels
         if(flagVal.verbose > 0)
            cerr << "Clipping bad pixels\n";
         MArray<float, 2>::IndexIterator i_i = rfull.indexBegin();
         MArray<float, 2>::iterator r_i = rfull.begin();
         MArray<float, 2>::iterator e_i = e_rfull.begin();
         while(!e_i.done())
         {
            const float e_val = *e_i;
            if(e_val != 0.0f)
            {
               const MArray<float, 2> v_circ = circ *
                  rfull( Range(i_i(1) - flagVal.length,
                               i_i(1) + flagVal.length),
                         Range(i_i(2) - flagVal.length,
                               i_i(2) + flagVal.length));
               const float val = *r_i;
               const float horizon = e_val * flagVal.kappa;
               const float floor_limit = val - horizon;
               const float ceil_limit = val + horizon;
               const int clip_limit = flagVal.clip_limit;
               int floorctr, ceilctr;
               floorctr = ceilctr = 0;
               MArray<float, 2>::const_iterator v_i = v_circ.begin();
               while( !v_i.done() &&
                      (floorctr < clip_limit) &&
                      (ceilctr < clip_limit))
               {
                  const float cval = *v_i;
                  if(cval != 0.0f)
                  {
                     floorctr += (cval < floor_limit);
                     ceilctr  += (cval > ceil_limit);
                  }
                  ++v_i;
               }
               if( !((floorctr < clip_limit) &&
                     (ceilctr < clip_limit)) )
                  *e_i = 0.0f;
            }
            ++i_i;
            ++r_i;
            ++e_i;
         }
         e_r *= e_r; // square error
         if(flagVal.verbose > 1)
            cerr << "Clipping left: " << count(e_r != 0.0f) << endl;

         // get polynome fit
         if(flagVal.verbose > 0)
            cerr << "Fitting polynome of order " << flagVal.interpol << endl;

         MArray<float, 2> polfit;
         string comment;
         switch(flagVal.interpol)
         {
            case 1:
               polfit.makeReference(PolynomFit<double, float, 1, true, 2>::fit(r, e_r, comment));
               break;
            case 2:
               polfit.makeReference(PolynomFit<double, float, 2, true, 2>::fit(r, e_r, comment));
               break;
            case 3:
               polfit.makeReference(PolynomFit<double, float, 3, true, 2>::fit(r, e_r, comment));
               break;
            case 4:
               polfit.makeReference(PolynomFit<double, float, 4, true, 2>::fit(r, e_r, comment));
               break;
            default:
               throw(LinearAlgebraException("Only polynomes from order 1 to 4 are implemented! (Feel free to expand ... just add another case statement!)"));
               break;
         }

         if(flagVal.verbose > 1)
         {
            const string outpolname = prepend_filename(filename, "pol.");
            FitsOut outpol(outpolname,
                           *infile_i, flagVal.quiet_please);
            outpol.addHistory("File is polynome:");
            outpol.addHistory(comment);
            if(flagVal.verbose > 0)
               cerr << "Saving " << outpolname << endl;
            outpol << polfit;
         }

         if(flagVal.verbose > 0)
            cerr << "Multiplying with polynome:" << endl
                 << comment << endl;

         // apply to image and error and save result
         const string outname = prepend_filename(filename, flagVal.prefix);
         FitsOut outfile(outname,
                         *infile_i, flagVal.quiet_please);
         outfile.addHistory("multiplied with polynome:");
         outfile.addHistory(comment);
         a = merge(a != flagVal.nan_value, a * polfit, a);
         if(flagVal.verbose > 0)
            cerr << "Saving " << outname << endl;
         outfile << a;
         a.free();

         const string outerror = prepend_filename(outname, "e.");
         outfile.setFilename(outerror);
         e_a = merge(e_a > 0.0f, e_a * polfit, e_a);
         if(flagVal.verbose > 0)
            cerr << "Saving " << outerror << endl;
         outfile << e_a;

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
      ++infile_i; ++inerror_i;
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: comparefits [-FLAGS [option]] FILES\n") +
      string("Normalize largest scale structure to given reference\n") +
      string("using a polynome sum_i (x^i * sum_j a_ij y^j)\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Arno Riffeser <arri@usm.uni-muenchen.de>\n") +
      string("                        Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new IntOption("interpol_order", "2",
                                    "polynome order (1 - 4)",
                                    'i', &flagVal.interpol ) );
      flags.addOption(new FloatOption("kappa", "4.0",
                                      "kappa for worst pixel clipping (drop pixels with at least clip_limit high or clip_limit low outliers within clip_radius)",
                                      'k', &flagVal.kappa) );
      flags.addOption(new IntOption("clip_radius", "10",
                                    "clip radius (of circular stencil), also see kappa option",
                                    'c', &flagVal.length ) );
      flags.addOption(new IntOption("clip_limit", "3",
                                    "clip limit, also see kappa option",
                                    0, &flagVal.clip_limit ) );
      flags.addOption(new DoubleOption("norm_kappa", "2.0",
                                       "kappa for kappa-median clipping for norm factor",
                                       0, &flagVal.norm_kappa) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end normalisation determination region (default is center quarter)",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new StringOption("reference_file", "",
                                       "use as reference (default is using 1st file)",
                                       'f', &flagVal.ref_file) );
      flags.addOption(new StringOption("reference_keyword", "",
                                       "use as reference the file with value of keyword, which is min( value > limit ) or if none obeyes max( value )",
                                       0, &flagVal.ref_key) );
      flags.addOption(new DoubleOption("reference_limit", "0.0",
                                       "see option reference_keyword",
                                       0, &flagVal.ref_limit) );
//       flags.addOption(new BoolOption("nan", "FALSE",
//                                      "keep nan_value in either file dominant",
//                                      'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan in addition to 0.0",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("prefix", "d", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

