/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: cosmicfits.cpp 165 2015-03-23 09:54:40Z kodric $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2013 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *                         Arno Riffeser <arri@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

//#define LTL_RANGE_CHECKING
#define LTL_TEMPLATE_LOOP_LIMIT 32

#include "cosmicfits.h"

#include <ltl/marray_io.h>

#ifndef HAVE_ISNAN
template<class T>
bool isnan(const T x)
{
  return (x!=x);
}

#endif

// main
int main(int argc, char* argv[])
{

  // class for flag values
  flagValues flagVal;

  // set options with configfile
  const string cfgfilename = string(argv[0]) + string(".cfg");
  ConfigFileReader cfgfile(cfgfilename);
  OptionParser      flags(&cfgfile);
  try {
    add_options(flags, flagVal);           // save values to class
    flags.parseOptions();
  }
  catch( UException e ) {
    if(flagVal.verbose > 0)
      {
        cerr << e.what() << endl;
        print_usage(flags);
      }
    exit(1);
  }

  // set options with commandline
  CommandLineReader comline( argc, argv ); // read from command line
  flags.changeReader(&comline);
  try {
    flags.parseOptions();
  }
  catch( UException e ) {
    if(flagVal.verbose > 0)
      {
        cerr << e.what() << endl;
        print_usage(flags);
      }
    exit(1);
  }

  if(flagVal.verbose > 1)
    cerr << flags << endl;
  flagVal.quiet_please = (flagVal.verbose == 0);
  const bool refit_always = !flagVal.refit_failed;

  const list<string> files = comline.fileArguments(); // get the filenames
  if(files.size() < 1) print_usage(flags);

  // instance for FitsTools class methods
  FitsTools<float, 2> ft(flagVal.verbose);

  // for all FILES
  for( list<string>::const_iterator file_i = files.begin();
       file_i != files.end();
       ++file_i ) {
    try{
      // arrays for input and output
      MArray<float, 2> a, e_a;

      // load image and error image
      FitsIn infile(*file_i, flagVal.quiet_please, flagVal.ignore_header);
      const bool have_error = ft.loadImageError2(infile, a, e_a);
      if(!have_error)
        {
          // if there's no error image calculate a naive one
          if(flagVal.verbose > 0)
            cerr << "Using SQRT( signal [+ sky] / gain ) + readnoise for error."
                 << endl;
          //e_a.makeReference(MArray<float, 2>( (a + flagVal.sky_offset) /
          //                                    flagVal.gain) );
          const float ron2 = flagVal.readnoise * flagVal.readnoise;
          e_a.makeReference(MArray<float, 2>(
                                             ( (a + flagVal.sky_offset) / flagVal.gain ) +
                                             ron2 ));

          if(flagVal.nan)
            e_a = merge(a != flagVal.nan_value, e_a, flagVal.nan_error);
        }
      MArray<float, 2> b(a.shape()), e_b(a.shape());
      b = a;
      e_b = e_a;
      // mask indicating where to check for cosmics
      MArray< unsigned char , 2> mask(a.shape());

      // use region?
      Region fregion(flagVal.region);
      if(fregion.getStart(1) != 0)
        {
          // adjust to data of file
          ft.adjustRegion2File(infile, fregion);
          MArray<float, 2> e_ra;
          ft.referenceRegion(e_ra, e_a, fregion);
        }
      mask = (e_a > 0.0f);

      // list for results
      list<FitResult<double, 5, 2> > cosmics;
         
      // fitting classes: ltl/fmatrix/marquardt.h, ltl/fmatrix/mrqfuncs.h
      Marquardt< PolyGaussian<double, float, 5, 2>,
        double, float, 5, 2 >
        gaussmrq(flagVal.iter_max,
                 flagVal.alamda[0], flagVal.alamda[1],
                 flagVal.alamda[2], flagVal.alamda[3]);
      Marquardt< Gaussian<double, float, 3, 2>,
        double, float, 3, 2 >
        cgaussmrq(flagVal.iter_max,
                  flagVal.alamda[0], flagVal.alamda[1],
                  flagVal.alamda[2], flagVal.alamda[3]);

      // vectors holding the startparameters
      // global startvalues are set here,
      // rest will be set for individual test pixel
      FVector<double, 5> startpar;
      //startpar(1) = TPAR(mindat);
      //startpar(2) = a(coord(1), coord(2)) - startpar(1);
      startpar(3) = flagVal.start_fwhm;// * M_2_SQRTPI;
      startpar(4) = flagVal.start_fwhm;
      startpar(5) = flagVal.start_angle;
      FVector<double, 3> cstartpar;
      //startpar(1) = TPAR(mindat);
      //startpar(2) = a(coord(1), coord(2)) - startpar(1);
      cstartpar(3) = flagVal.circ_start;

      // counters for condition hits, passes, cosmics and pixels
      size_t passcounter = 1, hitcounter = 0,
        totpcounter = 0, totccounter = 0;
      // if auto_adjustment for cosmicfwhm width is on
      // adjust after 1st pass
      //          bool auto_adjust = (flagVal.auto_adjust != 0.0);
         
      do {
        if(flagVal.verbose > 0) {
          cerr << "------------------------------------------------------------------------" << endl;
          cerr << "Pass " << passcounter << ":" << endl;
        }
        // unmask outside region
        if(fregion.getStart(1) != 0)
          setMAoutsideRegion((unsigned char) 0, mask, fregion);
        // unmask unfittable border
        setMAborder( (unsigned char) 0, mask, flagVal.fit_radius);
        // get local maxima in remaining
        hitcounter = searchlocmax(a.beginRA(), mask.beginRA(),
                                  a.length(1), a.length(2));
        if(flagVal.verbose > 0)
          cerr << "  Found " << hitcounter << " local maxima." << endl;
        // unmask isolated pixel
        if(flagVal.isolation > 0)
          {
            size_t dropcounter = masknanneighbors(a.beginRA(), mask.beginRA(),
                                                  mask.length(1), mask.length(2),
                                                  flagVal.nan_value, flagVal.isolation);
            if(flagVal.verbose > 0)
              cerr << "  Dropped " << dropcounter << " isolated maxima." << endl;
            hitcounter -= dropcounter;
          }
        if(have_error)
          {
            // unmask pixels with saturated neighbours
            size_t dropcounter = masknanneighbors(e_a.beginRA(), mask.beginRA(),
                                                  mask.length(1), mask.length(2),
                                                  -1.0f, 2);
            if(flagVal.verbose > 0)
              cerr << "  Dropped " << dropcounter << " close to saturated maxima." << endl;
            hitcounter -= dropcounter;
            const unsigned char uc0 = 0;
            const unsigned char uc1 = 1;
            // unmask pixels with big errors
            if(flagVal.upper_error > 0.0f)
              {
                const float e_limit = pow2(flagVal.upper_error) / flagVal.gain;
                mask = merge( (mask != uc0) && (e_a < (e_limit * a)),
                              uc1, uc0);
                dropcounter = hitcounter - count(mask);
                if(flagVal.verbose > 0)
                  cerr << "  Dropped " << dropcounter << " maxima with big errors." << endl;
                hitcounter -= dropcounter;
              }
          }
        if(flagVal.verbose > 0)
          cerr << "  Testing " << hitcounter << " local maxima." << endl;

        size_t smoothcount = 0, circfitcount = 0, 
          singmatcount = 0, divergcount = 0,
          nopeakcount = 0, cosmiccount = 0;
            
        MArray< unsigned char , 2>::iterator m_i = mask.begin();
        MArray< unsigned char , 2>::IndexIterator i_i = mask.indexBegin();
         
        while(!m_i.done()) // test array once
          {
            if(*m_i){ // check mask
              *m_i = 0; // this one will be done

              FitResult<double, 5, 2> result;
              result.coord = i_i();
              result.niter = 0;
              const Range xRange(result.coord(1) - flagVal.fit_radius,
                                 result.coord(1) + flagVal.fit_radius);
              const Range yRange(result.coord(2) - flagVal.fit_radius,
                                 result.coord(2) + flagVal.fit_radius);
              MArray<float, 2> ra( a(xRange, yRange) );
              const float mindat = min(ra, flagVal.nan_value);
              if( mindat > ( a(result.coord) -
                             (flagVal.smooth_scale * sqrt(e_a(result.coord))) ) )
                {
                  result.niter = 0;
                  ++smoothcount;
                }
              else
                {
                  startpar(1) = double(mindat);
                  startpar(2) = double(a(i_i(1), i_i(2)) - mindat);
                  //startpar(3) = flagVal.start_fwhm;// * M_2_SQRTPI;
                  //startpar(4) = flagVal.start_fwhm;
                  //startpar(5) = flagVal.start_angle;
                  MArray<float, 2> e_ra( e_a(xRange, yRange) );
                  try
                    {
                      // ELLIPTICAL GAUSSFIT: see ltl/fmatrix/mrqfuncs.h        
                      if(flagVal.cosmic_fwhm > 0.0)
                        {
                          gaussmrq.eval(ra,
                                        flagVal.nan_value,
                                        e_ra,
                                        startpar);
                          result.parameter  = gaussmrq.getResult();
                          if( !(isnan(result.parameter(1)) ||
                                isnan(result.parameter(2)) ||
                                (result.parameter(2) < 0) ||
                                isnan(result.parameter(3)) ||
                                (result.parameter(3) < 0) ||
                                isnan(result.parameter(4)) ||
                                (result.parameter(4) < 0) ||
                                isnan(result.parameter(5))) )
                            {
                              result.paramerror = gaussmrq.getVariance();
                              result.chisquare  = gaussmrq.getChiSquare();
                              result.niter      = gaussmrq.getNIteration();
                            }
                        }
                    }
                  catch( SingularMatrixException e )
                    {
                      if(flagVal.verbose > 3)
                        cerr << e.what() << " @ "
                             << result.coord(1) << ", " 
                             << result.coord(2) << endl;
                    }
                  catch( DivergenceException e )
                    {
                      if(flagVal.verbose > 3)
                        cerr << e.what() << " @ "
                             << result.coord(1) << ", " 
                             << result.coord(2) << endl;
                    }
                  catch( LinearAlgebraException e )
                    {
                      if(flagVal.verbose > 0)
                        {
                          cerr << e.what() << " @ "
                               << result.coord(1) << ", " 
                               << result.coord(2) << endl;
                          cerr << "Consider setting isolation/saturation flag appropriately.\n";                         
                        }
                    }
                     
                  // check if cosmic, if not -> retry with circ fit
                  if(refit_always && (result.niter > 0))
                    {
                      if(!is_cosmic(result, flagVal))
                        result.niter = 0;
                    }

                  // CIRCULAR GAUSSFIT:                  
                  if( (result.niter == 0) && (flagVal.circ_fwhm > 0.0) )
                    {
                      // 5-param fit did not work =>
                      // retry with 3-param fit
                      if(flagVal.verbose > 2)
                        cerr << "  Circular refit: ";
                      cstartpar(1) = startpar(1);
                      cstartpar(2) = startpar(2);
                      //cstartpar(3) = flagVal.start_fwhm;
                      try
                        {
                          cgaussmrq.eval(ra,
                                         flagVal.nan_value,
                                         e_ra,
                                         cstartpar);
                          FVector<double, 3> cparam = cgaussmrq.getResult();
                          FVector<double, 3> cparamerror = cgaussmrq.getVariance();
                          result.parameter(1) = cparam(1);
                          result.parameter(2) = cparam(2);
                          result.parameter(4) = result.parameter(3) = cparam(3);
                          result.parameter(5)  = 0.0;
                          result.paramerror(1) = cparamerror(1);
                          result.paramerror(2) = cparamerror(2);
                          result.paramerror(3) = cparamerror(3);
                          result.paramerror(5) = result.paramerror(4)  = 0.0;
                          result.chisquare  = cgaussmrq.getChiSquare();
                          result.niter      = cgaussmrq.getNIteration();
                          ++circfitcount;
                        }
                      catch( SingularMatrixException ec )
                        {
                          if(flagVal.verbose > 2)
                            cerr << ec.what() << " @ "
                                 << result.coord(1) << ", " 
                                 << result.coord(2) << endl;
                          result.niter = 0;
                          ++singmatcount;
                        }
                      catch( DivergenceException ec )
                        {
                          if(flagVal.verbose > 2)
                            cerr << ec.what() << " @ "
                                 << result.coord(1) << ", " 
                                 << result.coord(2) << endl;
                          result.niter = 0;
                          ++divergcount;
                        }
                    }
                     
                  if(result.niter > 0) // fit ok
                    {
                      if( (result.parameter(2) > 0.0) // and peak (not hole)
                          && (result.parameter(3) > 0.0) // and elliptic,
                          && (result.parameter(4) > 0.0) // i.e. not para- or hyperbolic
                          && (result.paramerror(1) > 0.0) // make sure fit
                          && (result.paramerror(2) > 0.0) // has not gone
                          && (result.paramerror(3) > 0.0) // wrong some how
                          )
                        {
                          ++cosmiccount;                  
                          cosmics.push_back(result); // put on list
                           
                          if(flagVal.verbose > 2)
                            // old fitstools filter output format
                            //                               cerr << i_i(2) << '\t'
                            //                                    << i_i(1) << '\t'
                            //                                    << result.niter << '\t'
                            //                                    << result.parameter(3) << '\t'
                            //                                    << result.parameter(4) << '\t'
                            //                                    << result.parameter(2) << '\t'
                            //                                    << result.parameter(1) << '\t'
                            //                                    << sqrt(result.chisquare) << endl;
                            cerr << i_i(1) << ",\t" 
                                 << i_i(2) << ":\t"
                                 << result.niter << " iter, paramter => "
                                 << result.parameter << ", chi: "
                                 << sqrt(result.chisquare) << endl;
                        }
                      else ++nopeakcount;
                    }
                }
            }
            ++m_i;
            ++i_i;
          } // end of testing array once loop

        hitcounter -= smoothcount + singmatcount + divergcount;
        if(flagVal.verbose > 0)
          {
            cerr << "  Successfully fitted " << hitcounter
                 << " candidates:" << endl;
            cerr << "  " << hitcounter - circfitcount
                 << " with 5 parameter Gaussian," << endl;
            cerr << "  "<< circfitcount
                 << " refitted with 3 parameter Gaussian." << endl;
            cerr << "  Dropped " << smoothcount
                 << " candidates with small start amplitude." << endl;
            cerr << "  Dropped " << singmatcount
                 << " candidates because of singular Matrix in refit." << endl;
            cerr << "  Dropped " << divergcount
                 << " candidates because of divergent refit." << endl;
            cerr << "  Dropping " << nopeakcount 
                 << " candidates, because of amplitude or any width <= 0." << endl;
          }
        hitcounter -= nopeakcount;
        if(flagVal.verbose > 0)
          {
            cerr << "  => " << hitcounter << " remaining ("
                 << " = hitcounter, cosmic count = " << cosmiccount << " )" << endl;
          }

        // readjustment of fwhm criterium
        //             if(auto_adjust)
        //             {
        //                estimate_cosmicfwhm(cosmics, flagVal);
        //                auto_adjust = false;
        //             }

        size_t ccounter = 0, pcounter = 0;
        // replacement loop
        list< FitResult<double, 5, 2> >::const_iterator cosmic_i = 
          cosmics.begin();
        while(cosmic_i != cosmics.end())
          {
            if( is_cosmic(*cosmic_i, flagVal) )
              {
                ++ccounter;
                if(flagVal.verbose > 1) {
                  cerr << " " << setw(6) << ccounter 
                       << " " << setw(6) << (*cosmic_i).coord(1) 
                       << " " << setw(6) << (*cosmic_i).coord(2);
                  if(flagVal.verbose > 2)
                    cerr << "   :  " 
                         << (*cosmic_i).parameter << "; "
                         << (*cosmic_i).paramerror ;
                  cerr << endl;
                }
                pcounter +=
                  replace_cosmic(*cosmic_i, flagVal,
                                 e_b, a, e_a, mask);
              }
            ++cosmic_i;
          }
        cosmics.clear();
        b = a;
        e_b = e_a;

        if( (flagVal.verbose > 0) && (ccounter > 0))
          {
            if(flagVal.mask) cerr << "  Masked ";
            else cerr << "Replaced ";
            cerr << pcounter << " pixels in "
                 << ccounter << " cosmics." << endl;
          }

        ++passcounter;
        totpcounter += pcounter;
        totccounter += ccounter;
        hitcounter = ccounter;

      } while(hitcounter > 0);
         
      ////////////////////////////////// median filter in addition ////////////////////////////
      if(flagVal.median_kappa != 0.0)
        {
          e_a = merge(e_a < 0.0f, 0.0f, e_a);
          do{
            mask = (e_a > 0.0f);
            if(flagVal.verbose > 0) {
              cerr << "------------------------------------------------------------------------" << endl;
              cerr << "Pass " << passcounter << ": Median clipping\n";
            }
            // unmask outside region
            if(fregion.getStart(1) != 0)
              setMAoutsideRegion((unsigned char) 0, mask, fregion);
            // unmask unfittable border
            setMAborder( (unsigned char) 0, mask, flagVal.fit_radius);
            // get local maxima in remaining
            hitcounter = searchlocmax(a.beginRA(), mask.beginRA(),
                                      a.length(1), a.length(2));
            if(flagVal.verbose > 0)
              cerr << "  Found " << hitcounter << " local maxima." << endl;
            // unmask isolated pixel
            if(flagVal.isolation > 0)
              {
                size_t dropcounter = masknanneighbors(a.beginRA(), mask.beginRA(),
                                                      mask.length(1), mask.length(2),
                                                      flagVal.nan_value, flagVal.isolation);
                if(flagVal.verbose > 0)
                  cerr << "  Dropped " << dropcounter << " isolated maxima." << endl;
                hitcounter -= dropcounter;
              }
            if(have_error)
              {
                // unmask pixels with saturated neighbours
                size_t dropcounter = masknanneighbors(e_a.beginRA(), mask.beginRA(),
                                                      mask.length(1), mask.length(2),
                                                      -1.0f, 2);
                if(flagVal.verbose > 0)
                  cerr << "  Dropped " << dropcounter << " close to saturated maxima." << endl;
                hitcounter -= dropcounter;
                const unsigned char uc0 = 0;
                const unsigned char uc1 = 1;
                // unmask pixels with big errors
                if(flagVal.upper_error > 0.0f)
                  {
                    const float e_limit = pow2(flagVal.upper_error / flagVal.gain);
                    mask = merge( (mask != uc0) && (e_a < (e_limit * a)),
                                  uc1, uc0);
                    dropcounter = hitcounter - count(mask);
                    if(flagVal.verbose > 0)
                      cerr << "  Dropped " << dropcounter << " maxima with big errors." << endl;
                    hitcounter -= dropcounter;
                  }
              }
            if(flagVal.verbose > 0)
              cerr << "  Testing " << hitcounter << " local maxima." << endl;
            MArray< unsigned char , 2>::iterator m_i = mask.begin();
            MArray< unsigned char , 2>::IndexIterator i_i = mask.indexBegin();
         
            hitcounter = 0;
            const int box_size = 2 * flagVal.fit_radius + 1;
            MArray<float, 2> ra( box_size, box_size );
            const int center_coord = flagVal.fit_radius+1;
            while(!m_i.done()) // test array once
              {
                if(*m_i){ // check mask
                  *m_i = 0; // this one is done

                  const Range xRange(i_i(1) - flagVal.fit_radius,
                                     i_i(1) + flagVal.fit_radius);
                  const Range yRange(i_i(2) - flagVal.fit_radius,
                                     i_i(2) + flagVal.fit_radius);
                  ra = a(xRange, yRange);
                  ra(center_coord, center_coord) = flagVal.nan_value;
                  const double med = median_exact(ra, flagVal.nan_value);
                  // error estimate from error frame will "generate" cosmics in sources...
                  //if(have_error){
                  //ra = e_a(xRange, yRange); 
                  //ra(center_coord, center_coord) = flagVal.nan_value;
                  //}
                  //const double sig = have_error ?
                  //sqrt( average( ra, flagVal.nan_value ) ) :
                  //stddev(ra, flagVal.nan_value);
                  // ... while just looking at local variations won't be fooled
                  const double sig = stddev(ra, flagVal.nan_value);
                  if( b(i_i(1), i_i(2)) >= med + sig * flagVal.median_kappa)
                    {
                      ++hitcounter;
                      if(flagVal.mask)
                      {
                         b(i_i(1), i_i(2)) = flagVal.nan_value;
                         e_b(i_i(1), i_i(2)) = flagVal.nan_error;
                      }
                      else
                      {
                        b(i_i(1), i_i(2)) = med;
                        e_b(i_i(1), i_i(2)) += sig;
                      }
                    }
                }
                ++m_i; ++i_i;
              }
            ++passcounter;
            if(flagVal.verbose > 0)
              cerr << "  Clipped " << hitcounter << " pixels." << endl;
            totpcounter += hitcounter; totccounter += hitcounter;
          } while(hitcounter > 0);
        }

      --passcounter;

      if(flagVal.verbose > 0)
        cerr << "========================================================================" << endl;

      string result = (flagVal.mask) ?
        string("Masked") :
        string("Replaced");
      {
        char buf[256];
        sprintf(buf, " %lu pixels in %lu cosmics within %lu passes.",
                (unsigned long)(totpcounter),
                (unsigned long)(totccounter),
                (unsigned long)(passcounter));
        result += string(buf);
      }
      infile.addHistory(string(argv[0]) + string(":"));
      infile.addHistory(flags.toString());
      infile.addHistory(result);   
      if(flagVal.verbose > 0)
        {
          cerr << result << endl;
        }

      // compare with original if masking is requested (not replacing)
      if(flagVal.mask)
        {
          if(flagVal.verbose > 2)
            cerr << "Creating masked image by comparing result with original\n";
          infile >> a;
          IndexList<2> tobemasked = where( a!= b);
          b[tobemasked] = flagVal.nan_value;
          e_b[tobemasked] = flagVal.nan_error;
        }
      // save file
      a.free(); e_a.free();
      FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                      infile, flagVal.quiet_please);
      ft.saveImageError2(b, e_b, outfile);

    } // this file should have worked, otherwise ->
    catch (exception &e){
      if(flagVal.verbose > 0)
        cerr << e.what() << endl;
    }
  }

  if(flagVal.verbose > 0)
    cerr << endl;

  return 0;
}


void print_usage(OptionParser& flags)
{
  const string usage =
    string("Usage: cosmicfits [-FLAGS [option]] FILES\n") +
    string("run cosmicfilter on FILES\n\n");
  const string copyright = 
    string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n") +
    string("                        Arno Riffeser <arri@usm.uni-muenchen.de>\n");
  cerr << usage;
  flags.printUsage( cerr );
  cerr << copyright;
  exit(0);
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
  try {
    flags.addOption(new DoubleArrayOption("alamda", "1.0e-5, 12.0, 1.0e-10, 1.0e10",
                                          "start step, step modifier, min. step and max. step for Levenberg-Marquardt fit",
                                          0, 4, flagVal.alamda ) );
    flags.addOption(new IntOption("divergence", "160",
                                  "maximum number of iterations for Levenberg-Marquardt fit",
                                  0, &flagVal.iter_max ) );
    flags.addOption(new DoubleOption("start_fwhm", "1.1",
                                     "start width value for fit (one axis, other is M_2_SQRTPI times)",
                                     0, &flagVal.start_fwhm) );
    flags.addOption(new DoubleOption("circ_start", "3.0",
                                     "start width value for circular refit",
                                     0, &flagVal.circ_start) );
    flags.addOption(new DoubleOption("start_angle", "0.0",
                                     "start angle value for fit",
                                     0, &flagVal.start_angle) );
    flags.addOption(new IntOption("fit_radius", "3",
                                  "fitbox is (2*radius + 1)^2",
                                  0, &flagVal.fit_radius ) );
    flags.addOption(new BoolOption("refit_failed", "FALSE",
                                   "try a circular refit only if elliptical fit failed, default is to refit every no cosmic elliptical",
                                   0, &flagVal.refit_failed ) );
    flags.addOption(new IntOption("isolation", "4",
                                  "do not filter pixels with more than isolation NaN neighbors, set to zero to turn off!",
                                  0, &flagVal.isolation ) );
    flags.addOption(new FloatOption("upper_error", "3.14159265",
                                    "do not filter pixels with error > upper_error * SQRT(signal / gain), set to zero to turn of!",
                                    0, &flagVal.upper_error) );
    flags.addOption(new FloatOption("smooth_scale", "5.0",
                                    "do not filter pixels with value - (smooth_scale * error) < minimum of fitbox, set to zero to turn of!",
                                    0, &flagVal.smooth_scale) );
    //       flags.addOption(new DoubleOption("auto", "0.0",
    //                                        "automatically readjust --cosmic_fwhm, if evaluated histogram is auto times sigma confidence level",
    //                                        'a', &flagVal.auto_adjust ) );
    flags.addOption(new DoubleOption("cosmic_fwhm", "1.5",
                                     //                                        "maximum fwhm for cosmic for smaller axis, default if no auto or auto does not work",
                                     "maximum fwhm for cosmic for smaller axis, set = 0 to turn off elliptical fit",
                                     'c', &flagVal.cosmic_fwhm) );
    flags.addOption(new DoubleOption("circ_fwhm", "1.5",
                                     "fwhm limit for circular refit, set = 0.0 to turn off circular fit",
                                     0, &flagVal.circ_fwhm) );
    flags.addOption(new DoubleOption("confidence_kappa", "1.7",
                                     "apply confidence_kappa times error of fitparameters to fitparameters before evaluating cosmic criteria (+ more save, - more risky)",
                                     0, &flagVal.confidence_kappa) );
    flags.addOption(new FloatOption("gain", "1.0", "gain",
                                    'g', &flagVal.gain) );
    flags.addOption(new DoubleOption("kappa", "5.0",
                                     "minimum amplitude for cosmic: kappa * SQRT(fitssurface / gain)",
                                     'k', &flagVal.kappa) );
    flags.addOption(new DoubleOption("median_kappa", "4.5",
                                     "kappa for median filter, set to 0.0 to turn off median filtering",
                                     0, &flagVal.median_kappa) );
    flags.addOption(new DoubleOption("replace_kappa", "2.0",
                                     "replace pixels where (fit_function - fit_surface) > replace_kappa * SQRT(fit_surface / gain)",
                                     0, &flagVal.replace_kappa) );
    flags.addOption(new IntOption("replace_radius", "5",
                                  "evaluated replacement box is (2*radius + 1)^2",
                                  0, &flagVal.replace_radius ) );
    flags.addOption(new FloatOption("sky_offset", "0.0",
                                    "already subtracted constant sky surface, needed to calculate correct noise estimation from surface fit and gain",
                                    0, &flagVal.sky_offset) );
    flags.addOption(new FloatOption("readnoise", "0.0",
                                    "i.e. noise in bias / overscan region; only needed if it's dominant, e.g. \"empty frames\"",
                                    0, &flagVal.readnoise) );
    flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                          "x_start:x_end,y_start:y_end only filter region",
                                          'r', 2, &flagVal.region) );
    flags.addOption(new StringOption("prefix", "c", "output file prefix",
                                     'p', &flagVal.prefix) );
    flags.addOption(new BoolOption("mask", "FALSE",
                                   "set cosmics to NaN_value in image, default is to replace with surface fit",
                                   'm', &flagVal.mask ) );
    flags.addOption(new BoolOption("ignore_header", "FALSE",
                                   "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                   0, &flagVal.ignore_header ) );
    flags.addOption(new BoolOption("nan", "FALSE",
                                   "keep nan_value in either file dominant",
                                   'n', &flagVal.nan ) );
    flags.addOption(new FloatOption("nan_value", "0.0",
                                    "use value as NaN (see nan/mask option)",
                                    0, &flagVal.nan_value) );
    flags.addOption(new FloatOption("nan_error", "0.0",
                                    "use value as NaN in error image (see nan/mask option)",
                                    0, &flagVal.nan_error) );
    flags.addOption(new IntOption("verbose", "1",
                                  "verbosity level 0 = shut up, 1 standard, etc.",
                                  'v', &flagVal.verbose ) );
  }
  catch( UException e ) {
    cerr << e.what() << endl;
    exit(1);
  }
  
}



