/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: expandfits.cpp 131 2008-12-04 16:48:13Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe of any but output will be of type FT
#define FT float

#include "fitstools.h"


// class for flag values
class flagValues
{
   public:
      flagValues(int dimension) : region(dimension), verbose(1) { };
      int all;
      int left;
      int right;
      int bottom;
      int top;
      FT value;
      FT error;
      Region region; // region class defined in util/option.h
      string prefix;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

int main( int argc, char *argv[] ) 
{
   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);
   if( flagVal.all )
   {
      flagVal.left = flagVal.all; flagVal.right = flagVal.all;
      flagVal.bottom = flagVal.all; flagVal.top = flagVal.all;
   }

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // array for input and output
         MArray<FT, dimension> a;
         string to_history = "";

         // open file
         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.region.getStart(1) > 0)
         {
            infile.setRegion(flagVal.region);
            to_history += string("cut region ") + flagVal.region.toString()
               + string(", ");
         }
         // read data from file
         ft.loadImage(infile, a);
         // prepare output
         FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                         infile, flagVal.quiet_please);
         // correct WCS, i.e. modify CRPIX if applicable
         try{
            outfile.addValueCard("CRPIX1  ",
                                 outfile.getFloat("CRPIX1  ") + double(flagVal.left));
            outfile.addValueCard("CRPIX2  ",
                                 outfile.getFloat("CRPIX2  ") + double(flagVal.bottom));
         }
         catch (exception &e){
            if(flagVal.verbose > 1)
               cerr << "No WCS\n";
         }

         to_history += ft.expandArray( a, flagVal.left, flagVal.right,
                                       flagVal.bottom, flagVal.top, flagVal.value );
         {
            char buf[80];
            sprintf(buf, ", set error of expanded region to %g", flagVal.error);
            to_history += string(buf);
         }
         if(flagVal.verbose > 0)
            cerr << to_history << endl;
         outfile.addHistory(to_history);

         // save file
         ft.saveImage(a, outfile);
         a.free();
         try{
            // open file
            FitsIn inerror( ft.prependFilename(*file_i, "e."),
                            flagVal.quiet_please, true);
            if(flagVal.region.getStart(1) > 0)
               inerror.setRegion(flagVal.region);
            ft.loadImage(inerror, a);            
            ft.expandArray( a, flagVal.left, flagVal.right,
                            flagVal.bottom, flagVal.top, flagVal.error );
            outfile.setFilename(ft.prependFilename(outfile.getFilename(), "e."));
            // save file
            ft.saveImage(a, outfile);
         }
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << "no errorfile\n";
         }

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: expandfits [-FLAGS [option]] FILES\n") +
      string("add left, right, bottom and/or top region or a complete border to image\n") +
      string("negative settings crop image regions\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new IntOption("all", "0",
                                    "add a surrounding border, overwrites left, right, bottom and top settings",
                                    'a', &flagVal.all) );
      flags.addOption(new IntOption("left", "0",
                                    "add a region to the left side of the image",
                                    0, &flagVal.left) );
      flags.addOption(new IntOption("right", "0",
                                    "add a region to the right side of the image",
                                    0, &flagVal.right) );
      flags.addOption(new IntOption("bottom", "0",
                                    "add a region to the bottom side of the image",
                                    0, &flagVal.bottom) );
      flags.addOption(new IntOption("top", "0",
                                    "add a region to the top side of the image",
                                    0, &flagVal.top) );
      flags.addOption(new FloatOption("value", "0.0",
                                      "sets the value of the new region",
                                      0, &flagVal.value) );
      flags.addOption(new FloatOption("value_error", "0.0",
                                      "sets the error of the new region",
                                      0, &flagVal.error) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end expand only region",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new StringOption("prefix", "e", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

