/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: extractfits.cpp 130 2008-12-04 16:35:15Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include "fitstools.h"

// class for flag values
class flagValues
{
   public:
      flagValues() : region(999), verbose(1) { };
      Region region; // region class defined in util/option.h
      string prefix;
      int extctwidth;
      int verbose;
      bool quiet_please;
      bool ignore_header;
      bool no_clean_header;
};

string crpix(const int i);

void clean_header(FitsHeader& header);

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

int main( int argc, char *argv[] ) 
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<float, 2> ft(flagVal.verbose, flagVal.ignore_header);

   // for all FILES
   for( list<string>::const_iterator file_i = files.begin();
        file_i != files.end();
        ++file_i )
   {
      try{
         // open IO-file(s)
         FitsExtensionIn infile(*file_i,
                                      flagVal.quiet_please,
                                      flagVal.ignore_header);
         if(infile.getNaxis() > 0)
         {
            const string outfilename =
               ft.prependFilename(*file_i, flagVal.prefix);
            FitsOut outfile(outfilename, infile, flagVal.quiet_please);
            if(!flagVal.no_clean_header)
               clean_header(outfile);
            // correct WCS, i.e. modify CRPIX if applicable
            try{
               for(int i = 1, j = 1; j <= flagVal.region.getDim(); ++i, ++j)
               {
                  if(i <= flagVal.region.getDim())
                  {
                     if(flagVal.region.getStart(i) == flagVal.region.getEnd(i))
                        --j; // this one will collapse -> replace in next pass
                     else
                     {
                        const double crpix_out =
                           outfile.getFloat(crpix(i)) - double(flagVal.region.getStart(i) - 1);
                        outfile.addValueCard(crpix(j), crpix_out);
                     }
                  }
                  else // erase "trailing" collapsed dims
                     outfile.eraseCard(crpix(j));
               }
            }
            catch (exception &e){
               if(flagVal.verbose > 1)
                  cerr << "No WCS\n";
            }
            

            // set region to be extracted
            Region fregion(flagVal.region.getSlice(1, infile.getNaxis()));
            const bool isFullFrame = ft.adjustRegion2File(infile, fregion);

            const string buf = string("Cut region ") + 
               fregion.toString();
            outfile.addHistory(buf);
            if(flagVal.verbose > 0)
               cerr << buf << " of " << *file_i << endl
                    << "Saving " << outfilename << endl;
            if(!isFullFrame)
               infile.setRegion(fregion);
            infile >> outfile;
            infile.freeData();

            try{
               const string errorname = ft.prependFilename(*file_i, "e.");
               FitsIn inerror(errorname, flagVal.quiet_please, true);

               const string outerrorname = ft.prependFilename(outfilename,"e.");
               outfile.setFilename(outerrorname);

               if(flagVal.verbose > 0)
                  cerr << buf << " of " << errorname << endl
                       << "Saving " << outerrorname << endl;
               if(!isFullFrame)
                  inerror.setRegion(fregion);
               inerror >> outfile;
            }
            catch (exception &e){
               if(flagVal.verbose > 0)
                  cerr << "No errorfile\n";
            }
         }
         else
         {
            if(flagVal.verbose > 0)
               cerr << "Splitting Extensions of "
                    << *file_i << " into files:" << endl;
            try
            {
               int i = 1;
               while(true)
               {
                  if(flagVal.verbose > 0)
                  {
                     cerr << i << " ... ";
                     cerr.flush();
                  }
                  FitsIn extin_i(infile.getNextExtension());
                  if(flagVal.verbose > 0)
                  {
                     cerr << "opened ... ";
                     cerr.flush();
                  }
                  string suffix = "_";
                  if(i > 99 || flagVal.extctwidth>2)
                  {
                     suffix += char( (i/100) + 0x30 );
                     suffix += char( ((i%100) / 10) + 0x30 );
                     suffix += char( (i%10) + 0x30 );
                  }
                  else
                  {
                     if(i > 9 || flagVal.extctwidth>1)
                     {
                        suffix += char( (i / 10) + 0x30 );
                        suffix += char( (i % 10) + 0x30 );
                     }
                     else
                        suffix += char( i + 0x30 );
                  }
                  string outname = *file_i;
                  outname.insert(outname.find_last_of('.'), suffix);
                  FitsOut outfile(outname, extin_i);
                  extin_i >> outfile;
                  if(flagVal.verbose > 0)
                     cerr << "saved\n";
                  ++i;
               }
            }
            catch(FitsException e)
            {
               if(flagVal.verbose > 0)
                  cerr << e.what() << endl;
            }
         }
      }
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }
   if(flagVal.verbose > 0)
      cerr << endl;

   return 0;
}

string crpix(const int i)
{
   char buf[16];
   sprintf(buf, "CRPIX%-3d", i);
   return string(buf);
}

void clean_header(FitsHeader& header)
{
   header.eraseCard("BIASSEC ");
   header.eraseCard("TRIMSEC ");
   header.eraseCard("DATASEC ");
   header.eraseCard("CCDSEC  ");
   header.eraseCard("ORIGSEC ");
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: extractfits [-FLAGS [option]] FILES\n") +
      string("extract extensions or region of FILES\n") +
      string("unset NAXIS ranges default to full range\n") +
      string("NAXIS ranges beyond valid range are clipped\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new RegionArrayOption("region", "",
                                            "NAXIS1_start:NAXIS1_end,...,NAXISnaxis_start:NAXISnaxis_end (region to be extracted)",
                                            'r', 999, &flagVal.region) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new BoolOption("dont_clean_header", "FALSE",
                                     "do not erase the FITS cards BIASSEC, TRIMSEC, DATASEC, CCDSEC and ORIGSEC",
                                     0, &flagVal.no_clean_header ) );
      flags.addOption(new StringOption("prefix", "e", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new IntOption("extctwidth", "0",
                                    "minimal width of extension suffix, i.e. zero padding",
                                    0, &flagVal.extctwidth ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}
