/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: extractkeys.cpp 177 2018-08-01 23:15:21Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2002 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fits.h>
using namespace ltl;
using std::exception;
using std::string;

#include<iostream>
using std::cout;
using std::cerr;
using std::endl;
#include<list>
using std::list;


// structure for flag values
class flagValues
{
   public:
      string keys;
      bool card_comment;
      bool history;
      bool comment;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   // parse comma separated list of FITS-keys
   list<string> fits_keys;
   {
      string str = flagVal.keys, token;
      string::size_type end = 1;

      // comma separated string of strings
      while( end != str.npos )
      {
         end = str.find_first_of( ",", 0 );
         if( end != str.npos )
         {
            token = str.substr( 0, end );
            str.erase( 0, end+1 );
         }
         else
         {
            token = str.substr( 0, str.length() );
            str.erase(0, str.length() );
         }
         fits_keys.push_back(token);
      }
   }

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   cout.precision(17);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);

         // get value of FITS key
         list<string>::const_iterator k_i = fits_keys.begin();
         const list<string>::const_iterator k_end = fits_keys.end();
         
         while( k_i != k_end )
         {
            const string key_i = infile.adjustKeyword(*k_i);
            try
            {               
               const double dbl_key = infile.getFloat( key_i );
               cout << *k_i << '=' << dbl_key;
               if(flagVal.card_comment)
                  cout << " # " << infile.getComment( key_i );
               cout << endl;
            }
            catch(FitsException e)
            {
               try
               {
                  const string str_key = infile.getString( key_i );
                  cout << *k_i << '='
                       << "'" << str_key << "'";
                  if(flagVal.card_comment)
                     cout << " # " << infile.getComment( key_i );
                  cout << endl;
               }
               catch(FitsException e)
               {
                  if(flagVal.verbose > 1)
                     cerr << "key " << *k_i
                          << " not in " << *file_i << endl;
               }
            }
            ++k_i;
         }
         try
         {               
            if(flagVal.history)
            {
               const string history = infile.getHistory();
               string::size_type start=0;
               string::size_type end=0;               
               while( (end=history.find('\n', start)) != string::npos)
               {
                  const string line = string("# ") +
                     history.substr(start, end - start);
                  cout << line.substr(0, line.find_last_not_of(' ', end-1)+1)
                       << endl;
                  start=end+1;
               }
            }
         }
         catch(FitsException e)
         {
            if(flagVal.verbose > 1)
               cerr << "No HISTORY in " << *file_i << endl;
         }
         try
         {               
            if(flagVal.comment)
            {
               const string comment = infile.getHistory();
               string::size_type start=0;
               string::size_type end=0;               
               while( (end = comment.find('\n', start)) != string::npos)
               {
                  const string line = string("# ") +
                     comment.substr(start, end - start);
                  cout << line.substr(0,
                                      line.find_last_not_of(' ', end - 1) + 1)
                       << endl;
                  start = end + 1;
               }
            }
         }
         catch(FitsException e)
         {
            if(flagVal.verbose > 1)
               cerr << "No COMMENT in " << *file_i << endl;
         }


      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: extractkeys -u key1,key2,...,keyn [-FLAGS [option]] FILES\n") +
      string("extract FITS keys from FILES and print them to stdout\n\n");

   cerr << usage;
   flags.printUsage( cerr );
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new StringOption("use_key", "",
                                       "comma separated list of FITS keys to extract",
                                       'u', &flagVal.keys) );
      flags.addOption(new BoolOption("card_comment", "TRUE",
                                     "print CARD comment as comment",
                                     0, &flagVal.card_comment ) );
      flags.addOption(new BoolOption("history", "FALSE",
                                     "print HISTORY as comment",
                                     'h', &flagVal.history ) );
      flags.addOption(new BoolOption("comment", "FALSE",
                                     "print COMMENT as comment",
                                     'c', &flagVal.comment ) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

