/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: fillfits.cpp 177 2018-08-01 23:15:21Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fitsio.h>
#include <ltl/statistics.h>
#include <ltl/linlsqfit.h>
using namespace ltl;

using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues() : region(2) { };
//       bool boolean;
//       int integer;
//       float floatprec;
//       double doubleprec;
//       string str;
      
//       int int_array[5];
//       float float_array[2];
//       double double_array[3];

      Region region; // region class defined in util/region.h

      int radius_limit;
      float min_fill;
      float nan_value;
      float nan_error;
//      bool nan;
      string prefix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   // array holding pixel distances to center of array
   MArray<float, 2>c_dist( Range(-flagVal.radius_limit, flagVal.radius_limit),
                           Range(-flagVal.radius_limit, flagVal.radius_limit) );
   c_dist = sqrt( pow2(indexPosDbl(c_dist, 1)) +  pow2(indexPosDbl(c_dist, 2)) );

   // minimum filling -> filling in parts of unity circle
   flagVal.min_fill *= float(M_PI);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<float, 2> a, e_a;

         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
         infile >> a; infile.freeData();

         bool have_error = true;
         try{
            const string errorname = prepend_filename(*file_i, "e.");
            FitsIn inerror(errorname,
                           flagVal.quiet_please, true);
            if(flagVal.verbose > 0)
               cerr << "Reading " << errorname << endl;
            inerror >> e_a; inerror.freeData();
         }
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << e.what() << endl;
            e_a.makeReference(MArray<float, 2>(a.shape()));
            e_a = merge(a != flagVal.nan_value, 1.0f, flagVal.nan_error);
            have_error = false;
         }


         if(flagVal.verbose > 0)
            cerr << "Filling bad pixels ..." << endl;

         MArray<float, 2> b(a.shape()), e_b(a.shape());
         b = a; e_b = e_a; // copy input data
         // input errors -> errors^2 and all NaN errors -> 0.0f
         e_a = merge(e_a > 0.0f, e_a * e_a, 0.0f);

         const int x0 = (flagVal.region.getStart(1) > b.minIndex(1)) ?
            flagVal.region.getStart(1) : b.minIndex(1) + 1;
         const int x1 = flagVal.region.getEnd(1) &&
            (flagVal.region.getEnd(1) < b.maxIndex(1)) ?
            flagVal.region.getEnd(1) : b.maxIndex(1) - 1;
         const int y0 = (flagVal.region.getStart(2) > b.minIndex(2)) ?
            flagVal.region.getStart(2) : b.minIndex(2) + 1;
         const int y1 = flagVal.region.getEnd(2) &&
            (flagVal.region.getEnd(2) < b.maxIndex(2)) ?
            flagVal.region.getEnd(2) : b.maxIndex(2) - 1;

         // loop over pixels
         const Range x_b(x0, x1);
         const Range y_b(y0, y1);
         MArray<float, 2> rb( b(x_b, y_b) ), e_rb( e_b(x_b, y_b) );
         MArray<float, 2>::iterator b_i(rb.begin()), e_i(e_rb.begin());

         for(int y = y0; y <= y1; ++y)
         {
            for(int x = x0; x <= x1; ++x)
            {
               // needs replacement?
               if( (*b_i == flagVal.nan_value) &&
                   (*e_i == flagVal.nan_error) )
               {
                  if(flagVal.verbose > 1)
                     cerr << "@: " << x << ", " << y << " =\t";
                  bool not_finished = true;
									const int xr0 = (x-flagVal.radius_limit >= b.minIndex(1)) ?
										x-flagVal.radius_limit  : b.minIndex(1) ;
									const int xr1 = (x+flagVal.radius_limit <= b.maxIndex(1)) ?
										x+flagVal.radius_limit  : b.maxIndex(1) ;
									const int yr0 = (y-flagVal.radius_limit >= b.minIndex(2)) ?
										y-flagVal.radius_limit  : b.minIndex(2) ;
									const int yr1 = (y+flagVal.radius_limit <= b.maxIndex(2)) ?
										y+flagVal.radius_limit  : b.maxIndex(2) ;
                  if (sum(merge(a(Range(xr0,xr1),Range(yr0,yr1))!=flagVal.nan_value,1,0))>0) {
                    int radius = 1;
                    // while not replaced and radius <= maximum
                    while(not_finished)
                    {
                       const int xstart = x - radius;
                       const int xend   = x + radius;
                       const int ystart = y - radius;
                       const int yend   = y + radius;
                       // radius grown beyond image?
                       not_finished = (xstart >= b.minIndex(1)) &&
                          (ystart >= b.minIndex(2)) &&
                          (xend <= b.maxIndex(1)) &&
                          (yend <= b.maxIndex(2));
                       if(not_finished)
                       {
                          // get subarray
                          const Range xrange(xstart, xend);
                          const Range yrange(ystart, yend);
                          MArray<float, 2> box(a(xrange, yrange));
                          const MArray<float, 2> e_box(e_a(xrange, yrange));
                          // make circular stencil utilizing the error array
                          const Range crange(-radius, radius);
                          const MArray<float, 2> e_stencil( c_dist(crange, crange) );
                          const float r2 = float(radius * (radius + 1));
                          const float r_lim = float(sqrt(r2));
                          const MArray<float, 2> e_circ =
                             (e_stencil <= r_lim) * // in circle
                             e_stencil * e_box;
                          const int valid = count(e_circ);
                          // need more than 3 valid pixels for 3 parameter fit
                          // also need pixel to be not in one "row"
                          // and in addition must have more than filling minimum
                          const bool enough =
                             (valid > (1 + 2 * radius)) &&
                             (valid >= ((flagVal.min_fill * r2) - 1.0f));
                          if(enough)
                          {
                             try{
                                // make the fit
                                box.setBase(-radius, -radius);
                                FVector<float, 3> result = 
                                   PolynomFit<float, float, 1, false, 2>::eval(box,
                                                                               e_circ);
                                *b_i = result(1); // solution at coord. (0, 0)
                                *e_i = sqrt( sum(e_circ) / valid );
                                not_finished = false;
                             }
                             catch(LinearAlgebraException e)
                             {
                                // radius at maximum?
                                ++radius;
                                not_finished = (radius <= flagVal.radius_limit);
                             }
                          }
                          else {
                             // radius at maximum?
                             ++radius;
                             not_finished = (radius <= flagVal.radius_limit);
                          }
                       }
										}
                  }
                  if(flagVal.verbose > 1)
                     cerr << b(x, y) << endl;                  
               } // end pixel == NaN
               ++b_i;
               ++e_i;
            } // end x
         } // end y

         a.free();
         e_a.free();

         const string outname = prepend_filename(*file_i, flagVal.prefix);
         FitsOut outfile(outname,
                         infile, flagVal.quiet_please);
         outfile.addHistory(string("interpolated bad pixels") + *file_i);
         if(flagVal.verbose > 0)
            cerr << "Saving " << outname << endl;
         outfile << b;

         if(have_error)
         {
            const string errorname = prepend_filename(outname, "e.");
            outfile.setFilename(errorname);
            if(flagVal.verbose > 0)
               cerr << "Saving " << errorname << endl;
            outfile << e_b;
         }

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: fillfits [-FLAGS [option]] FILES\n") +
      string("replace NaN pixels by interpolation of neighbours\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "perform only within region",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new IntOption("radius_limit", "12",
                                    "maximum radius for fit stencil",
                                    'l', &flagVal.radius_limit ) );
      flags.addOption(new FloatOption("min_fill", "0.6",
                                      "minimum filling factor to try interpolation",
                                      'm', &flagVal.min_fill) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as NaN",
                                      0, &flagVal.nan_value) );
      flags.addOption(new FloatOption("nan_error", "0.0",
                                      "use value as error NaN ",
                                      0, &flagVal.nan_error) );
      flags.addOption(new StringOption("prefix", "i", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

