/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: fits2jpeg.cpp 175 2018-07-31 12:53:05Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fitsio.h>
#include <ltl/statistics.h>
using namespace ltl;

#include <string>
#include <iostream>
using namespace std;

#include <cstdio>

extern "C" {
#include <jpeglib.h>
#include <jerror.h>
//#include <jconfig.h>
#ifndef JPEG_MAX_DIMENSION // hack to avoid double inclusion of jmorecfg in unguarded implementations
#include <jmorecfg.h>
#endif
}


// structure for flag values
class flagValues
{
   public:
      flagValues(const int dimension) : region(dimension) { };
      float nan_value;
      bool nan;
      bool invert;
      bool autocuts;
      double logscale;
      double cuts[2];
      int quality;
      Region region; // region class defined in util/option.h
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string jpg_filename(string path)
{
   path.erase(path.find_last_of('.') + 1);
   path.append(string("jpg"));
   return path;
}

int main(int argc, char* argv[])
{
   // structure for flag values
   flagValues flagVal(2);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if( (flagVal.cuts[0] == 0.0) &&
       (flagVal.cuts[1] == 0.0) &&
       (flagVal.logscale == 0.0) )
   {
      flagVal.autocuts = true;
      flagVal.logscale = 10.0;
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   if(flagVal.verbose > 0)
      cerr << argv[0] << ":" << endl;

   // cuts ok?
   if( !flagVal.autocuts && (flagVal.cuts[0] == flagVal.cuts[1]) )
   {
      cerr << "Either select autocuts or set 'reasonable' cutlevels!\n";
      exit(-1);
   }
   

   // variables for scaling
   const double numcolor = 255.0;
   double diff = flagVal.cuts[1] - flagVal.cuts[0];;
   
   // with fixed cutlevels they must only be set once
   if(!flagVal.autocuts)
   {
      if(flagVal.verbose > 0)
         cerr << "Using fixed cuts ["
              << flagVal.cuts[0] << ", " 
              << flagVal.cuts[1] << "]."
              << endl;
      if(diff == 0.0) diff = 1.0;
   }

   if(flagVal.verbose > 0)
   {
      cerr << "Scaling ";
      if(flagVal.logscale == 0.0)
         cerr << " linear." << endl;
      else
      {
         if(flagVal.logscale > 0.0)
            cerr << " logarithmic low signal enhancement of " 
                 << flagVal.logscale << "." << endl;
         else
            cerr << " with x^" << -flagVal.logscale << "." << endl;
      }
   }
   
   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // array for input
         MArray<float, 2> a;
         FitsIn infile(*file_i, flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.region.getStart(1) > 0)
            infile.setRegion(flagVal.region);

         // if no autocuts, scale while reading
         // scale to [0, 1] range
         if(!flagVal.autocuts)
         {
            const double bzero = infile.getBzero() - flagVal.cuts[0];
            const double bscale = infile.getBscale();
            if(bzero != 0.0)
               infile.setBzero(bzero / diff);
            infile.setBscale(bscale / diff);
         }
         
         // read file
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
         infile >> a;
         infile.freeData();

         // if autocuts, scale now ... use statistics
         // scale to [0, 1] range
         if(flagVal.autocuts)
         {
            double median, sigma, min;
            float fmin, fdiff;
            // get statistics
            if(flagVal.nan)
               kappa_sigma_median(a, 6.0, flagVal.nan_value, &median, &sigma);
            else
               kappa_sigma_median(a, 6.0, &median, &sigma);

            // full auto cuts
            if( (flagVal.cuts[0] == 0.0) && (flagVal.cuts[1] == 0.0) )
            {
               min = median - 0.5 * sigma;
               diff = 5.5 * sigma;
            }
            // auto cuts using cutlevels as median +- cut*sigma
            else
            {
               min = median + flagVal.cuts[0] * sigma;
               diff = (flagVal.cuts[1] - flagVal.cuts[0]) * sigma;
            }
            fdiff = float(diff) + (diff == 0.0) * 1.0f;
            fmin = float(min);
            if(flagVal.verbose > 0)
               cerr << "Using auto cuts ["
                    << fmin << ", " 
                    << fmin + fdiff << "]."
                    << endl;
            a = (a - fmin) / fdiff;
         }
         
         // clip to scaled cuts [0, 1]
         MArray<float, 2>::iterator a_i = a.begin();
         while(!a_i.done())
         {
            if(*a_i < 0.0f) *a_i = 0.0f;
            else if(*a_i > 1.0f) *a_i = 1.0f;
            ++a_i;
         }
         a.reverseSelf(2);

         // build byte array
         // since jpeglib is plain C code we have to use a
         // char array for the data, but for easier access we also
         // wrap it in a MArray
         unsigned char* const chardata = new unsigned char [a.nelements()];
         // charlines are needed by jpeglib
         unsigned char** const charlines = new unsigned char* [a.length(2)];
         {
            unsigned char* const enddata = chardata + a.nelements();
            unsigned char* lineptr = chardata;
            unsigned char** line_i = charlines;
            while(lineptr != enddata)
            {
               *line_i = lineptr;
               ++line_i;
               lineptr += a.length(1);
            }
         }
         int dim_ [2] = {a.length(1), a.length(2)};
         MArray<unsigned char, 2> c(chardata, dim_);

         // now calculate image
         // logarithmic brightness scaling
         if(flagVal.logscale != 0.0)
         {
	    if(flagVal.logscale > 0.0)
            {
               const double upscale = numcolor / log(1.0 + flagVal.logscale);
               c = cast<unsigned char>()( log( a * flagVal.logscale + 1.0) * upscale );
            }
	    else
            {
               const double upscale =
                  numcolor / (pow(2.0, - flagVal.logscale) - 1.0);
               c = cast<unsigned char>()( (pow(a + 1.0, -flagVal.logscale) - 1.0) * upscale );
            }
         }
         else
            c = cast<unsigned char>()(numcolor * a);

         if(flagVal.invert)
            c = ~c;

         const string outfilename = jpg_filename(*file_i);
         if(flagVal.verbose > 0)
            cerr << "Writing " << outfilename << endl;
          
         struct jpeg_compress_struct cinfo;
         struct jpeg_error_mgr jerr;
	    
         cinfo.err = jpeg_std_error(&jerr);
         jpeg_create_compress(&cinfo);
	    
         FILE * outfile;
         if ((outfile = fopen(outfilename.c_str(), "wb")) == NULL)
            throw IOException(string("Can't open ") + outfilename);

         jpeg_stdio_dest(&cinfo, outfile);
         cinfo.image_width = a.length(1); /* image width and height, in pixels */
         cinfo.image_height = a.length(2);
         cinfo.input_components = 1;	/* # of color components per pixel */
         cinfo.in_color_space = JCS_GRAYSCALE; /* colorspace of input image */
	    
         jpeg_set_defaults(&cinfo);
         jpeg_set_quality (&cinfo, flagVal.quality, FALSE);
	    
         jpeg_start_compress(&cinfo, TRUE);
	    
         if( jpeg_write_scanlines(&cinfo, charlines, a.length(2)) < 
             (unsigned int)a.length(2) )
               if(flagVal.verbose > 0)
                  cerr << "Have not written all lines.\n";
         
         jpeg_finish_compress(&cinfo);
         fclose(outfile);
	    
         // clean up
         jpeg_destroy_compress(&cinfo);
         
         // chardata should be deleted by MArray destructor
         delete [] charlines;

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: fits2jpeg [-FLAGS [option]] FILES\n") +
      string("Convert FITS FILES to JPEG images.\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "ingore nan_value when calculating autocuts",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new BoolOption("invert", "FALSE",
                                     "invert image",
                                     'i', &flagVal.invert ) );
      flags.addOption(new BoolOption("autocuts", "FALSE",
                                     "automatically set cuts (is also turned on by giving no cuts)",
                                     'a', &flagVal.autocuts ) );
      flags.addOption(new DoubleOption("logscale", "0.0",
                                       "logscale: 0 = linear, >0 = logarithmic, <0 = rootscaling",
                                       'l', &flagVal.logscale) );
      flags.addOption(new DoubleArrayOption("cuts", "0.0,0.0",
                                            "absolute (or with auto sigma to median) cuts: locut,hicut",
                                            'c', 2, flagVal.cuts ) );
      flags.addOption(new IntOption("quality", "90",
                                    "quality of jpeg compression",
                                    'q', &flagVal.quality ) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end only on region",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

