/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: templatefits.cpp 94 2005-02-21 12:18:38Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#include <ltl/fitsio.h>
#include <ltl/statistics.h>

//#include <fstream>

using namespace ltl;
using namespace util;
using namespace std;

// structure for flag values
class flagValues
{
   public:

      float nan_value;
      bool nan;
      string postfix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
// inline string postpend_filename(string path, const string& prefix)
// {
//    path.replace(path.find_last_of('.') + 1, prefix);
//    return path;
//}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<int, 2> a;

         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
         infile >> a; infile.freeData();
         
         const int minimum = min(a);
         const int maximum = max(a);
         MArray<int, 1> hist = histogram(a, maximum - minimum + 2, minimum, maximum+1);
         
         MArray<int, 1>::const_iterator h_i = hist.begin();
         int i = minimum;
         while(!h_i.done())
         {
            cout << i << '\t' << *h_i << endl;
            ++i;
            ++h_i;
         }

         //const string outname = postpend_filename(*file_i, flagVal.prefix);
         a.free();

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: fitshist [-FLAGS [option]] FILE(s)\n") +
      string("print a histogram (value <TAB> count) to stdout.\n\n");
   const string copyright = 
      string("Copyright (C) 2009 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
//       flags.addOption(new StringOption("postfix", "tab", "output file postfix",
//                                        'p', &flagVal.postfix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

