/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: fitstools.h 143 2010-04-26 17:30:18Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// functions suitable for many parts of the fitstools

#ifndef __FITSTOOLS_H__
#define __FITSTOOLS_H__

#include <ltl/fitsio.h>
#include <ltl/marray.h>
#include <ltl/statistics.h>

#ifdef LTL_USING_NAMESPACE
using namespace ltl;
#endif

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#ifdef UTIL_USING_NAMESPACE
using namespace util;
#endif

#include <string>
#include <iostream>
#include <list>
using namespace std;

#include <cstdio>


// class for image/error pixel
template<class T>
class image_error_pixel
{
   public:
      image_error_pixel(const T image, const T error) :
         image_(image), error_(error) { };
      image_error_pixel() :
         image_(T(0)), error_(T(0)) { };
      T image_;
      T error_;
   protected:
};

// help class to sort pixel including errors
template<class T>
class less_iep // : public binary_function<T, T, T>
{
   public:
      less_iep()
      { };
      inline bool operator()(image_error_pixel<T> i,
                             image_error_pixel<T> j) const
      {
         return ( ((i.image_ <= j.image_) && (i.error_ > T(0))) ||
                  (j.error_ <= T(0)) );
      }
};

// help class to sort pixels including errors and nan
template<class T>
class less_iep_nan// : public binary_function<T, T, T>
{
   public:
      less_iep_nan(const T& nan)
         : nan_(nan)
      { };
      inline bool operator()(image_error_pixel<T> i,
                             image_error_pixel<T> j) const
      {
         return ( ((i.image_ <= j.image_) && (i.image_ != nan_) && 
                   (i.error_ > T(0))) ||
                  (j.image_ == nan_) || (j.error_ <= T(0) ));
      }
   private:
      const T nan_;
};


// class for fitstools methods
template<class T, int N>
class FitsTools
{
   public:
      FitsTools(const int verbose = 1, const bool ignore_header = false) :
         verbose_(verbose), quiet_please_( verbose == 0 ),
         ignore_header_(ignore_header)
      { };
      
// prototypes start
      void addArray(MArray<T, N>& a, const MArray<T, N>& b,
                    const bool nan = false, const T nan_value = T(0));

      void addArrayConstant(MArray<T, N>& a,
                            const MArray<T, N>& b, const T s,
                            const bool nan = false,
                            const T nan_value = T(0));

      void addPhotonNoise2(MArray<T, N>& e_a,
                           const MArray<T, N>& a,
                           const T g = T(1.0),
                           const bool nan = false,
                           const T nan_value = T(0.0));

      void addZeroPhotonNoise2(MArray<T, N>& e_a,
                               const MArray<T, N>& a,
                               const T g = T(1.0),
                               const bool nan = false,
                               const T nan_value = T(0.0));

      bool adjustRegion2File(const FitsIn& infile, Region& region);

      string averageArray(double& mean, double& sigma,
                          const MArray<T, N>& m, const double kappa,
                          const bool sigma_flag = false,
                          const bool twist =false,
                          const bool nan = false,
                          const T nan_value = T(0));

      void convert2AbsoluteError2(MArray<T, N>& e_a,
                                  const MArray<T, N>& a,
                                  const bool nan = false,
                                  const T nan_value = T(0.0));

      void convert2RelativeError2(MArray<T, N>& e_a,
                                  const MArray<T, N>& a,
                                  const bool nan = false,
                                  const T nan_value = T(0.0));

      void divideArrayConstant(MArray<T, N>& a,
                               const MArray<T, N>& b, const T d,
                               const bool nan = false,
                               const T nan_value = T(0));

      string expandArray(MArray<T, N>& a,
                         const int left, const int right,
                         const int bottom, const int top,
                         const T value);

      bool loadError2(FitsIn& infile, MArray<T, N>& error);

      bool loadError2(const string& filename, MArray<T, N>& error);

      bool loadError2(const string& filename, MArray<T, N>& error,
                      const Region& region);

      void loadImage(FitsIn& infile, MArray<T, N>& image);

      void loadImage(const string& filename, MArray<T, N>& image);

      void loadImage(const string& filename, MArray<T, N>& image,
                     const Region& region);

      bool loadImageError2(FitsIn& infile,
                           MArray<T, N>& image,
                           MArray<T, N>& error);

      bool loadImageError2(const string& filename,
                           MArray<T, N>& image,
                           MArray<T, N>& error);

      bool loadImageError2(const string& filename,
                           MArray<T, N>& image,
                           MArray<T, N>& error,
                           const Region& region);

      void loadImageOffset(FitsIn& infile, const double offset,
                           MArray<T, N>& image);

      string medianArray(double& median, double& sigma,
                         const MArray<T, N>& m, const double kappa,
                         const bool sigma_flag = false,
                         const bool twist = false,
                         const bool nan = false,const T nan_value = T(0));

      void multiplyArrayConstant(MArray<T, N>& a,
                                 const MArray<T, N>& b,
                                 const T d,
                                 const bool nan, const T nan_value);

      void multiplyConstant(MArray<T, N>& a, const T d,
                            const bool nan = false,
                            const T nan_value = T(0));

      string prependFilename(string path, const string& prefix) const;

      void propagateError2(MArray<T, N>& e_a,
                           const MArray<T, N>& e_b,
                           const T e_s);

      void propagateZeroError2(MArray<T, N>& e_a,
                               const MArray<T, N>& e_b,
                               const T e_s);

      void referenceRegion(MArray<T, N>& ref,
                           const MArray<T, N>& input,
                           const Region& region);

      string rotateReverseArray(MArray<T, N>& a,
                                const int rot, const bool x, const bool y);

// changes error! error -> sqrt(error)
      void saveError2(MArray<T, N>& error, FitsOut& outfile);

      void saveImage(MArray<T, N>& image, FitsOut& outfile);

      void saveImageError2(MArray<T, N>& a, MArray<T, N>& e_a,
                           FitsOut& outfile);

      void selectRegionFile(FitsIn& infile, Region region);

      void subtractArrayConstant(MArray<T, N>& a,
                                 const MArray<T, N>& b, const T s,
                                 const bool nan = false,
                                 const T nan_value = T(0));

// prototypes end
   private:
      const int verbose_;
      const bool quiet_please_;
      const bool ignore_header_;
};



// fill in prefix to filename for given path
template<class T, int N>
inline string FitsTools<T, N>::prependFilename(string path,
                                               const string& prefix) const
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}


template<class T, int N>
string FitsTools<T, N>::averageArray(double& mean, double& sigma,
                                     const MArray<T, N>& m,
                                     const double kappa,
                                     const bool sigma_flag, const bool twist,
                                     const bool nan, const T nan_value)
{
   char buf[128];
   string result = "";

   const int n_valid = (nan) ?
      count(m != nan_value) :
      m.nelements();

   if(kappa) // clipped
   {
      int n; // for number of clipping remnants
      if(nan) // with nan
      {
         if(twist)
            n = kappa_median_average( m, kappa, nan_value, &mean, &sigma );
         else
            n = kappa_sigma_average( m, kappa, nan_value, &mean, &sigma );
      }      
      else // without nan
      {
         if(twist)
            n = kappa_median_average( m, kappa, &mean, &sigma );
         else
            n = kappa_sigma_average( m, kappa, &mean, &sigma );
      }

      const double kappa_correction( erf( kappa / M_SQRT2 ) );
      sigma /= kappa_correction;
      sprintf(buf, "clipping left %i of %i values, ", n, n_valid);
      result += string(buf);
   }
   else // not clipped
   {
      if(nan) // with nan
         sigma = stddev( m, nan_value, &mean );
      else sigma = stddev( m, &mean ); // without nan
   }
   
   if( !sigma_flag ) sigma /= sqrt(double(n_valid)); // use SEM
   //else sigma += sigma / sqrt(double(m.nelements())); // use stddev + SEM
   
   sprintf( buf, "mean %g with error %g",
            mean, sigma );
   result += string(buf);
   return result;
}

template<class T, int N>
string FitsTools<T, N>::medianArray(double& median, double& sigma,
                                    const MArray<T, N>& m,
                                    const double kappa,
                                    const bool sigma_flag, const bool twist,
                                    const bool nan, const T nan_value)
{
   char buf[128];
   string result = "";

   const int n_valid = (nan) ?
      count(m != nan_value) :
      m.nelements();

   if(kappa) // clipped
   {
      int n; // for number of clipping remnants
      if(nan) // with nan
      {
         if(twist)
            n = kappa_average_median( m, kappa, nan_value, &median, &sigma );
         else
            n = kappa_sigma_median( m, kappa, nan_value, &median, &sigma );
      }
      else // without nan
      {
         if(twist)
            n = kappa_average_median( m, kappa, &median, &sigma );
         else
            n = kappa_sigma_median( m, kappa, &median, &sigma );
      }
      
      const double kappa_correction( erf( kappa / M_SQRT2 ) );
      sigma /= kappa_correction;
      sprintf(buf, "clipping left %i of %i values, ", n, n_valid);
      result += string(buf);
   }
   else // not clipped
   {
      const double n_m1 = double(n_valid - 1);
      if(n_m1 > 0.0)
      {
         if(nan) // with nan
         { 
            median = median_exact( m, nan_value );
            sigma =
               sqrt( sum( merge(m != nan_value,
                                pow2(m - median),
                                T(0.0)) ) / n_m1 );
         }
         else // without nan
         {
            median = median_exact( m );
            sigma = sqrt( sum( pow2(m - median) ) / n_m1 );
         }
      }
      else
         sigma = 0.0;
   }
   
   if( !sigma_flag ) sigma /= sqrt(double(n_valid)); // use SEM
   //else sigma += sigma / sqrt(double(m.nelements())); // use stddev + SEM

   sprintf( buf, "median %g with error %g", median, sigma );
   result += string(buf);
   return result;
}

template<class T, int N>
void FitsTools<T, N>::addArray(MArray<T, N>& a,
                               const MArray<T, N>& b,
                               const bool nan, const T nan_value)
{
   if( !a.isConformable(b) ) // do sizes match?
      throw (RangeException("NAXIS parameter of files do not match"));
   if( nan ) // with nan
   {
      a = merge( ( a != nan_value ) && (b != nan_value ),
                 a + b, nan_value);
   }
   else // without nan
      a += b;
}

template<class T, int N>
void FitsTools<T, N>::divideArrayConstant(MArray<T, N>& a,
                                          const MArray<T, N>& b,
                                          const T d,
                                          const bool nan, const T nan_value)
{
   // calculate divided file
   const bool divbyconst = d && (d  !=  T(1.0));
   
   if( b.isAllocated() ) // divide by file?
   {
      if( !a.isConformable(b) ) // do sizes match?
         throw (RangeException("NAXIS parameter of files do not match"));
      if( nan ) // with nan
      {
         if ( divbyconst ) // divide by constant?
            a = merge( ( a != nan_value ) && (b != nan_value ) && (b != T(0)),
                       a / (b * d), nan_value);
         else // no constant!
            a = merge( ( a != nan_value ) && (b != nan_value ) && (b != T(0)),
                       a / b, nan_value);
      }
      else // without nan
      {
         if( divbyconst )
            a = merge( b != T(0), a / (b * d), a); // divide constant?
         else
            a = merge( b != T(0), a / b , a); // no constant!
      }
   }
   else
   {
      if( divbyconst )
      { // no file, but divide by constant?
         if( nan ) a = merge(a != nan_value, a / d, nan_value);
         else a /= d;
      }
   }
}

template<class T, int N>
void FitsTools<T, N>::multiplyArrayConstant(MArray<T, N>& a,
                                            const MArray<T, N>& b,
                                            const T d,
                                            const bool nan, const T nan_value)
{
   // calculate multiplied file
   const bool mulbyconst = d  !=  T(1.0);
   
   if( b.isAllocated() ) //  multiply with file?
   {
      if( !a.isConformable(b) ) // do sizes match?
         throw (RangeException("NAXIS parameter of files do not match"));
      if( nan ) // with nan
      {
         if ( mulbyconst ) // multiply with constant?
            a = merge( ( a != nan_value ) && (b != nan_value ),
                       a * b * d, nan_value);
         else // no constant!
            a = merge( ( a != nan_value ) && (b != nan_value ),
                       a * b, nan_value);
      }
      else // without nan
      {
         if( mulbyconst ) a *= b * d; // multiplay with constant?
         else a *= b; // no constant!
      }
   }
   else
   {
      if( mulbyconst )
      { // no file, but multiply with constant?
         if( nan ) a = merge(a != nan_value, a * d, nan_value);
         else a *= d;
      }
   }
}

template<class T, int N>
void FitsTools<T, N>::addArrayConstant(MArray<T, N>& a,
                                       const MArray<T, N>& b,
                                       const T s,
                                       const bool nan, const T nan_value)
{
   // calculate added file
   if( b.isAllocated() ) // add file?
   {
      if( !a.isConformable(b) ) // do sizes match?
         throw (RangeException("NAXIS parameter of files do not match"));
      if( nan ) // with nan
      {
         if ( s ) // add constant?
            a = merge( (a != nan_value) && (b != nan_value),
                       a + b + s, nan_value);
         else // no constant!
            a = merge( (a != nan_value) && (b != nan_value),
                       a + b, nan_value);
      }
      else // without nan
      {
         if( s ) a += b + s; // add constant?
         else a += b; // no constant!
      }
   }
   else
      if( s )
      { // no file, but add constant?
         if( nan )
            a = merge( a != nan_value, a + s, nan_value );
         else
            a += s;
      }
}

template<class T, int N>
void FitsTools<T, N>::subtractArrayConstant(MArray<T, N>& a,
                                            const MArray<T, N>& b,
                                            const T s,
                                            const bool nan, const T nan_value)
{
   // calculate difference file
   if( b.isAllocated() ) // subtract file?
   {
      if( !a.isConformable(b) ) // do sizes match?
         throw (RangeException("NAXIS parameter of files do not match"));
      if( nan ) // with nan
      {
         if ( s ) // subtract constant?
            a = merge( (a != nan_value) && (b != nan_value),
                       a - (b + s), nan_value);
         else // no constant!
            a = merge( (a != nan_value) && (b != nan_value),
                       a - b, nan_value);
      }
      else // without nan
      {
         if( s ) a -= b + s; // subtract constant?
         else a -= b; // no constant!
      }
   }
   else
      if( s )
      { // no file, but subtract constant?
         if( nan )
            a = merge( a != nan_value, a - s, nan_value);
         else
            a -= s;
      }
}

template<class T, int N>
inline void FitsTools<T, N>::multiplyConstant(MArray<T, N>& a,
                                              const T d,
                                              const bool nan,
                                              const T nan_value)
{
   if( nan )
      a = merge(a != nan_value, a * d, nan_value);
   else a *= d;
}


// add linear error^2 of two arrays and one constant
// store result to 1st array
template<class T, int N> // keep zeros dominant
void FitsTools<T, N>::propagateError2(MArray<T, N>& e_a,
                                      const MArray<T, N>& e_b,
                                      const T e_s)
{
   if(e_a.isAllocated()) // in/out error present
   {
      if(e_b.isAllocated()) // to be added error present
      {
         if( e_s ) // error of constant present
            e_a = merge( (e_a > T(0)) && (e_b > T(0)),
                         e_a + e_b + e_s,
                         merge(e_b < e_a, e_b, e_a) );
         else // no error of constant
            e_a = merge( (e_a > T(0)) && (e_b > T(0)),
                         e_a + e_b,
                         merge(e_b < e_a, e_b, e_a) );
      }
      else if( e_s ) // add only error of constant
         e_a = merge(e_a > T(0), e_a + e_s, e_a);
   }
   else // input error missing => create new error
      throw(RangeException("no output array allocated in error propagation"));
}
template<class T, int N> // add on zero errors
void FitsTools<T, N>::propagateZeroError2(MArray<T, N>& e_a,
                                          const MArray<T, N>& e_b,
                                          const T e_s)
{
   if(e_a.isAllocated()) // in/out error present
   {
      if(e_b.isAllocated()) // to be added error present
      {
         if( e_s ) // error of constant present
            e_a = merge( (e_a >= T(0)) && (e_b >= T(0)),
                         e_a + e_b + e_s,
                         merge(e_b < e_a, e_b, e_a) );
         else // no error of constant
            e_a = merge( (e_a >= T(0)) && (e_b >= T(0)),
                         e_a + e_b,
                         merge(e_b < e_a, e_b, e_a) );
      }
      else if( e_s ) // add only error of constant
         e_a = merge(e_a >= T(0), e_a + e_s, e_a);
   }
   else // input error missing => create new error
      throw(RangeException("no output array allocated in error propagation"));
}


// e_a = e_a * a * a except for
// e_a < 0 or a == nan_value
template<class T, int N>
void FitsTools<T, N>::convert2AbsoluteError2(MArray<T, N>& e_a,
                                             const MArray<T, N>& a,
                                             const bool nan,
                                             const T nan_value)
{
   if( nan )
      e_a *= ((a != nan_value) && (e_a > T(0))) * pow2(a) +
         (e_a <= T(0)) * T(1.0);
   else
      e_a *= (e_a > T(0)) * pow2(a) + (e_a < T(0)) * T(1.0);
}
// e_a = e_a / (a * a) except for
// e_a < 0, a == 0 or a == nan_value
template<class T, int N>
void FitsTools<T, N>::convert2RelativeError2(MArray<T, N>& e_a,
                                             const MArray<T, N>& a,
                                             const bool nan,
                                             const T nan_value)
{
   if( nan )
      e_a = merge( (a != nan_value) && (a != T(0)) && (e_a > T(0)),
                   e_a / pow2(a),
                   ( (a != T(0)) && (e_a > T(0)) ) * nan_value +
                   ( e_a < T(0) ) * e_a);
   else
      e_a = merge( (a != T(0)) && (e_a > T(0)),
                   e_a / pow2(a),
                   ( e_a < T(0) ) * e_a);
}

// determine and add photon noise
template<class T, int N>
void FitsTools<T, N>::addPhotonNoise2(MArray<T, N>& e_a,
                                      const MArray<T, N>& a,
                                      const T g,
                                      const bool nan, const T nan_value)
{
   if(g != T(1))
   {
      if(nan)
         e_a += ( (a != nan_value) && (a > T(0)) && (e_a > T(0)) ) * (a/g);
      else
         e_a += ( (a > T(0)) && (e_a > T(0)) ) * (a/g);
   }
   else
   {
      if(nan)
         e_a += ( (a != nan_value) && (a > T(0)) && (e_a > T(0)) ) * a;
      else
         e_a += ( (a > T(0)) && (e_a > T(0)) ) * a;
   }
}
template<class T, int N>
void FitsTools<T, N>::addZeroPhotonNoise2(MArray<T, N>& e_a,
                                          const MArray<T, N>& a,
                                          const T g,
                                          const bool nan, const T nan_value)
{
   if(g != T(1))
   {
      if(nan)
         e_a += ( (a != nan_value) && (a > T(0)) && (e_a >= T(0)) ) * (a/g);
      else
         e_a += ( (a > T(0)) && (e_a >= T(0)) ) * (a/g);
   }
   else
   {
      if(nan)
         e_a += ( (a != nan_value) && (a > T(0)) && (e_a >= T(0)) ) * a;
      else
         e_a += ( (a > T(0)) && (e_a >= T(0)) ) * a;
   }
}

template<class T, int N>
string FitsTools<T, N>::expandArray(MArray<T, N>& a, 
                                    const int left, const int right,
                                    const int bottom, const int top,
                                    const T value)
{

   const int x = a.length(1);
   const int y = a.length(2);

   const int new_x = x + left + right;
   const int new_y = y + bottom + top;
   
   const bool expand_bottom = (bottom > 0);
   const bool expand_left = (left > 0);
   const bool expand_right = (right > 0);
   const bool expand_top = (top > 0);

   MArray<T, N> new_a(new_x, new_y);
   
   if(expand_bottom)
      new_a( Range::all(), Range( 1, bottom ) ) = value;
   if(expand_left)
      new_a( Range( 1, left ), Range::all() ) = value;
   if(expand_right)
      new_a( Range( (new_x - right + 1), new_x ), Range::all() ) = value;
   if(expand_top)
      new_a( Range::all(), Range( (new_y - top + 1), new_y ) ) = value;
   
   new_a( Range( (1 + expand_left * left), (new_x - expand_right * right) ),
          Range( (1 + expand_bottom * bottom), (new_y - expand_top * top) )) =
      a( Range( (1 - (!expand_left) * left), (x + (!expand_right) * right) ),
         Range( (1 - (!expand_bottom) * bottom), (y + (!expand_top) * top) ));
   

   a.makeReference(new_a);

   string history = "";
   char buf[128];
   
   if( (left == right) && (left == bottom) && (left == top) )
   {
      sprintf(buf, "added a %d pixel border with value = %g", left, value);
      history += string(buf);
   }
   else
   {
      if(left)
      {
         if(expand_left)
            sprintf(buf, "added %d pixel with value = %g region to the left side of the image", left, value);
         
         else sprintf(buf, "cropped %d pixel region from the left side of the image", 0-left);
         history += string(buf);
         if( right || bottom || top ) history += string(", ");
      }
      if(right)
      {
         if(expand_right)
            sprintf(buf, "added %d pixel with value = %g region to the right side of the image", right, value);
         
         else sprintf(buf, "cropped %d pixel region from the right side of the image", 0-right);
         history += string(buf);
         if( bottom || top ) history += string(", ");
      }
      if(bottom)
      {
         if(expand_bottom)
            sprintf(buf, "added %d pixel with value = %g region to the bottom side of the image", bottom, value);
         
         else sprintf(buf, "cropped %d pixel region from the bottom side of the image", 0-bottom);
         history += string(buf);
         if( top ) history += string(", ");
      }
      if(top)
      {
         if(expand_top)
            sprintf(buf, "added %d pixel with value = %g region to the top side of the image", top, value);
         
         else sprintf(buf, "cropped %d pixel region from the top side of the image", 0-top);
         history += string(buf);
      }
   }
   return history;
}


template<class T, int N>
string FitsTools<T, N>::rotateReverseArray(MArray<T, N>& a,
                                           const int rot,
                                           const bool x, const bool y)
{
   string history = "";   
   if( (rot==90) || (rot==180) || (rot==270) )
   {
      char buf[64];
      sprintf(buf, "rotated image clockwise %d deg", rot);
      history += string(buf);
      if(y || x)
         history += string(", ");
      if(rot==180)
      {
         a.reverseSelf(1);
         a.reverseSelf(2);
      }
      else
      {
         a.transposeSelf();
         if(rot==90) a.reverseSelf(2);
         else a.reverseSelf(1);
      }
   }
   if( x)
   {
      history += string("reversed x axis");
      if(y)
         history += string(", ");
      a.reverseSelf(1);
   }
   if( y)
   {
      history += string("reversed y axis");
      a.reverseSelf(2);
   }
   return history;
}

template<class T, int N>
inline bool FitsTools<T, N>::loadError2(FitsIn& infile,
                                        MArray<T, N>& error)
{
   try{
      if(verbose_ > 0)
         cerr << "Reading " << infile.getFilename() << endl;
      // build MArray
      infile >> error;
      infile.freeData();
      // square error
      error = merge(error > T(0), error * error, error);
      return true;
   }
   catch (exception &e){
      if(verbose_ > 0)
         cerr << e.what() << " No errorfile\n";
   }
   return false;
}

template<class T, int N>
bool FitsTools<T, N>::loadError2(const string& filename,
                                 MArray<T, N>& error)
{
   try{
      // open file
      FitsIn infile( prependFilename(filename, "e."),
                           quiet_please_, true );
      return loadError2(infile, error);
   }
   catch (exception &e){
      if(verbose_ > 0)
         cerr << e.what() << " No errorfile\n";
   }
   return false;
}

template<class T, int N>
bool FitsTools<T, N>::loadError2(const string& filename,
                                 MArray<T, N>& error,
                                 const Region& region)
{
   try{
      // open file
      FitsIn infile( prependFilename(filename, "e."),
                           quiet_please_, true );
      selectRegionFile(infile, region);
      return loadError2(infile, error);
   }
   catch (exception &e){
      if(verbose_ > 0)
         cerr << e.what() << " No errorfile\n";
   }
   return false;
}

template<class T, int N>
inline void FitsTools<T, N>::loadImage(FitsIn& infile,
                                       MArray<T, N>& image)
{
   if(verbose_ > 0)
      cerr << "Reading " << infile.getFilename() << endl;
   // build MArray
   infile >> image;
   infile.freeData();
}

template<class T, int N>
inline void FitsTools<T, N>::loadImage(const string& filename,
                                       MArray<T, N>& image)
{
   FitsIn infile( filename, quiet_please_, ignore_header_ );
   loadImage(infile, image);
}

template<class T, int N>
inline void FitsTools<T, N>::loadImage(const string& filename,
                                       MArray<T, N>& image,
                                       const Region& region)
{
   // open file
   FitsIn infile( filename, quiet_please_, ignore_header_ );
   selectRegionFile(infile, region);
   loadImage(infile, image);
}

template<class T, int N>
inline bool FitsTools<T, N>::loadImageError2(FitsIn& infile,
                                             MArray<T, N>& image,
                                             MArray<T, N>& error)
{
   loadImage(infile, image);
   return loadError2(infile.getFilename(), error);
}

template<class T, int N>
inline bool FitsTools<T, N>::loadImageError2(const string& filename,
                                             MArray<T, N>& image,
                                             MArray<T, N>& error)
{
   loadImage(filename, image);
   return loadError2(filename, error);
}

template<class T, int N>
inline bool FitsTools<T, N>::loadImageError2(const string& filename,
                                             MArray<T, N>& image,
                                             MArray<T, N>& error,
                                             const Region& region)
{
   loadImage(filename, image, region);
   return loadError2(filename, error, region);
}

template<class T, int N>
void FitsTools<T, N>::loadImageOffset(FitsIn& infile,
                                      const double offset,
                                      MArray<T, N>& image)
{
   if(offset)
   {
      const double old_bzero = infile.getBzero();
      infile.setBzero(old_bzero + offset);
      // altered BZERO, so subtraction is done while reading
   }
   loadImage(infile, image);
}

template<class T, int N>
inline void FitsTools<T, N>::saveError2(MArray<T, N>& error,
                                        FitsOut& outfile)
{
   error = merge(error > T(0), sqrt(error), error);
   outfile.setFilename( prependFilename(outfile.getFilename(), "e.") );
   saveImage(error, outfile);
}

template<class T, int N>
inline void FitsTools<T, N>::saveImage(MArray<T, N>& image,
                                       FitsOut& outfile)
{
   if(verbose_ > 0)
      cerr << "Saving " << outfile.getFilename() << endl;
   outfile << image;
}

template<class T, int N>
void FitsTools<T, N>::saveImageError2(MArray<T, N>& image,
                                      MArray<T, N>& error,
                                      FitsOut& outfile)
{
   saveImage(image, outfile);
   saveError2(error, outfile);
}

template<class T, int N>
void FitsTools<T, N>::referenceRegion(MArray<T, N>& ref,
                                      const MArray<T, N>& input,
                                      const Region& region)
{
   const string mismatchtxt = "Selected region does not conform with array.";
   if(region.getDim() != N)
      throw RangeException(mismatchtxt);
   if(region.getStart(1))
   {
      for( int i = 1; i <= N; ++i )
         if( (region.getStart(i) < 1) ||
             (region.getEnd(i) > input.length(i)) ||
             (region.getStart(i) > region.getEnd(i)) )
            throw RangeException(mismatchtxt);
      // adjust next line for other N
      switch(N)
      {
         case 1:
            ref.makeReference( input( Range(region.getStart(1),
                                            region.getEnd(1)) ) );
            break;
         case 2:
            ref.makeReference( input( Range(region.getStart(1),
                                            region.getEnd(1)),
                                      Range(region.getStart(2),
                                            region.getEnd(2)) ) );
            break;
         case 3:
            ref.makeReference( input( Range(region.getStart(1),
                                            region.getEnd(1)),
                                      Range(region.getStart(2),
                                            region.getEnd(2)),
                                      Range(region.getStart(3),
                                            region.getEnd(3)) ) );
            break;
         default:
            throw RangeException("Adjust method FitsTools<T, N>::referenceRegion()!");
      };
   }
   else ref.makeReference(input);
}

template<class T, int N>
inline void FitsTools<T, N>::selectRegionFile(FitsIn& infile,
                                              Region region)
{
   // set region to be extracted
   const bool isFullFrame = adjustRegion2File(infile, region);
   if(isFullFrame)
      infile.resetRegion();
   else
      infile.setRegion(region);
}

// if region is exceeding FITS file clip according to FITS file boundaries
// return true if resulting region matches full FITS file
template<class T, int N>
inline bool FitsTools<T, N>::adjustRegion2File(const FitsIn& infile,
                                               Region& region)
{
   // set region to be extracted
   const int naxis = infile.getNaxis();
   bool isFullFrame = true;
   for(int i = 1; i <= naxis; ++i)
   {
      int start = region.getStart(i);
      if(start < 0)
      {
         start = 1;
         if(verbose_ > 0)
            cerr << "NAXIS" << i
                 << " start clipped to 1" << endl;
      }
      int end = region.getEnd(i);
      const int naxis_i = infile.getNaxis(i);
      if(end > naxis_i)
      {
         end = naxis_i;
         if(verbose_ > 0)
            cerr << "NAXIS" << i 
                 <<" end clipped to " << naxis_i << endl;
      }
      if(start && end)
         region.setRange(i, start, end);
      else
      {
         region.setRange(i, 1, naxis_i);
         if(verbose_ > 1)
            cerr << "NAXIS" << i 
                 << " set to full range 1:" << naxis_i << endl;
      }
      if( (start != 1) || (end != naxis_i) )
         isFullFrame = false;
   }
   return isFullFrame;
}

#endif
