/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: flatfits.cpp 145 2010-10-04 13:36:16Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Arno Riffeser <arri@usm.uni-muenchen.de>
 *                         Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */


#define LTL_RANGE_CHECKING

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#ifdef UTIL_USING_NAMESPACE
using namespace util;
#endif

#include <ltl/fitsio.h>
#include <ltl/statistics.h>

#ifdef LTL_USING_NAMESPACE
using namespace ltl;
#endif

#include <string>
#include <iostream>
#include <list>
using namespace std;

#include <cstdio>

// structure for flag values
class flagValues
{
   public:
      flagValues() : region(2) { };
      string outfile;
      Region region;
      //float tresh;
      float kappa;
      float highkappa;
      double normkappa;
      int box;

      bool one_flat;
//       bool sub;

      float nan_value;
      bool nan;
      string prefix;

      int maxmem;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char **argv)
{
   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   try{
      const size_t n_files = files.size();

      // open input files
      list<FitsIn> infiles;
      list<FitsIn> inerrors;

      list<string>::const_iterator file_i = files.begin();
      string history = string("Combined via flatfits:");
      while( file_i != files.end() )
      {
         const string filename = *file_i;
         if(flagVal.verbose > 0)
            cerr << "Reading " << filename << endl;
         infiles.push_back( FitsIn(filename,
                                   flagVal.quiet_please,
                                   flagVal.ignore_header) );
         const string errorname = prepend_filename(filename, "e.");
         if(flagVal.verbose > 0)
            cerr << "Reading " << errorname << endl;
         inerrors.push_back( FitsIn(errorname,
                                    flagVal.quiet_please,
                                    true) );
         history += string(" ") + filename;
         ++file_i;
      }

      // average per image
      MArray<float, 1> ave(n_files);

      const list<FitsIn>::iterator first_file = infiles.begin();
      list<FitsIn>::iterator infile_i = first_file;
      MArray<float, 1>::iterator ave_i = ave.begin();

      if(flagVal.region.getStart(1) == 0)
      {
         const int xmin = 1;
         const int xmax = (*first_file).getNaxis(1);
         const int ymin = 1;
         const int ymax = (*first_file).getNaxis(2);
         flagVal.region.setRange( 1,
                                  xmin + (xmax + 1 - xmin) / 4,
                                  xmax - (xmax + 1 - xmin) / 4 );
         flagVal.region.setRange( 2,
                                  ymin + (ymax + 1 - ymin) / 4,
                                  ymax - (ymax + 1 - ymin) / 4 );
      }

      if(flagVal.verbose > 0)
      {
         cerr << "Evaluating normalisation";
         if( flagVal.region.getStart(1) != 0)
            cerr << " of region " << flagVal.region.toString();
         cerr << ":" << endl;
      }

      while( !ave_i.done() )
      {
         if( flagVal.region.getStart(1) != 0)
            (*infile_i).setRegion(flagVal.region);
         double holder;
         MArray<float, 2> image;
         (*infile_i) >> image;
         if(flagVal.nan)
            kappa_median_average(image, flagVal.normkappa,
                                 flagVal.nan_value, &holder);
         else
            kappa_median_average(image, flagVal.normkappa,
                                 &holder);
         *ave_i = float(holder);
         if( flagVal.region.getStart(1) != 0)
            (*infile_i).resetRegion();
         else
            (*infile_i).freeData();

         if(flagVal.verbose > 0)
            cerr << (*infile_i).getFilename() << "\t" << holder << endl;

         ++infile_i;
         ++ave_i;
      }

      // calculate result
      const int xlen = (*first_file).getNaxis(1);
      const int ylen = (*first_file).getNaxis(2);
      const list<FitsIn>::iterator first_error = inerrors.begin();

      // result, e_result, control u, control b
      MArray< short int, 2> umat;
      if(flagVal.verbose > 1)
         umat.makeReference(MArray< short int, 2>(xlen, ylen));
      MArray< int, 2> bmat;
      if(flagVal.verbose > 2)
         bmat.makeReference(MArray< int, 2>(xlen, ylen));

      MArray<char, 3> maskcube(n_files, xlen, ylen);
      maskcube = 1;

      MArray<float, 2> files_rows;
      MArray<float, 2> errors_rows;
      MArray<float, 2> outarray(xlen, ylen);
      MArray<float, 2> outerror(xlen, ylen);
      MArray<float, 2>::iterator out_i = outarray.begin();
      MArray<float, 2>::iterator out_e = outerror.begin();

      MArray<float, 1> refholder;
      MArray<float, 1> e_refholder;

      list < MArray<float, 2> > mafifo;

      if(flagVal.verbose > 0)
      {
         cerr << "Calculating result image" << endl;
      }

      // number of lines for simultaneous read
      const int n_lines = (flagVal.maxmem * 1024 * 1024) / (n_files * xlen * 16);
      Region y_block(2);
      y_block.setRange(1, 1, xlen);
      MArray<float, 3> fileblock;
      MArray<float, 3> errorblock;

      for (int y = 1, yy = 1 - (1 + flagVal.box); yy <= ylen; ++y, ++yy)
      {
         if(y <= ylen)
         {
            if(flagVal.verbose > 1)
            {
               cerr << "highkappa: " << (y * 100.0) / ylen << " %        \t";
               cerr.flush();
            }
            list<FitsIn>::iterator infile_i = first_file;
            list<FitsIn>::iterator inerror_i = first_error;

            // read y-th row of all files, read in blocks
            if(!( (y-1) % n_lines )){ // allocate new block and read into it
               y_block.setRange(2, y, (y+n_lines-1<=ylen) ? y+n_lines-1 : ylen);
               fileblock.makeReference(MArray<float, 3>(int(n_files), xlen, int(y_block.getLength(2))));
               errorblock.makeReference(MArray<float, 3>(int(n_files), xlen, int(y_block.getLength(2))));
               MArray<float, 2> blockref;
               for(size_t z = 1; z <= n_files; ++z){
                  if(flagVal.verbose > 1)
                     cerr << "Reading " << (*infile_i).getFilename() << y_block.toString() << endl;
                  blockref.makeReference( fileblock(z, Range::all(), Range::all()) );
                  (*infile_i).setRegion(y_block);
                  (*infile_i) >> blockref;
                  ++infile_i;
                  blockref.makeReference( errorblock(z, Range::all(), Range::all()) );
                  (*inerror_i).setRegion(y_block);
                  (*inerror_i) >> blockref;
                  ++inerror_i;
               }
            }
         
            //files_rows.makeReference(MArray<float, 2>(n_files, xlen));
            //errors_rows.makeReference(MArray<float, 2>(n_files, xlen));
            const int y_inblock = ((y-1) % n_lines) + 1;
            files_rows.makeReference( fileblock(Range::all(), Range::all(), y_inblock) );
            mafifo.push_back(files_rows);
            errors_rows.makeReference( errorblock(Range::all(), Range::all(), y_inblock) );
            mafifo.push_back(errors_rows);
            
            for (int x = 1; x <= xlen; ++x)
            { 
               // pixel stack
               MArray<float,1> pixel(files_rows(Range::all(), x));
               MArray<float,1> epixel(errors_rows(Range::all(), x));
      
               int z = count(pixel != flagVal.nan_value);
               MArray<char, 1> refmask(maskcube(Range::all(), x, y));
               refmask = merge(pixel != flagVal.nan_value,
                               refmask, 0);

               // if any not nan
               if (z > 0)
               {
                  // normed pixel stack
                  MArray<float,1> npixel =
                     merge(pixel != flagVal.nan_value,
                           pixel / ave, flagVal.nan_value);
                  MArray<float,1> enpixel =
                     merge(epixel > 0.0f,
                           flagVal.highkappa * epixel / ave,
                           0.0f);
                  MArray<float,1> em2npixel =
                     merge(epixel > 0.0f,
                           pow2(ave / epixel), 0.0f);

                  const int x0 = x - flagVal.box;
                  const int x1 = x + flagVal.box;
                  const int y0 = y - flagVal.box;
                  const int y1 = y + flagVal.box;
                  MArray<char, 3> submask = 
                     maskcube( Range::all(),
                               Range( ((x0 < maskcube.minIndex(2)) ?
                                       maskcube.minIndex(2) : 
                                       x0),
                                      ((x1 > maskcube.maxIndex(2)) ?
                                       maskcube.maxIndex(2) : 
                                       x1) ),
                               Range( ((y0 < maskcube.minIndex(3)) ?
                                       maskcube.minIndex(3) : 
                                       y0),
                                      ((y1 > maskcube.maxIndex(3)) ?
                                       maskcube.maxIndex(3) : 
                                       y1) ) );
                  int zold;
                  // clip ceiling - highkappa
                  do // q*sig Ausreisser klippen
                  {
                     zold = z;
                     // error weighted mean
                     const float eval = sum(em2npixel);
                     const float val  = sum(npixel * em2npixel);
                     const float c = val / eval; // error weighted mean
                     // clip ceiling resp. to per pixel error
                     MArray<float, 1>::iterator em2np_i = em2npixel.begin();
                     MArray<float, 1>::const_iterator np_i = npixel.begin();
                     MArray<float, 1>::const_iterator enp_i = enpixel.begin();
                     for (int i = 1; i <= int(n_files); ++i) { // HIGHKAPPA
                        if ( *em2np_i &&
                             (*np_i - c) > (*enp_i) )
                        { // maxima
                           --z;
                           *em2np_i = 0.0f; // nan in per pixel stack
                           // nan in result "cube" using box
                           submask(i, Range::all(), Range::all()) = 0;  // BOX
                        }
                        ++em2np_i; ++np_i; ++enp_i;
                     }
                  } while (z != zold);
               } // end of if(z > 0)
            } // end of x-loop
         }
         else
         {
            if(flagVal.verbose > 1)
               cerr << "highkappa: " << 100.0 << " %        \t";
         }
         
         if(yy >= 1)
         {
            if(flagVal.verbose > 1)
               cerr << "kappa: " << (yy * 100.0) / ylen << " %        \t\r";

            files_rows.makeReference(*(mafifo.begin()));
            mafifo.pop_front();
            errors_rows.makeReference(*(mafifo.begin()));
            mafifo.pop_front();

            // loop for symmetric clipping (ceil + floor)
            for (int x = 1; x <= xlen; ++x)
            { 
               // pixel stack
               MArray<float,1> pixel(files_rows(Range::all(), x));
               MArray<float,1> epixel(errors_rows(Range::all(), x));
      
               MArray<char, 1> refmask(maskcube(Range::all(), x, yy));
               int z = count(refmask != char(0));
               // if any not nan
               if (z > 1 || (flagVal.one_flat && z > 0))
               {
                  // normed pixel stack
                  MArray<float,1> npixel =
                     merge(pixel != flagVal.nan_value,
                           pixel / ave, flagVal.nan_value);
                  MArray<float,1> enpixel =
                     merge(epixel > 0.0f,
                           epixel / ave,
                           0.0f);
                  MArray<float,1> em2npixel =
                     merge(refmask,
                           pow2(ave / epixel), 0.0f);

                  int zold = 0;
                  // clip worst pixel (biggest error weighted offset from mean)
                  while( (z > 1) && (z != zold) )
                  {
                     zold = z;
                     // error weighted mean
                     const float eval = sum(em2npixel);
                     const float val  = sum(npixel * em2npixel);
                     const float c = val / eval; // error weighted mean
                     // clip resp. to per pixel error
                     MArray<float, 1>::iterator em2np_i = em2npixel.begin();
                     MArray<float, 1>::const_iterator np_i = npixel.begin();
                     MArray<float, 1>::const_iterator enp_i = enpixel.begin();
                     int maxi = 0; float maxp = 0.0f;
                     for (int i = 1; i <= int(n_files); ++i) {
                        if (*em2np_i) {
                           const float holder = float(fabs(*np_i - c)) / *enp_i;
                           if(holder > maxp)
                           { maxi = i; maxp = holder;}
                        }
                        ++em2np_i; ++np_i; ++enp_i;
                     }
                     if(maxp > flagVal.kappa)
                     { // clip maximum
                        em2npixel(maxi) = 0.0f; // nan in per pixel stack
                        refmask(maxi) = char(0); // nan in result "cube" using box
                        --z;
                     }
                  }
                  if(z > 1)
                  {
                     em2npixel = refmask;
                     const float aves = sum(ave * em2npixel);
                     *out_i = sum(em2npixel * pixel) / aves;
                     *out_e = float( sqrt( sum( pow2(em2npixel * epixel) ) ) ) / aves;
                     if(umat.isAllocated())
                        umat(x, yy) = z;
                     if(bmat.isAllocated())
                     {
                        int tmp = 0;
                        MArray<char ,1>::iterator ok_i = refmask.begin();
                        for (int i = 0; i < int(n_files); ++i)
                        {
                           tmp +=  int(*ok_i) * (1 << i);
                           ++ok_i;
                        }
                        bmat(x, yy) = tmp;
                     }
                  }
                  else // one valid pixel
                  {
                     int i = 1;
                     while(refmask(i) == char(0)) ++i;
                     *out_i = npixel(i); *out_e = enpixel(i);
                     if(umat.isAllocated())
                        umat(x, yy) = 1;
                     if(bmat.isAllocated())
                        umat(x, yy) = (1 << (i-1));
                  }
               } // end of if(z > 0)
               else // no valid pixel
               {
                  *out_i = flagVal.nan_value; 
                  *out_e = 0.0f;
                  if(umat.isAllocated())
                     umat(x, yy) = 0;
                  if(bmat.isAllocated())
                     bmat(x, yy) = 0;
               }
               ++out_i;
               ++out_e;
            } // end of x-loop
         }
         else
         {
            if(flagVal.verbose > 1)
            {
               cerr << "kappa: " << 0.0 << " %        \t\r";
            }
         }
      } // end of y-loop

      if(flagVal.verbose > 1)
         cerr << endl;

      inerrors.clear();
      ave.free();
      maskcube.free();
      files_rows.free(); errors_rows.free();
      refholder.free(); e_refholder.free();
      mafifo.clear();

      if( flagVal.outfile.size() == 0 )
         flagVal.outfile = prepend_filename(*files.begin(), flagVal.prefix);

      FitsOut out(flagVal.outfile, *first_file, flagVal.quiet_please);
      infiles.clear();
      
      out.addHistory(history);
      if(flagVal.verbose > 0)
         cerr << "Saving " << out.getFilename() << endl;
      out << outarray;
      outarray.free();
      
      out.setFilename( prepend_filename(flagVal.outfile, string("e.")) );
      if(flagVal.verbose > 0)
         cerr << "Saving " << out.getFilename() << endl;
      out << outerror;
      outerror.free();

      if(umat.isAllocated())
      {
         cerr << "Number of pixels per flats used:\n";
         for(short int i = 0; i <= (short int)(n_files); ++i)
         {
            cerr << "Flats:\t" << i
                 << "\t= " << count(umat == i) << endl;
         }
         out.setFilename( prepend_filename(flagVal.outfile, string("u.")) );
         out.setBitpixOut(16);
         out.setBzero(32768);
         umat += ((short)-32768); // -= 32768 is illegal as there is no signed short of that size
         cerr << "Saving " << out.getFilename() << endl;
         out << umat;
         umat.free();
      }
      
      if(bmat.isAllocated())
      {
         out.setFilename( prepend_filename(flagVal.outfile, string("b.")) );
         out.setBitpixOut(32);
         out.setBzero(0);
         cerr << "Saving " << out.getFilename() << endl;
         out << bmat;
         bmat.free();
      }
      
   }
   catch (exception &e)
   {
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
   }
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: flatfits [-FLAGS [option]] FILES\n") +
      string("build a maximum S/N normed average frame of FILES\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Arno Riffeser <arri@usm.uni-muenchen.de>\n") +
      string("                        Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags,
                 flagValues& flagVal)
{
   try {
      flags.addOption(new RegionArrayOption("region","0:0,0:0",
                                            "normalize by clipping region x_start:x_end,y_start:y_end (default is center quarter)",
                                            'r', 2, &flagVal.region));
      flags.addOption(new FloatOption("kappa", "3",
                                      "kappa for symmetric clipping ( k < q !!!)", 'k', &flagVal.kappa));
      flags.addOption(new FloatOption("high_kappa", "5",
                                      "kappa for clip ceiling ( q > k !!!)", 'q', &flagVal.highkappa));
      flags.addOption(new DoubleOption("norm_kappa", "3",
                                       "kappa for clipping to determine normalisation factor", 0, &flagVal.normkappa));
      flags.addOption(new IntOption("box", "2",
                                    "box radius", 'b', &flagVal.box));
      flags.addOption(new BoolOption("force_flat", "FALSE",
                                     "default only gives result, where at least two flats remain after high clipping, this flag overrides the default",
                                     'f', &flagVal.one_flat ) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan_value in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("outfile","",
                                       "output filename",'o',&flagVal.outfile));
      flags.addOption(new StringOption("prefix", "m.", "output file prefix, if not using oufile",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("maxmem", "256",
                                    "do not use more than maxmem (MB) memory => use tiled file reads",
                                    0 , &flagVal.maxmem ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}


