/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: headfits.cpp 142 2009-10-15 13:53:37Z snigula $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) Claus A. Goessl <cag@usm.uni-muenchen.de>
 *               Niv Drory       <drory@astro.as.utexas.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fits.h>
#include <ltl/misc/exceptions.h>

using namespace ltl;

#include <stdio.h>
#include <iostream>
#include <string>
#include <sstream>
#include <list>
#include <cstdlib>
using namespace std;

#define MAXKEYS 32

// globals for flag values
struct globals
{
      bool help;
      bool print;       // print the whole header (default)
      bool read;        // read a single keyword
      bool pfile;       // print filename with header info
      bool del;         // delete a keyword
      bool add;         // add a keyword, error if already exists
      bool comment;     // add a COMMENT line
      bool history;     // add a HISTORY line
      string keyword[MAXKEYS];
      string type;
      string value;
      string value_comment;
} g;

void print_usage( OptionParser& flags );

void add_options( OptionParser& flags );

int main(int argc, char* argv[])
{
   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   OptionParser      flags( &comline );
   flags.changeReader(&comline);

   add_options( flags );

   try
   {
      flags.parseOptions();
   }
   catch( UException& e )
   {
      cerr << "headfits ERROR: Caught exception: " << e.what() << endl;
      print_usage( flags );
      exit(1);
   }

   if( g.help )
   {
      print_usage( flags );
      exit(0);
   }

   try
   {
      if( g.read || g.add || g.del || g.comment || g.history )
         g.print = false;

      const list<string> files = comline.fileArguments(); // get the filenames
      if(files.size() < 1)
         throw IOException( "At least one file must be given" );

      // for all FILES
      for( list<string>::const_iterator file_i=files.begin();
         file_i != files.end();
           ++file_i )
      {
         bool modified = false;

         // print the entire header to stdout
         if( g.print )
         {
            if( g.pfile )
               cout << *file_i << endl;

            FitsIn infile(*file_i, true, true );
            string header=infile.writeHeader(true);
            int n = header.length();
            for( int j=0; j<n; j+=80 )
               cout << header.substr(j, 80) << endl;
            continue;
         }

         // open the file and trim the keyword
         FitsIn infile(*file_i, true, false );

         // read a keyword or a list of keywords
         if( g.read )
         {
            if( g.pfile )
               cout << *file_i << "  ";

            int j=0;
            while( j<MAXKEYS && g.keyword[j].length() > 0 )
            {
               if( j>0 ) cout << " ";
               string tmpk = g.keyword[j];
               string key = infile.adjustKeyword( tmpk );
               if( key == "COMMENT " )
                  cout << infile.getComment();
               else if( key == "HISTORY " )
                  cout << infile.getHistory();
               else
                  cout << infile.getValueAsString( key );

               ++j;
            }
            cout << endl;
         }
         else  // not read
         {
            string tmpk = g.keyword[0];
            if( tmpk.length() == 0 )
                  throw IOException( "No header keyword named for modification!" );

            string key = infile.adjustKeyword( tmpk );
            string tmpfile = *file_i+".tmp";
            FitsOut outfile( tmpfile, infile.getHeader(), true, false );

            // add/write a keyword
            if( g.add )
            {
               cout << "headfits: adding card " << key << ":" << g.value << endl;
               if( g.type == "string" )
                  outfile.addValueCard( key, g.value, g.value_comment );
               else if ( g.type == "bool" )
               {
                  bool b;
                  istringstream tmp;
                  tmp.str( g.value );
                  tmp >> b;
                  outfile.addValueCard( key, b, g.value_comment );
               }
               else if ( g.type == "int" )
               {
                  int b;
                  istringstream tmp;
                  tmp.str( g.value );
                  tmp >> b;
                  outfile.addValueCard( key, b, g.value_comment );
               }
               else if ( g.type == "float" )
               {
                  float b;
                  istringstream tmp;
                  tmp.str( g.value );
                  tmp >> b;
                  outfile.addValueCard( key, b, g.value_comment );
               }
               else
                  throw IOException( "Unsupported header data type" );

               modified = true;
            }

            // add COMMENT or HISTORY line
            if( g.comment )
               infile.addComment( g.value );
            if( g.history )
               infile.addHistory( g.value );

            // remove a keyword
            if( g.del )
            {
               cout << "headfits: deleting card " << key << endl;
               outfile.eraseCard( key );
               modified = true;
            }

            // write the file if modified
            if( modified )
            {
               cout << "headfits: updating file " << *file_i << endl;
               outfile << infile;
               infile.freeData();
               rename( tmpfile.c_str(), (*file_i).c_str() );
            }

         } // not read

      } // for all files
   }
   catch (FitsException& e)
   {
      cerr << "headfits ERROR: caught exception: " << e.what() << endl;
      exit(1);
   }
   return 0;
}


void print_usage( OptionParser& flags )
{
   cout << "Usage: headfits [OPTIONS] FILES ... " << endl;
   cout << "   Prints/edit FITS header information" << endl;

   flags.printUsage( cout );

   exit(1);
}

void add_options(OptionParser& flags)
{
   try
   {
      flags.addOption(new BoolOption("help", "FALSE", "Print usage information",
                                     'h', &g.help) );

      flags.addOption(new BoolOption("print", "TRUE", "Print entire header [default]",
                                     'p', &g.print) );

      flags.addOption(new BoolOption("read", "FALSE", "Read a single keyword or a list of keywords",
                                     'r', &g.read) );

      flags.addOption(new BoolOption("filename", "FALSE", "Print name of the processed file along with the header information",
                                     'f', &g.pfile) );

      flags.addOption(new BoolOption("add", "FALSE", "Add a keyword/value, error if it already exists",
                                     'a', &g.add) );

      flags.addOption(new BoolOption("delete", "FALSE", "Delete a single keyword",
                                     'd', &g.del) );

      flags.addOption(new BoolOption("comment", "FALSE", "Add a COMMENT line",
                                     'c', &g.comment) );

      flags.addOption(new BoolOption("history", "FALSE", "Add a HISTORY line",
                                     'j', &g.history) );

      flags.addOption(new StringOption("type", "string", "the value type to be added [string|bool|int|float]. Default is string.",
                                       't', &g.type) );

      flags.addOption(new StringArrayOption("keyword", ",,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,",
                                            "The keyword to read or write, or a comma-separated list of keywords to read (max 32).",
                                            'k', 32, &g.keyword[0]) );

      flags.addOption(new StringOption("value", "", "The value",
                                       'v', &g.value) );

      flags.addOption(new StringOption("value-commnet", "", "The value comment",
                                       'V', &g.value_comment) );

   }

   catch( UException e )
   {
      cerr << e.what() << endl;
      exit(1);
   }

}
