/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: joinfits.cpp 177 2018-08-01 23:15:21Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2006 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#include <ltl/fitsio.h>

using namespace ltl;
using namespace util;
using namespace std;

// structure for flag values
class flagValues
{
   public:
      int naxis;
      string prefix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

template<class T, int N>
void fill_outfile(FitsIn& infile, FitsOut& outfile, const Region& region)
{
   MArray<T, N> a;
   infile >> a;
   outfile.setRegion(region);
   outfile << a;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 2) print_usage(flags);
   
   try{
      // open input files
      list<FitsIn> infiles;
//       list<FitsIn> inerrors;
//       list<bool> have_errors;
      list<string>::const_iterator file_i = files.begin();
      while( file_i != files.end() )
      {
         const string filename = *file_i;
         if(flagVal.verbose > 0)
            cerr << filename << endl;
         infiles.push_back( FitsIn(filename,
                                   flagVal.quiet_please,
                                   flagVal.ignore_header) );
         ++file_i;
      }
      
      // calculate output file dimensions
      const list<FitsIn>::iterator first_file = infiles.begin();
      const int bitpix = (*(first_file)).getBitpix();
      const int naxis = (*(first_file)).getNaxis();
      Region region(naxis);
      for(int i = 1; i<=naxis; ++i)
         if(i!=flagVal.naxis)
            region.setRange( i, 1, (*(first_file)).getNaxis(i) );
      {
         int join_axis = 0;
         for(list<FitsIn>::iterator infile_i = first_file;
             infile_i != infiles.end();
             ++infile_i)
         {
            join_axis += (*(infile_i)).getNaxis(flagVal.naxis);
         }
         region.setRange( flagVal.naxis, 1, join_axis );
      }
      
      const string outname = prepend_filename(*(files.begin()), flagVal.prefix);
      FitsOut outfile(outname);
      outfile.setGeometry(bitpix, region);
      {
         int join_axis = 0;
         for(list<FitsIn>::iterator infile_i = first_file;
             infile_i != infiles.end();
             ++infile_i)
         {
            const int naxis_i = (*(infile_i)).getNaxis(flagVal.naxis);
            region.setRange(flagVal.naxis, join_axis + 1, join_axis + naxis_i);
            join_axis += naxis_i;
            switch(bitpix)
            {
               case LTL_FITS_BITPIX_CHAR:
               {
                  switch(naxis)
                  {
                     case 1:fill_outfile<unsigned char, 1>(*infile_i, outfile, region);break;
                     case 2:fill_outfile<unsigned char, 2>(*infile_i, outfile, region);break;
                     case 3:fill_outfile<unsigned char, 3>(*infile_i, outfile, region);break;
                     case 4:fill_outfile<unsigned char, 4>(*infile_i, outfile, region);break;
                     case 5:fill_outfile<unsigned char, 5>(*infile_i, outfile, region);break;
                  }
               }
               break;
               case LTL_FITS_BITPIX_SHORT:
               {
                  switch(naxis)
                  {
                     case 1:fill_outfile<short int, 1>(*infile_i, outfile, region);break;
                     case 2:fill_outfile<short int, 2>(*infile_i, outfile, region);break;
                     case 3:fill_outfile<short int, 3>(*infile_i, outfile, region);break;
                     case 4:fill_outfile<short int, 4>(*infile_i, outfile, region);break;
                     case 5:fill_outfile<short int, 5>(*infile_i, outfile, region);break;
                  }
               }
               break;
               case LTL_FITS_BITPIX_INT:
               {
                  switch(naxis)
                  {
                     case 1:fill_outfile<int, 1>(*infile_i, outfile, region);break;
                     case 2:fill_outfile<int, 2>(*infile_i, outfile, region);break;
                     case 3:fill_outfile<int, 3>(*infile_i, outfile, region);break;
                     case 4:fill_outfile<int, 4>(*infile_i, outfile, region);break;
                     case 5:fill_outfile<int, 5>(*infile_i, outfile, region);break;
                  }
               }
               break;
               case LTL_FITS_BITPIX_FLOAT:
               {
                  switch(naxis)
                  {
                     case 1:fill_outfile<float, 1>(*infile_i, outfile, region);break;
                     case 2:fill_outfile<float, 2>(*infile_i, outfile, region);break;
                     case 3:fill_outfile<float, 3>(*infile_i, outfile, region);break;
                     case 4:fill_outfile<float, 4>(*infile_i, outfile, region);break;
                     case 5:fill_outfile<float, 5>(*infile_i, outfile, region);break;
                  }
               }
               break;
               case LTL_FITS_BITPIX_DOUBLE:
               {
                  switch(naxis)
                  {
                     case 1:fill_outfile<double, 1>(*infile_i, outfile, region);break;
                     case 2:fill_outfile<double, 2>(*infile_i, outfile, region);break;
                     case 3:fill_outfile<double, 3>(*infile_i, outfile, region);break;
                     case 4:fill_outfile<double, 4>(*infile_i, outfile, region);break;
                     case 5:fill_outfile<double, 5>(*infile_i, outfile, region);break;
                  }
               }
               break;
               default: throw FitsException("illegal BITPIX");
            }

         }
      }
      outfile.resetRegion();
   } // these files should have worked, otherwise ->
   catch (exception &e){
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: joinfits [-FLAGS [option]] FILES\n") +
      string("join data of FILES along given NAXIS\n\n");
   const string copyright = 
      string("Copyright (C) 2006 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new IntOption("naxis", "1",
                                    "NAXIS along which to join",
                                    'n', &flagVal.naxis ) );
      flags.addOption(new StringOption("prefix", "j.", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

