/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: fits2jpeg.cpp 94 2005-02-21 12:18:38Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>
using namespace util;

#include <ltl/fits.h>
using namespace ltl;

#include <string>
#include <iostream>
using namespace std;

#include <cstdio>

extern "C" {
#include <jpeglib.h>
#include <jerror.h>
//#include <jconfig.h>
#ifndef JPEG_MAX_DIMENSION // hack to avoid double inclusion of jmorecfg in unguarded implementations
#include <jmorecfg.h>
#endif
}


// structure for flag values
class flagValues
{
   public:
      flagValues() { };
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string fits_filename(string path)
{
   path.erase(path.find_last_of('.') + 1);
   path.append(string("fits"));
   return path;
}

int main(int argc, char* argv[])
{
   // structure for flag values
   flagValues flagVal;

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   OptionParser      flags(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   if(flagVal.verbose > 0)
      cerr << argv[0] << ":" << endl;
   
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_decompress(&cinfo);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // read file
	FILE * infile;
	if ((infile = fopen((*file_i).c_str(), "rb")) == NULL) {
           cerr << "can't open " << *file_i << endl;
           throw;
	}
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, TRUE);
        cinfo.out_color_space = JCS_GRAYSCALE;
        jpeg_start_decompress(&cinfo);

        // prepare save file
        const int bitpix = 8;
        const int naxis = 2;
        const int naxis_i[2] = { static_cast<int>(cinfo.image_width), static_cast<int>(cinfo.image_height) };
        const string outfilename = fits_filename(*file_i);
        FitsOut outfile(outfilename);
        outfile.openData(bitpix, naxis, naxis_i);
        unsigned char * const start = outfile.begin();
        unsigned char * const end   = outfile.end();

        // read lines into FITS-file, reversing
        for(unsigned char * line = end - naxis_i[0];
            line >= start; line -= naxis_i[0]){
           jpeg_read_scanlines(&cinfo, &line, 1);
        }

        jpeg_finish_decompress(&cinfo);        
        outfile.closeData();

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }
   jpeg_destroy_decompress(&cinfo);

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: jpeg2fits [-FLAGS [option]] FILES\n") +
      string("Convert JPEG FILES to FITS images.\n\n");
   const string copyright = 
      string("Copyright (C) 2007-2008 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new IntOption("verbose", "0",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

