/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: maskfits.cpp 147 2010-11-18 17:02:10Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include "fitstools.h"

// structure for flag values
class flagValues
{
   public:
      flagValues() : region(2) { };
      double circle[2];
      int radius;
      Region region; // region class defined in util/region.h

      bool accumulate;
      bool invert;

      float kappa;

      string inname;
      string outname;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

template<class T>
class stencil{
   public:
      MArray<T, 2> getCircle(const int radius) const;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   MArray<int, 2> cstencil;
   if(flagVal.circle[0])
   {
      const int x = int(flagVal.circle[0] + 0.5);
      const int y = int(flagVal.circle[1] + 0.5);      
      flagVal.region.setRange(1, x - flagVal.radius,
                              x + flagVal.radius);
      flagVal.region.setRange(2, y - flagVal.radius,
                              y + flagVal.radius);
      stencil<int> sten;
      cstencil.makeReference(sten.getCircle(flagVal.radius));
      if(flagVal.invert)
         cstencil = (cstencil-1) * (-1);
   }

   // instance for FitsTools class methods
   FitsTools<float, 2> ft(flagVal.verbose, flagVal.ignore_header);

   const list<string> files = comline.fileArguments(); // get the filenames
   if((flagVal.inname.length() == 0) && (files.size() < 1)) print_usage(flags);

   string to_history = "";
   if(flagVal.accumulate)
      to_history += "cumulative ";
   to_history += "mask of";
   if(flagVal.region.getStart(1) != 0)
      to_history += string(" the region ") + flagVal.region.toString();
   
   MArray<int, 2> mask;
   if(flagVal.inname.length() > 0)
   {
      FitsIn inmask(flagVal.inname, flagVal.quiet_please, false);
      inmask >> mask;
      to_history += string(" starting from mask ") + flagVal.inname;
   }
   else
   {
      FitsHeader hdr(*(files.begin()), true, true);   
      mask.makeReference( MArray<int, 2>(Range(1, hdr.getNaxis(1)),
                                         Range(1, hdr.getNaxis(2))) );
      mask = 0;
   }
   MArray<int, 2> rmask;
   {
      FitsTools<int, 2> it;
      it.referenceRegion(rmask, mask, flagVal.region);
   }

   if(flagVal.kappa != 0.0f)
   {
      to_history += ": ";
      // for all FILES
      for( list<string>::const_iterator file_i=files.begin();
           file_i != files.end();
           ++file_i ) {
         try{
            // arrays for input and output
            MArray<float, 2> a, ra, e_a, e_ra;
            
            FitsIn infile(*file_i,
                          flagVal.quiet_please, flagVal.ignore_header);
            ft.loadImage(infile, a);
            if(!a.isConformable(mask))
               throw FitsException("file does not conform (NAXIS mismatch)");
            ft.referenceRegion(ra, a, flagVal.region);
            try{
               ft.loadImage(ft.prependFilename(*file_i, "e."), e_a);
            }
            catch (exception &e){
               if(flagVal.verbose > 0)
                  cerr << e.what() << " estimating error with SQRT()" <<endl;
               e_a.makeReference( MArray<float, 2>(sqrt(fabs(a))) );
            }
            ft.referenceRegion(e_ra, e_a, flagVal.region);
            
            if(flagVal.invert)
            {
               if(flagVal.accumulate)
                  rmask += ( (fabs(ra) < flagVal.kappa * e_ra) &&
                             (e_ra > 0.0f) );
               else
                  rmask |=  ( (fabs(ra) < flagVal.kappa * e_ra) &&
                              (e_ra > 0.0f) );
            }
            else
            {
               if(flagVal.accumulate)
                  rmask += ( (fabs(ra) >= flagVal.kappa * e_ra) &&
                             (e_ra > 0.0f) );
               else
                  rmask |=  ( (fabs(ra) >= flagVal.kappa * e_ra) &&
                              (e_ra > 0.0f) );
            }
            to_history += *file_i + ' ';
            
         } // this file should have worked, otherwise ->
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << e.what() << endl;
         }
      }
      
   }
   else
   {
      if(cstencil.isAllocated())
      {
         if(flagVal.invert)
            rmask *= cstencil;
         else
         {
            if(flagVal.accumulate)
               rmask += int(files.size()) * cstencil;
            else
               rmask = cstencil;
         }            
      }
      else
      {
         if(flagVal.invert)
            rmask = 0;
         else
         {
            if(flagVal.accumulate)
               rmask += files.size();
            else
               rmask = 1;
         }
      }
   }
   mask -= 32768;

   FitsOut outfile(flagVal.outname + string(".fits"));
   outfile.setBscale(1);
   outfile.setBzero(32768);
   outfile.addHistory(to_history);
   if(flagVal.verbose > 0)
      cerr << "Saving " << outfile.getFilename() << endl;
   outfile.setBitpixOut(16);
   outfile << mask;

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

template<class T>
MArray<T, 2> stencil<T>::getCircle(const int radius) const
{
   const int horizon2 = radius * radius + 1;
   MArray<T, 2> circle(Range(-radius, radius), Range(-radius, radius));
   circle = cast<T>()( (pow2(indexPosInt(circle, 1)) + pow2(indexPosInt(circle, 2))) <= horizon2 );
   return circle;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: maskfits [-FLAGS [option]] FILE(s)\n") +
      string("build (cumulative) mask of FILE(s)\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("accumulate", "FALSE",
                                     "build a cumulative mask",
                                     'a', &flagVal.accumulate ) );
      flags.addOption(new FloatOption("kappa", "0.0", "mask true where signal >= kappa * error",
                                      'k', &flagVal.kappa) );
      flags.addOption(new BoolOption("invert", "FALSE",
                                     "invert condition",
                                     'i', &flagVal.invert ) );      
      flags.addOption(new DoubleArrayOption("circle", "0.0,0.0",
                                            "center coordinates of circular region",
                                            'c', 2, flagVal.circle ) );
      flags.addOption(new IntOption("radius", "1",
                                    "radius of circular region",
                                    0, &flagVal.radius ) );
      flags.addOption(new StringOption("inmaskfile", "", "start building from inmaskfile",
                                       'f', &flagVal.inname) );
      flags.addOption(new StringOption("outmask", "mask",
                                       "name of output mask file",
                                       'o', &flagVal.outname) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform only within region",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

