/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: meanfits.cpp 171 2018-02-27 13:39:30Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2008 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT float

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#ifdef UTIL_USING_NAMESPACE
using namespace util;
#endif

#include <ltl/fitsio.h>
#include <ltl/statistics.h>

#ifdef LTL_USING_NAMESPACE
using namespace ltl;
#endif

#include <string>
#include <iostream>
#include <list>

using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues()
      { };
      bool median;
      bool new_error;
      bool sigma;
      bool twist;
      FT kappa;
      bool nan;
      FT nan_value;
      //Region region; // region class defined in util/option.h
      string exptime;
      string juldate;
      string prefix;
      bool zero;
      int maxmem;
      int verbose;
      bool quiet_please;
      bool ignore_header;
      string outfile;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

inline string prependFilename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main( int argc, char *argv[] ) 
{
   // class for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 2) print_usage(flags);
   
   try{
      const int n_files = files.size();
      const double kappa_correction( erf( flagVal.kappa / M_SQRT2 ) );

      string history = "";
      if(flagVal.kappa)
         history += "kappa sigma clipped ";
      if(flagVal.twist)
         history += string("(respective to ") +
            (flagVal.median ? string("average") : string("median")) +
            string(") ");
      history += (flagVal.median ? string("median") : string("average")) +
         string(" of:");

      if(flagVal.verbose > 0)
         cerr << "Building " << history << endl;

      // open input files
      list<FitsIn> infiles;
      list<FitsIn> inerrors;
      bool have_errors = true;
      if(flagVal.new_error) have_errors = false;
      list<string>::const_iterator file_i = files.begin();
      while( file_i != files.end() )
      {
         const string filename = *file_i;
         if(flagVal.verbose > 0)
            cerr << filename << endl;
         infiles.push_back( FitsIn(filename,
                                   flagVal.quiet_please,
                                   flagVal.ignore_header) );
         if(have_errors)
         {
            const string errorname = prependFilename(filename, "e.");
            try
            {
               inerrors.push_back( FitsIn(errorname,
                                          flagVal.quiet_please,
                                          true) );
            }
            catch (exception &e){
               cerr << e.what() << endl;
               have_errors = false;
            }
         }
         history += string(" ") + filename;
         ++file_i;
      }

      const list<FitsIn>::iterator first_error = inerrors.begin();
      const list<FitsIn>::iterator first_file = infiles.begin();
      const int xlength = (*(first_file)).getNaxis(1);
      const int ylength = (*(first_file)).getNaxis(2);

      // estimate reasonable number of lines to read in one run
      // assume at least ~400 MB workspace / 4(8) bytes per pixel =>
      // 50 Mpixel / (NAXIS1 * n_files)
      const int n_lines = (flagVal.maxmem * 1024 * 1024) / (n_files * xlength * 8);

      //MArray<FT, 2> files_rows(n_files, xlength);
      //MArray<FT, 2> errors_rows(n_files, xlength);
      MArray<FT, 3> files_rows(n_files, xlength, n_lines);
      MArray<FT, 3> errors_rows(n_files, xlength, n_lines);
      MArray<FT, 2> outarray(xlength, ylength);
      MArray<FT, 2> outerror(xlength, ylength);
      MArray<FT, 2>::iterator out_i = outarray.begin();
      MArray<FT, 2>::iterator out_e = outerror.begin();

      MArray<FT, 2> refbholder; // for image block
      MArray<FT, 2> e_refbholder;
      MArray<FT, 1> refholder; // for one pixel of all files
      MArray<FT, 1> e_refholder;

      // loop over all n_lines rows blocks of all files
      Region y_block(2);
      y_block.setRange(1, 1, xlength);
      for(int y = 1; y <= ylength; y+=n_lines) // ++y)
      {
         y_block.setRange(2, y, (y+n_lines-1<=ylength) ? y+n_lines-1 : ylength);
         if(flagVal.verbose > 1)
            //if(!(y%10))
            cerr << "lines from " << y << " up to " << y_block.getEnd(2) << endl;
         list<FitsIn>::iterator infile_i = first_file;
         list<FitsIn>::iterator inerror_i = first_error;

         // read y-th row block of all files
         for(int z = 1; z <= n_files; ++z)
         {
            refbholder.makeReference( files_rows(z, Range::all(), Range(1, y_block.getLength(2))) );
            (*infile_i).setRegion(y_block);
            if(flagVal.verbose > 2)
            {
               cerr << (*infile_i).getFilename();
               if((*infile_i).getRegionLength())
                  cerr << " " << (*infile_i).getRegion().toString();
               cerr << endl;
            }
            (*infile_i) >> refbholder;
            ++infile_i;
            if(have_errors)
            {
               e_refbholder.makeReference( errors_rows(z, Range::all(), Range(1, y_block.getLength(2))) );
               (*inerror_i).setRegion(y_block);
               (*inerror_i) >> e_refbholder;
               ++inerror_i;
               e_refbholder = merge(e_refbholder > FT(0.0),
                                   pow2(e_refbholder),
                                   e_refbholder);
            }
         }

         // evaluate mean for every pixel of a row (i.e. block) in all files
         for(int nl = 1; nl <= n_lines && out_i!=outarray.end(); ++nl)
         for(int x = 1; x <= xlength && out_i!=outarray.end(); ++x)
         {
            refholder.makeReference( files_rows(Range::all(), x, nl) );
	    if(have_errors)  e_refholder.makeReference( errors_rows(Range::all(), x, nl) );
            double mean, sigma;

            if(flagVal.median)
            {
               if(flagVal.kappa) // clipped
               {
                  int n; // for number of clipping remnants
                  if(flagVal.nan) // with nan
                  {
                     if(flagVal.twist)
                        n = kappa_average_median( refholder, flagVal.kappa,
                                                  flagVal.nan_value, &mean, &sigma );
                     else
                        n = kappa_sigma_median( refholder, flagVal.kappa,
                                                flagVal.nan_value, &mean, &sigma );
                  }
                  else // without nan
                  {
                     if(flagVal.twist)
                        n = kappa_average_median( refholder, flagVal.kappa,
                                                  &mean, &sigma );
                     else
                        n = kappa_sigma_median( refholder, flagVal.kappa,
                                                &mean, &sigma );
                  }
                  sigma /= kappa_correction;
               }
               else // not clipped
               {
                  if(flagVal.nan) // with nan
                  { 
                     mean = median_exact( refholder, flagVal.nan_value );
                     const int n_m1 = count(refholder != flagVal.nan_value) - 1;
                     if(n_m1 > 0)
                        sigma =
                           sqrt( sum( merge(refholder != flagVal.nan_value,
                                            pow2(refholder - mean),
                                            FT(0.0)) ) ) / double(n_m1);
                     else
                        sigma = 0.0;
                  }
                  else // without nan
                  {
                     mean = median_exact( refholder );
                     const size_t n_m1 = refholder.nelements() - 1;
                     if(n_m1 > 0)
                        sigma = sqrt( sum( pow2(refholder - mean) ) ) / double(n_m1);
                     else
                        sigma = 0.0;
                  }
               }
            }
            else
            {
               if(flagVal.kappa) // clipped
               {
                  int n; // for number of clipping remnants
                  if(flagVal.nan) // with nan
                  {
                     if(flagVal.twist)
                        n = kappa_median_average( refholder, flagVal.kappa, flagVal.nan_value,
                                                  &mean, &sigma );
                     else
                        n = kappa_sigma_average( refholder, flagVal.kappa, flagVal.nan_value,
                                                 &mean, &sigma );
                  }
                  else // without nan
                  {
                     if(flagVal.twist)
                        n = kappa_median_average( refholder, flagVal.kappa,
                                                  &mean, &sigma );
                     else
                        n = kappa_sigma_average( refholder, flagVal.kappa,
                                                 &mean, &sigma );
                  }
                  sigma /= kappa_correction;
               }
               else // not clipped
               {
                  if(flagVal.nan) // with nan
                     sigma = stddev( refholder, flagVal.nan_value, &mean );
                  else sigma = stddev( refholder, &mean ); // without nan
               }
            }
            if( !flagVal.sigma ) sigma /= sqrt(double(refholder.nelements())); // use SEM

            if(have_errors) // use error from files
            {
               if(flagVal.nan)
               {
                  const FT e_min = min(e_refholder);
                  if(e_min < FT(0.0)) // dominant error
                     *out_e = e_min;
                  else // error of mean ignoring NaNs
                  {
                     const double ssum = sqrt(sum( e_refholder ));
                     const int n_valid = count(e_refholder > FT(0.0));
                     *out_e = FT(ssum) / FT(n_valid);
                  }
               }
               else // simple error of mean
                  *out_e = FT(sqrt(sum(e_refholder))) /
                     FT(e_refholder.nelements());
            }
            else // no error (at least one file missing or preset) -> use sem
               *out_e = FT( sigma );
            ++out_e;
            *out_i = FT( mean );
            ++out_i;
         }
      }

      // free some memory
      refholder.free();
      e_refholder.free();
      
      files_rows.free();

      const string outname = (flagVal.outfile == "" ?
                              prependFilename( (*first_file).getFilename(), 
                                               flagVal.prefix ) : flagVal.outfile);
      FitsOut outfile(outname, *first_file, flagVal.quiet_please);

      if(flagVal.juldate.size() > 0)
      {
         double juldate = 0.0;
         list<FitsIn>::const_iterator infile_i = first_file;
         const string jdkey = (*first_file).adjustKeyword(flagVal.juldate);
         const list<FitsIn>::const_iterator infile_end = infiles.end();
         size_t n_jd = size_t(0);
         while(infile_i != infile_end)
         {
            try{
               juldate += (*infile_i).getFloat(jdkey);
               ++n_jd;
            }
            catch(FitsException e){
               if(flagVal.verbose > 1)
                  cerr << e.what() << endl;
            }
            ++infile_i;
         }
         juldate /= double(n_jd);
         outfile.addValueCard(jdkey, juldate,
                              string("average Julian Date"));
      }

      if(flagVal.exptime.size() > 0)
      {
         double exptime = 0.0;
         list<FitsIn>::const_iterator infile_i = first_file;
         const string exptkey = (*first_file).adjustKeyword(flagVal.exptime);
         const list<FitsIn>::const_iterator infile_end = infiles.end();
         while(infile_i != infile_end)
         {
            try{
               exptime += (*infile_i).getFloat(exptkey);
            }
            catch(FitsException e){
               if(flagVal.verbose > 1)
                  cerr << e.what() << endl;
            }
            ++infile_i;
         }
         outfile.addValueCard(exptkey, exptime,
                              string("summed exposure time"));
      }
      
      infiles.clear();

      outfile.addHistory(history);
      if(flagVal.verbose > 0)
         cerr << "Saving " << outname << endl;
      outfile << outarray;
      const string outerrorname = prependFilename( outfile.getFilename(), "e." );
      outfile.setFilename(outerrorname);
      if(flagVal.verbose > 0)
         cerr << "Saving " << outerrorname << endl;
      outfile << outerror;
   }
   catch (exception &e){
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
   }
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: meanfits [-FLAGS [option]] FILES\n") +
      string("build a mean of FILES, keywords and filename taken from 1st file\n\n");

   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("median", "FALSE",
                                     "build median (not average which is default)",
                                     'm', &flagVal.median ) );
      flags.addOption(new BoolOption("new_error", "FALSE",
                                     "ignore input error files and use sem/sigma for new error which is default if at least one error file is missing",
                                     0, &flagVal.new_error ) );
      flags.addOption(new BoolOption("sigma", "FALSE",
                                     "add (kappa corrected) sigma to error; default is using just (kappa corrected) standard error of average/median",
                                     's', &flagVal.sigma ) );
      flags.addOption(new BoolOption("twist", "FALSE",
                                     "if average clip respective to median, if median clip respective to average",
                                     't', &flagVal.twist ) );
      flags.addOption(new FloatOption("kappa", "0.0",
                                      "for kappa sigma clipping, default: no clipping",
                                      'k', &flagVal.kappa) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("exposuretime", "",
                                       "also sum up value of FITS keyword exposuretime",
                                       'e', &flagVal.exptime) );
      flags.addOption(new StringOption("juldate", "",
                                       "calculate mean observation julian date using FITS keyword matching juldate",
                                       'j', &flagVal.juldate) );
      flags.addOption(new StringOption("outfile", "",
                                       "Output filename. Prefix + first file if none given.",
                                       'o', &flagVal.outfile) );
      flags.addOption(new StringOption("prefix", "m", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("maxmem", "256",
                                    "do not use more than maxmem (MB) memory => use tiled file reads",
                                    0 , &flagVal.maxmem ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

