/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: multiplyfits.cpp 143 2010-04-26 17:30:18Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT float

#include "fitstools.h"


// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) :
         region(dimension), of_region(dimension), verbose(1) { };
      FT multiplier[2]; // [0] for multiplier, [1] for error
      string multiplier_file;
      FT nan_value;
      bool nan;
      bool median;
      bool average;
      bool sigma;
      bool twist;
      double kappa;
      Region region; // region class defined in util/option.h
      Region of_region; // region class defined in util/option.h
      string prefix;
      string usekey;
      string useerror;
      string fitskey;
      string fitsfilekey;
      bool zero;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define(s) to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   // for file multiplication: readfile
   MArray<FT, dimension> b, e_b, rb, e_rb;
   if( flagVal.multiplier_file.length() > 0 )
   {
      const bool have_error =
         ft.loadImageError2( flagVal.multiplier_file, b, e_b );
   
      // get region to operate on, default is whole array
      ft.referenceRegion(rb, b, flagVal.region);
      if(have_error)
      {
         ft.referenceRegion(e_rb, e_b, flagVal.region);
         ft.convert2RelativeError2(e_rb, rb);
      }
   }
   
   flagVal.multiplier[1] *= flagVal.multiplier[1]; // square constant error

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<FT, dimension> a, e_a;

         // multiplier for this run
         FT d = flagVal.multiplier[0];
         FT e_d = flagVal.multiplier[1];

         // open IO-file(s), read data from file
         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         const bool have_error = ft.loadImageError2(infile, a, e_a);

         // get region to operate on, default is whole array
         MArray<FT, dimension> ra, e_ra;
         ft.referenceRegion(ra, a, flagVal.region);

         // if no input error start with zero
         if( !have_error )
         {
            e_a.makeReference( MArray<FT, dimension>( a.shape()) );
            e_a = 0.0f;
         }
         ft.referenceRegion( e_ra, e_a, flagVal.region );

         if( have_error )
            ft.convert2RelativeError2( e_ra, ra,
                                       flagVal.nan, flagVal.nan_value );

         // prepare outfile
         FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                         infile, flagVal.quiet_please);
         {
            string to_history = string("multiplied");
            if(flagVal.region.getStart(1))
               to_history += string(" region ") + flagVal.region.toString();
            to_history +=string(" with:");
            if(flagVal.verbose > 0)
               cerr << to_history << endl;
            outfile.addHistory(to_history);
         }

         // get multipliers apart from file

         if( (d!=1.0f) || e_d ) // get constant
         {
            char buf[80];
            sprintf(buf, "constant: %g with error %g", d, sqrt(e_d));
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }

         // build relative error
         if( d ) e_d /= ( d * d );

         // get value of FITS key
         if( ( flagVal.usekey.length() > 0 ) || ( flagVal.useerror.length() > 0 ) )
         {
            double holder = 0.0, errorholder = 0.0;
            char buf[128];
            strcpy(buf,"values of FITS keys: ");
            if( flagVal.usekey.length() > 0 )
            {
               holder = infile.getFloat( infile.adjustKeyword(flagVal.usekey) );
               d *= FT(holder);
               sprintf(buf + strlen(buf), "%s = %g",
                       flagVal.usekey.c_str(), holder);
               if( flagVal.useerror.length() > 0 ) strcat(buf, ", with ");
            }
            if( flagVal.useerror.length() > 0 )
            {
               errorholder = infile.getFloat( infile.adjustKeyword(flagVal.useerror) );
               e_d += FT(errorholder * errorholder / (holder * holder));
               sprintf(buf + strlen(buf), "error %s = %g",
                       flagVal.useerror.c_str(), errorholder);
            }
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }
         
         // get median or/and average of region
         if(flagVal.median || flagVal.average)
         {
            MArray<FT, dimension> m;
            ft.referenceRegion(m, a, flagVal.of_region);
            string result = "";
            
            if(flagVal.median) // median
            {
               double median, sigma;
               result += ft.medianArray(median, sigma, m, flagVal.kappa,
                                        flagVal.sigma, flagVal.twist,
                                        flagVal.nan, flagVal.nan_value);
               d   *= FT(median); // multiply to constant
               const double rel_error = sigma / median;
               e_d += FT(rel_error * rel_error); // add to error
            }
            
            if(flagVal.average) // average
            {
               double average, sigma;
               result += ft.averageArray(average, sigma, m, flagVal.kappa,
                                         flagVal.sigma, flagVal.twist,
                                         flagVal.nan, flagVal.nan_value);
               d   *= FT(average); // multiply to constant
               const double rel_error = sigma / average;
               e_d += FT(rel_error * rel_error); // add to error
            }

            if(flagVal.of_region.getStart(1))
               result += string(" of region ") + flagVal.of_region.toString();
            if(flagVal.verbose > 0)
               cerr << result << endl;
            outfile.addHistory(result);
         }
         
         //  multiply with file?
         if( b.isAllocated() )
         {
            const string buf = string("file: ") + flagVal.multiplier_file;
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }

         // calculate multiplied file
         ft.multiplyArrayConstant(ra, rb, d, flagVal.nan, flagVal.nan_value);

         // if requested write multiplier constant to key
         if( flagVal.fitskey.length() > 0 )
            outfile.addValueCard(flagVal.fitskey, d,
                                 "used for multiplier constant");

         // if requested write multiplier file name to key
         if( flagVal.fitsfilekey.length() > 0 )
            outfile.addValueCard(flagVal.fitsfilekey, flagVal.multiplier_file,
                                 "used for file multiplication");
         
         // calculate error
         if(flagVal.verbose > 0)
         {
            cerr << "Calculating error";
            cerr.flush();
         }
         if( (!have_error) || flagVal.zero ) // no initial error in file -> zero not dominant
         {
            if( e_b.isAllocated() || e_d ) // must I really?
               ft.propagateZeroError2(e_ra, e_rb, e_d);
         }
         // if there was a error given -> propagate with
         // negative or zero in file(s) dominant
         else
         {
            if( e_b.isAllocated() || e_d ) // must I really?
               ft.propagateError2(e_ra, e_rb, e_d);
            // save error to file anyways
         }
         if( have_error || e_b.isAllocated() || e_d )
         {
            if(flagVal.verbose > 0)
               cerr << "." << endl;
            ft.convert2AbsoluteError2(e_ra, ra,
                                      flagVal.nan, flagVal.nan_value);
            ft.saveImageError2(a, e_a, outfile);
         }
         else
         {
            e_ra.free(); e_a.free();
            if(flagVal.verbose > 0)
               cerr << ", but dropping it because it is zero." << endl;
            ft.saveImage(a, outfile);
         }

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: multiplyfits [-FLAGS [option]] FILES\n") +
      string("multiply FILES with flag dependent value or file\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("file", "", "multiply with file",
                                       'f', &flagVal.multiplier_file) );
      flags.addOption(new FloatArrayOption("constant", "1.0,0.0", "multiplier[, error]",
                                           'c', 2, flagVal.multiplier ) );
      flags.addOption(new BoolOption("average", "FALSE",
                                     "multiply with (clipped) average of region",
                                     'a', &flagVal.average ) );
      flags.addOption(new BoolOption("median", "FALSE",
                                     "multiply with (clipped) median of region",
                                     'm', &flagVal.median ) );
      flags.addOption(new DoubleOption("kappa", "0.0",
                                      "for kappa sigma clipping, default: no clipping",
                                      'k', &flagVal.kappa) );
      flags.addOption(new BoolOption("sigma", "FALSE",
                                     "add (kappa corrected) sigma to error; default is using just (kappa corrected) standard error of average/median",
                                     's', &flagVal.sigma ) );
      flags.addOption(new BoolOption("twist", "FALSE",
                                     "if average clip respective to median, if median clip respective to average",
                                     't', &flagVal.twist ) );
      flags.addOption(new RegionArrayOption("of_region", "0:0,0:0",
                                            "use x_start:x_end,y_start:y_end region to determine average/median",
                                            'o', 2, &flagVal.of_region) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform only on region, leave rest of frame(s) untouched",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new StringOption("use_key", "", "multiply with value of FITS keyword",
                                       0, &flagVal.usekey) );
      flags.addOption(new StringOption("use_error", "", "add to error value of FITS keyword",
                                       0, &flagVal.useerror) );
      flags.addOption(new StringOption("write_key", "", "write multiplier constant to FITS key",
                                       0, &flagVal.fitskey) );
      flags.addOption(new StringOption("write_file_key", "", "write multiplier file name to FITS key",
                                       0, &flagVal.fitsfilekey) );
      flags.addOption(new StringOption("prefix", "d", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("zero_error_not_dominant", "FALSE",
                                     "do NOT keep zero error dominant (only sometimes makes sense)",
                                     'z', &flagVal.zero ) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

