/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2008 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#include <ltl/fitsio.h>

using namespace ltl;
using namespace util;
using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues() : region(2) { };
      bool boolean;
      bool inv_boolean;
      int integer;
      float floatprec;
      double doubleprec;
      string str;
      
      int int_array[5];
      float float_array[2];
      double double_array[3];
      Region region; // region class defined in util/region.h

      float nan_value;
      bool nan;
      string prefix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<float, 2> a;

         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
         infile >> a; infile.freeData();
         a = merge(a==a, a, flagVal.nan_value);

         const string outname = prepend_filename(*file_i, flagVal.prefix);
         FitsOut outfile(outname,
                         infile, flagVal.quiet_please);
         outfile.addHistory(string("Replaced NaN within ") + *file_i);
         if(flagVal.verbose > 0)
            cerr << "Saving " << outname << endl;
         outfile << a; a.free();
         
      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: rmnan [-FLAGS [option]] FILES\n") +
      string("Replace NaN with some value in FILES\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2008 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as replacement for NaN",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("prefix", "n", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

