/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: rotrevfits.cpp 94 2005-02-21 12:18:38Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe of any but output will be of type FT
#define FT float

#include "fitstools.h"


// structure for flag values
class flagValues
{
   public:
      bool x;
      bool y;
      int rot;
      string prefix;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // class for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // array for input and output
         MArray<FT, dimension> a;

         // open file
         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         // read data from file
         ft.loadImage(infile, a);
         FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                         infile, flagVal.quiet_please);

         const string to_history =
            ft.rotateReverseArray( a, flagVal.rot, flagVal.x, flagVal.y );
         if(flagVal.verbose > 0)
            cerr << to_history << endl;
         outfile.addHistory(to_history);

         // save file
         ft.saveImage(a, outfile);
         a.free();

         try{
            // open file
            FitsIn inerror(ft.prependFilename(*file_i, "e."),
                           flagVal.quiet_please, true);
            ft.loadImage(inerror, a);
            outfile.setFilename( ft.prependFilename(outfile.getFilename(),
                                                    "e.") );
            ft.rotateReverseArray( a, flagVal.rot, flagVal.x, flagVal.y );
            if(flagVal.verbose > 0)
               cerr << "Applied operation also on error image" << endl;
            ft.saveImage(a, outfile);
         }
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << "No errorfile\n";
         }

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: rotrevfits [-FLAGS [option]] FILES\n") +
      string("rotate image (and error image) by 90, 180 or 270 deg.\n") +
      string("or/and (subsequently) reverse x (flip left side right)\n") +
      string("and/or y (flip upside down) coordinates \n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("reverse_x", "FALSE",
                                     "reverse x coordinates",
                                     'x', &flagVal.x ) );
      flags.addOption(new BoolOption("reverse_y", "FALSE",
                                     "reverse y coordinates",
                                     'y', &flagVal.y ) );
      flags.addOption(new IntOption("rotate", "0",
                                    "rotate clockwise by 90, 180 or 270 ",
                                    'r', &flagVal.rot) );
      flags.addOption(new StringOption("prefix", "r", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}
