/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: smoothfits.cpp 147 2010-11-18 17:02:10Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT float

//#define LTL_RANGE_CHECKING
//#define __SMOOTH_DEBUG__

#include "fitstools.h"

#include "smoothing.h"

// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) : region(dimension) { };
      int radius;
      bool circular;
      bool calcerror;
      bool median;
      bool nan;
      FT nan_value;
      Region region; // region class defined in util/option.h
      string prefix;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);
  
   // for radius == 1, circular is a box, so
   if(flagVal.radius == 1)
   {
      if(flagVal.verbose > 0)
         cerr << "For radius == 1, circular equals a box, so circular will be reset to false.\n";
      flagVal.circular = false;
   }

   stencil<FT> sten; // provide an instance for stencil methods

   const int boxlength = 2 * flagVal.radius + 1;
   // for all FILES
   for( list<string>::const_iterator file_i = files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // open IO-file(s)
         FitsIn infile(*file_i, flagVal.quiet_please, flagVal.ignore_header);
         if( (infile.getNaxis(1) < boxlength) ||
             (infile.getNaxis(1) < boxlength) )
            throw FitsException("Image too small for selected radius.");
         MArray<FT, DIMENSION> A;
         ft.loadImage(infile, A);
         MArray<FT, 2> B(A.shape());
         // median smoothing
         if(flagVal.median)
         {
            td_median_stencil<FT> medsmooth(A, flagVal.radius, flagVal.nan_value);
            // circular stencil
            if(flagVal.circular)
            {
               if(flagVal.radius < 1)
               {
                  // this is never true, but just to show how to do with no speed up
                  MArray<FT, 2>::IndexIterator ii = B.indexBegin();
                  const int limit = flagVal.radius * flagVal.radius + 1;
                  while( ! ii.done() ) {
                     MArray<FT, 2> C( A(Range( max((ii(1) - flagVal.radius),
                                                         B.minIndex(1)),
                                                        min((ii(1) + flagVal.radius),
                                                            B.maxIndex(1)) ),
                                                 Range( max((ii(2) - flagVal.radius),
                                                            B.minIndex(2)),
                                                        min((ii(2) + flagVal.radius),
                                                            B.maxIndex(2)) ) ));
                     C.setBase( max((0 - flagVal.radius), (1 - ii(1))), 
                                max((0 - flagVal.radius), (1 - ii(2))) );
                     IndexList<2> l = where( pow2(indexPosInt(C,1)) +
                                                   pow2(indexPosInt(C,2) <= limit) );
                     B(ii(1), ii(2)) =
                        median_exact(C(l));
                     ++ii;
                  }
               }
               else
               {
                  // so you will really go for the most nasty thing ...
                  if(flagVal.nan)
                     B = medsmooth.evaluateCircle_nan();
                  else
                     B = medsmooth.evaluateCircle();             
               }
            }
            // boxy stencil
            else
            {
               // box still can get complicated, if you wanna speed up things
               if(flagVal.radius < 1) {
                  // this is never true, but just to show how to do with no speed up
                  MArray<FT, 2>::IndexIterator ii = B.indexBegin();
                  while( ! ii.done() ) {
                     B(ii(1), ii(2)) =
                        median_exact( A( Range( max((ii(1) - flagVal.radius),
                                                    B.minIndex(1)),
                                                min((ii(1) + flagVal.radius),
                                                    B.maxIndex(1)) ),
                                         Range( max((ii(2) - flagVal.radius),
                                                    B.minIndex(2)),
                                                min((ii(2) + flagVal.radius),
                                                    B.maxIndex(2)) )) );
                     ++ii;
                  }
               }
               else {
                  // it is really tricky to do efficient median smoothing
                  // with bigger radius
                  // so this requires some extra methods, look in smoothing.h
                  if(flagVal.nan)
                     B = medsmooth.evaluateBox_nan();
                  else
                     B = medsmooth.evaluateBox();
               }
            }
         }
         // average smoothing
         else {
            MArray<FT, 2>::IndexIterator ii = B.indexBegin();

            // with circle stencil
            if(flagVal.circular) {
               const MArray<FT, 2> cstencil( sten.getCircle(flagVal.radius) );
#ifdef __SMOOTH_DEBUG__
               cout << cstencil << endl;
#endif
               while( ! ii.done() ) {
                  MArray<FT, 2> C( A(Range( max((ii(1) - flagVal.radius),
                                                B.minIndex(1)),
                                               min((ii(1) + flagVal.radius),
                                                   B.maxIndex(1)) ),
                                        Range( max((ii(2) - flagVal.radius),
                                                   B.minIndex(2)),
                                               min((ii(2) + flagVal.radius),
                                                   B.maxIndex(2)) ) ));
                  // keeping up with special cases, i.e. border is nasty
                  // only need partial stencil?
                  if(C.nelements() < cstencil.nelements()) {
                     // which part?
                     if(C.length(1) < cstencil.length(1)) {
                        // subarray of x
                        if(C.length(2) < cstencil.length(2)) {
                           // and of y ... but which parts ...
                           if( ii(1) <= flagVal.radius ) {
                              // ah ... this part of x ... and of y?
                              if( ii(2) <= flagVal.radius ) {
                                 B(ii(1), ii(2)) =
                                    average( C * cstencil(Range(0 - ii(1),
                                                                cstencil.maxIndex(1)),
                                                          Range(0 - ii(2), cstencil.maxIndex(2))),
                                             flagVal.nan_value);
                              }
                              else {
                                 B(ii(1), ii(2)) =
                                    average( C * cstencil(Range(0 - ii(1),
                                                                cstencil.maxIndex(1)),
                                                          Range(cstencil.minIndex(2),
                                                                C.length(2) - (flagVal.radius + 1))),
                                             flagVal.nan_value);
                              }
                           }
                           else {
                              // other part of x ... and y?
                              if( ii(2) <= flagVal.radius )
                                 B(ii(1), ii(2)) =
                                    average( C * cstencil(Range(cstencil.minIndex(1),
                                                                C.length(1) - (flagVal.radius + 1)),
                                                          Range(0 - ii(2),
                                                                cstencil.maxIndex(2))),
                                             flagVal.nan_value);
                              else {
                                 B(ii(1), ii(2)) =
                                    average( C * cstencil(Range(cstencil.minIndex(1),
                                                                C.length(1) - (flagVal.radius + 1)),
                                                          Range(cstencil.minIndex(2),
                                                                C.length(2) - (flagVal.radius + 1))),
                                             flagVal.nan_value);
                              }
                           }
                        }
                        // but no subarray of y
                        else {
                           // so which part of x?
                           if( ii(1) <= flagVal.radius)
                              B(ii(1), ii(2)) =
                                 average( C * cstencil(Range(0 - ii(1),
                                                             cstencil.maxIndex(1)),
                                                       Range::all()),
                                          flagVal.nan_value);
                           else // ok other part
                              B(ii(1), ii(2)) =
                                 average( C * cstencil(Range(cstencil.minIndex(1),
                                                             C.length(1) -
                                                             (flagVal.radius + 1)),
                                                       Range::all()),
                                          flagVal.nan_value);
                        }
                     }
                     // hmmm ... need complete x, so only part of y
                     else {
                        // but which part of y?
                        if( ii(2) <= flagVal.radius)
                           B(ii(1), ii(2)) =
                              average( C * cstencil(Range::all(),
                                                    Range(0 - ii(2),
                                                          cstencil.maxIndex(2))),
                                       flagVal.nan_value);
                        else // ok other part
                           B(ii(1), ii(2)) =
                              average( C * cstencil(Range::all(),
                                                    Range(cstencil.minIndex(2),
                                                          C.length(2) -
                                                          (flagVal.radius + 1))),
                                       flagVal.nan_value );
                     }
                  }
                  // no ... need complete stencil
                  // finally the "usual" case no border issues ... so
                  else
                     B(ii(1), ii(2)) = average(C * cstencil, flagVal.nan_value);
                  ++ii;
               }
            }
            // with box stencil, the most easiest case by far
            else {
               if(flagVal.nan) {
                  while( ! ii.done() ) {
                     B(ii(1), ii(2)) =
                        average( A( Range( max((ii(1) - flagVal.radius),
                                               B.minIndex(1)),
                                           min((ii(1) + flagVal.radius),
                                               B.maxIndex(1)) ),
                                    Range( max((ii(2) - flagVal.radius),
                                               B.minIndex(2)),
                                           min((ii(2) + flagVal.radius),
                                               B.maxIndex(2)) )),
                                 flagVal.nan_value );
                     ++ii;
                  }
               }
               else {
                  while( ! ii.done() ) {
                     B(ii(1), ii(2)) =
                        average( A( Range( max((ii(1) - flagVal.radius),
                                               B.minIndex(1)),
                                           min((ii(1) + flagVal.radius),
                                               B.maxIndex(1)) ),
                                    Range( max((ii(2) - flagVal.radius),
                                               B.minIndex(2)),
                                           min((ii(2) + flagVal.radius),
                                               B.maxIndex(2)) )));
                     ++ii;
                  }
               }
            }
         }

         const string outfilename = ft.prependFilename(*file_i, flagVal.prefix);
         FitsOut outfile(outfilename, infile, flagVal.quiet_please);
         ft.saveImage(B, outfile);

         if(flagVal.calcerror)
         {
            const string outerrorname = ft.prependFilename(outfilename, "e.");
            outfile.setFilename(outerrorname);      
            MArray<FT, 2> C(calculateSEM(B, A,
                                               flagVal.radius, flagVal.circular,
                                               flagVal.nan, flagVal.nan_value));
            ft.saveImage(C, outfile);
         }

         //           try{
         //              const string errorname = prepend_filename(*file_i, "e.");
         //              cerr<<"reading "<<errorname<<endl;
         //              FitsIn inerror(errorname);

         //              const string outerrorname = prepend_filename(outfilename,"e.");
         //              cerr<<"saving "<<outerrorname<<endl;
         //              outfile.setFilename(outerrorname);

         //           }
         //           catch (exception &e){
         //              cerr<<"no errorfile\n";
         //           }
      }
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: smoothfits [-FLAGS [option]] FILES\n") +
      string("smooth data of FILES\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new IntOption("radius", "1",
                                    "distance to smooth horizon",
                                    'a', &flagVal.radius) );
      flags.addOption(new BoolOption("circular", "FALSE",
                                     "use circular stencil",
                                     'c', &flagVal.circular) );
      flags.addOption(new BoolOption("error", "FALSE",
                                     "calculate SEM error image",
                                     'e', &flagVal.calcerror) );
      flags.addOption(new BoolOption("median", "FALSE",
                                     "use median, default is using average (median has no nan option)",
                                     'm', &flagVal.median) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan, so occasional bad pixel won't hurt",
                                      0, &flagVal.nan_value) );
      flags.addOption(new RegionArrayOption("region", "",
                                            "NAXIS1_start:NAXIS1_end,...,NAXISnaxis_start:NAXISnaxis_end",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new StringOption("prefix", "s.", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}
