/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: smoothing.h 167 2016-05-10 08:18:46Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// smoothing classes

// median smooth stencil:
// can still be optimized by sorting pointers to and not
// the fpixels themselves

// getCircle: returns a type T Circle in an MArray (inside 1, outside 0)


#ifndef __SMOOTHING_H__
#define __SMOOTHING_H__

//#define __SMOOTH_DEBUG__  // just for debugging

#ifndef __FITSTOOLS_H__
#include <ltl/marray.h>
#endif

#ifdef __SMOOTH_DEBUG__
#include <ltl/marray_io.h>
#include <iostream>
#endif

// STL headers
#include <vector>
#include <algorithm>

template<class T>
class fpixel
{
   public:
      fpixel(const T inv, const int inx, const int iny) :
         value(inv), x(inx), y(iny) { };
      //  fpixel(const fpixel<T> & in) :
      //    value(in.value), x(in.x), y(in.y) { };
      fpixel() { };
      T value;
      int x;
      int y;
};

// less for pixel with coordinates
template<class T>
inline bool operator<(const fpixel<T> & left, const fpixel<T> & right)
{ return (left.value < right.value); }


template<class T, int DIM>
class eq_dim_fp
{
   public:
      eq_dim_fp() { }
      inline bool operator()(const fpixel<T> & p, const int i) const
      {
         if(DIM == 1) return p.x == i;
         if(DIM == 2) return p.y == i;
      }
      inline bool operator()(const int i, const fpixel<T> & p) const
      {
         if(DIM == 1) return p.x == i;
         if(DIM == 2) return p.y == i;
      }
      inline bool operator()(const fpixel<T> & p, const fpixel<T> & q) const
      {
         if(DIM == 1) return p.x == q.x;
         if(DIM == 2) return p.y == q.y;
      }
};

template<class T, int DIM>
class neq_dim_fp
{
   public:
      neq_dim_fp() { }
      inline bool operator()(const fpixel<T> & p, const int i) const
      {
         if(DIM == 1) return p.x != i;
         if(DIM == 2) return p.y != i;
      }
      inline bool operator()(const int i, const fpixel<T> & p) const
      {
         if(DIM == 1) return p.x != i;
         if(DIM == 2) return p.y != i;
      }
      inline bool operator()(const fpixel<T> & p, const fpixel<T> & q) const
      {
         if(DIM == 1) return p.x != q.x;
         if(DIM == 2) return p.y != q.y;
      }
};


// help class to sort pixel including coordinates
template<class T>
class less_fp_nan // : public binary_function<T, T, T>
{
   public:
      less_fp_nan(const T & nan) : nan_(nan)
      { }
      inline bool operator()(const fpixel<T> & i, const fpixel<T> & j) const
      {
         if( i.value == nan_ ) return false;
         if( j.value == nan_ ) return true;
         return( i < j );

         //return ( (j.value == nan_) || ((i.value < j.value) && (i.value != nan_)) );
      }
      const T nan_;
};


template<class T>
class fpixelptr
{
   public:
      fpixelptr(const fpixel<T> & inpixel) :
         ptr(& inpixel) { };
      fpixel<T> * ptr;
};

template<class T>
inline bool operator<(fpixelptr<T> left, fpixelptr<T> right)
{ return ( left.ptr->value < right.ptr->value ); }


template<class T, int DIM>
void replaceInVec(const typename vector<fpixel<T> >::iterator begin,
		  const typename vector<fpixel<T> >::iterator end,
		  const int i,
		  const fpixel<T> & newpixel);
template<class T, int DIM, class P>
void replaceInVec(const typename vector<fpixel<T> >::iterator begin,
		  const typename vector<fpixel<T> >::iterator end,
		  const int i,
		  const fpixel<T> & newpixel,
		  const P & pred);

template<class T, int DIM>
void replaceVecCopy(const typename vector<fpixel<T> >::iterator newbegin,
		    const typename vector<fpixel<T> >::iterator newend,  
		    const typename vector<fpixel<T> >::const_iterator oldbegin,
		    const typename vector<fpixel<T> >::const_iterator oldend,  
		    const int i,
		    const typename vector<fpixel<T> >::const_iterator vecbegin,
		    const typename vector<fpixel<T> >::const_iterator vecend);
template<class T, int DIM, class P>
void replaceVecCopy(const typename vector<fpixel<T> >::iterator newbegin,
		    const typename vector<fpixel<T> >::iterator newend,  
		    const typename vector<fpixel<T> >::const_iterator oldbegin,
		    const typename vector<fpixel<T> >::const_iterator oldend,  
		    const int i,
		    const typename vector<fpixel<T> >::const_iterator vecbegin,
		    const typename vector<fpixel<T> >::const_iterator vecend,
		    const P & pred);


#ifdef __SMOOTH_DEBUG__
template<class T>
void showvec(typename vector<fpixel<T> >::const_iterator i,
	     const typename vector<fpixel<T> >::const_iterator end){
   while(i != end) {
      cerr << (*i).value << " " << (*i).x << " " << (*i).y << endl;
      ++i;
   }
}
#endif

// only for 2-D

template<class T>
class stencil{
   public:
      MArray<T, 2> getCircle(const int radius) const;
};


template<class T>
class td_median_stencil{

   public:
      td_median_stencil(const MArray<T, 2> ina, const int rad, const T nan = T(0)) :
         inarray(ina), outarray(inarray.shape()), outi(outarray.begin()),
         radius(rad),
         maxboxlength(2 * radius + 1), horizon2(radius * radius + 1),
         box(NULL), nextrowbox(NULL), dummybox(NULL),
         collist(inarray.length(1) - (radius + 1)),
         listi(collist.begin()),
         pred(nan),
         done_(false), done_box(false)
      { initStencil(); }

      MArray<T, 2> evaluateBox();
      MArray<T, 2> evaluateBox_nan();
      MArray<T, 2> evaluateCircle();
      MArray<T, 2> evaluateCircle_nan();

   private:
      const MArray<T, 2> inarray;
      MArray<T, 2> outarray;
      typename MArray<T, 2>::iterator outi;
      const int radius;
      const int maxboxlength;
      const int horizon2;
      // bool circle; // maybe same class for circular stencil
      vector<fpixel<T> > * box; // for the actual sorted vector
      vector<fpixel<T> > * nextrowbox; // used for the next row
      vector<fpixel<T> > * dummybox;
      vector<vector<fpixel<T> > > collist;
      typename vector<vector<fpixel<T> > >::iterator listi;
  
      less_fp_nan<T> pred;

      bool done_, done_box;

      int pos[2]; // actual position
      int start[2]; // border of stencil
      int end[2]; //
      // the internal methods
      // 1st:
      void initStencil();
      void initBox();
      void initBox_nan();
      void initBox_common();
      void initColListBox();
      void initColListBox_nan();
      void initColListCircle();
      void initColListCircle_nan();
      void resetX();
      void resetY();
      vector<fpixel<T> > getCol(const int x);
      vector<fpixel<T> > getCol_nan(const int x);
      vector<fpixel<T> > getCol_common(const int x);
      vector<fpixel<T> > getRow(const int y);
      vector<fpixel<T> > getRow_nan(const int y);
      vector<fpixel<T> > getRow_common(const int y);
      vector<fpixel<T> > * addVec(const vector<fpixel<T> > & colvec);
      vector<fpixel<T> > * addVec_nan(const vector<fpixel<T> > & colvec);
      void addCol(const vector<fpixel<T> > & colvec);
      void addCol_nan(const vector<fpixel<T> > & colvec);
      void delCol(const int x);
      vector<fpixel<T> > * delRow(const int y);
      void replaceColInBox(const int x,
                           const vector<fpixel<T> > & newvector);
      void replaceColInBox_nan(const int x,
                               const vector<fpixel<T> > & newvector);
      void replaceColInNewBox(vector<fpixel<T> > & newbox,
                              const vector<fpixel<T> > & oldvector,
                              const int x,
                              const vector<fpixel<T> > & newvector);
      void replaceColInNewBox_nan(vector<fpixel<T> > & newbox,
                                  const vector<fpixel<T> > & oldvector,
                                  const int x,
                                  const vector<fpixel<T> > & newvector);
      void replaceRowInNewBox(vector<fpixel<T> > & newbox,
                              const vector<fpixel<T> > & oldvector,
                              const int y,
                              const vector<fpixel<T> > & newvector);
      void replaceRowInNewBox_nan(vector<fpixel<T> > & newbox,
                                  const vector<fpixel<T> > & oldvector,
                                  const int y,
                                  const vector<fpixel<T> > & newvector);
      void addToColInList(const fpixel<T> & newpixel);
      void addToColInList_nan(const fpixel<T> & newpixel);
      void delFromColInList(const int y);
      void replaceInCol(vector<fpixel<T> > & colvec, const int y,
                        const fpixel<T> & newpixel);
      void replaceInCol_nan(vector<fpixel<T> > & colvec, const int y,
                            const fpixel<T> & newpixel);

      int getActLength(const int i)
      { return (end[i] - start[i] + 1); }

      int getActBoxSize()
      { return (getActLength(0) * getActLength(1)); }
  
      T nthInBox(const int n);

      T medianInBox();
      T medianInBox_nan();
      T medianInBox_common(const size_t length);

      bool isInCircle(const int x, const int y);
      T nthInCircle(const int n);

      T medianInCircle();
      T medianInCircle_nan();
      int medianInCircle_init(typename vector<bool>::iterator isini,
                              const typename vector<bool>::iterator isinend);
      T medianInCircle_common(typename vector<bool>::const_iterator isini,
                              const int length);

      int count_nan();
};

template<class T>
void td_median_stencil<T>::initStencil()
{
   for( int i=1; i<=2; ++i)
      if( maxboxlength > inarray.length(i) )
         throw RangeException("One dimension too small for selected smoothing radius.");
}


// get initial vector (1+radius) X (1+radius)
template<class T>
void td_median_stencil<T>::initBox()
{
   initBox_common();
   sort(box->begin(), box->end());
   if(end[1] < inarray.length(2))
      nextrowbox = addVec(getRow(end[1] + 1));
}
template<class T>
void td_median_stencil<T>::initBox_nan()
{
   initBox_common();
   sort(box->begin(), box->end(), pred);
   if(end[1] < inarray.length(2))
      nextrowbox = addVec_nan(getRow_nan(end[1] + 1));
}

template<class T>
void td_median_stencil<T>::initBox_common()
{
   // init box corners and "hot" spot
   resetY();
   resetX();
   const int boxlength = radius + 1;
   const MArray<T, 2> startarray(inarray(Range(1, boxlength),
                                         Range(1, boxlength)));
   box =
      new vector<fpixel<T> >(boxlength * boxlength);
   typename vector<fpixel<T> >::iterator si = box->begin();
   typename MArray<T, 2>::const_iterator ai = startarray.begin();
   typename MArray<T, 2>::IndexIterator a_ii = startarray.indexBegin();
   while( !ai.done() ) {
      *si = fpixel<T>(*ai, a_ii(1), a_ii(2));
      ++si;
      ++ai; ++a_ii;
   }
}

template<class T>
void td_median_stencil<T>::initColListBox()
{
   // 1st loops to init the colvector list (which actually is a vector)
   // x increasing box
   while( end[0] <= maxboxlength ) {
      *listi = getCol(end[0]);
      addCol(*listi);
      ++listi;

      *outi = medianInBox();
      ++outi; ++pos[0]; ++end[0];
   }

   // 1st allocation of placeholder
   dummybox = new vector<fpixel<T> >(box->size());
   // x constant box
   while(end[0] <= inarray.length(1)) {
      *listi = getCol(end[0]);
      replaceColInBox(start[0], *listi);
      ++start[0];
      ++listi;

      *outi = medianInBox();
      ++outi; ++pos[0]; ++end[0];
   }
   --end[0];
   // x decreasing box
   while(pos[0] <= inarray.length(1)) {
      delCol(start[0]);
      ++start[0];
    
      *outi = medianInBox();
      ++outi; ++pos[0];
   }

   // increase y
   ++pos[1]; ++end[1];
}
template<class T>
void td_median_stencil<T>::initColListBox_nan()
{
   // 1st loops to init the colvector list (which actually is a vector)
   // x increasing box
   while( end[0] <= maxboxlength ) {
      *listi = getCol_nan(end[0]);
      addCol_nan(*listi);
      ++listi;

      *outi = medianInBox_nan();
      ++outi; ++pos[0]; ++end[0];
   }

   // 1st allocation of placeholder
   dummybox = new vector<fpixel<T> >(box->size());
   // x constant box
   while(end[0] <= inarray.length(1)) {
      *listi = getCol_nan(end[0]);
      replaceColInBox_nan(start[0], *listi);
      ++start[0];
      ++listi;

      *outi = medianInBox_nan();
      ++outi; ++pos[0]; ++end[0];
   }
   --end[0];
   // x decreasing box
   while(pos[0] <= inarray.length(1)) {
      delCol(start[0]);
      ++start[0];
    
      *outi = medianInBox_nan();
      ++outi; ++pos[0];
   }

   // increase y
   ++pos[1]; ++end[1];
}

template<class T>
void td_median_stencil<T>::initColListCircle()
{
   // 1st loops to init the colvector list (which actually is a vector)
   // x increasing box
   while( end[0] <= maxboxlength ) {
      *listi = getCol(end[0]);
      addCol(*listi);
      ++listi;

      *outi = medianInCircle();
      ++outi; ++pos[0]; ++end[0];
   }

   // 1st allocation of placeholder
   dummybox = new vector<fpixel<T> >(box->size());
   // x constant box
   while(end[0] <= inarray.length(1)) {
      *listi = getCol(end[0]);
      replaceColInBox(start[0], *listi);
      ++start[0];
      ++listi;

      *outi = medianInCircle();
      ++outi; ++pos[0]; ++end[0];
   }
   --end[0];
   // x decreasing box
   while(pos[0] <= inarray.length(1)) {
      delCol(start[0]);
      ++start[0];
    
      *outi = medianInCircle();
      ++outi; ++pos[0];
   }

   // increase y
   ++pos[1]; ++end[1];
}
template<class T>
void td_median_stencil<T>::initColListCircle_nan()
{
   // 1st loops to init the colvector list (which actually is a vector)
   // x increasing box
   while( end[0] <= maxboxlength ) {
      *listi = getCol_nan(end[0]);
      addCol_nan(*listi);
      ++listi;

      *outi = medianInCircle_nan();
      ++outi; ++pos[0]; ++end[0];
   }

   // 1st allocation of placeholder
   dummybox = new vector<fpixel<T> >(box->size());
   // x constant box
   while(end[0] <= inarray.length(1)) {
      *listi = getCol_nan(end[0]);
      replaceColInBox_nan(start[0], *listi);
      ++start[0];
      ++listi;

      *outi = medianInCircle_nan();
      ++outi; ++pos[0]; ++end[0];
   }
   --end[0];
   // x decreasing box
   while(pos[0] <= inarray.length(1)) {
      delCol(start[0]);
      ++start[0];
    
      *outi = medianInCircle_nan();
      ++outi; ++pos[0];
   }

   // increase y
   ++pos[1]; ++end[1];
}

template<class T>
void td_median_stencil<T>::resetX(){
   pos[0] = start[0] = 1; end[0] = radius + 1;
   listi = collist.begin();
}

template<class T>
void td_median_stencil<T>::resetY(){
   pos[1] = start[1] = 1; end[1] = radius + 1;
   outi.reset();
}


template<class T>
MArray<T, 2> td_median_stencil<T>::evaluateBox()
{
   initBox();

   // first element!
   *outi = medianInBox();
   ++outi; ++pos[0]; ++end[0];

   // 1st loops to init the colvector list (which actually is a vector)
   initColListBox();

   // loop for y increasing box
   while(end[1] <= maxboxlength) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      // "next" box already at max?
      if(end[1] == maxboxlength) { // yes! => replace
         nextrowbox = new vector<fpixel<T> >(box->size());
         replaceRowInNewBox(*nextrowbox, *box, start[1], getRow(end[1] + 1));
      }
      else { // no! => grow
         nextrowbox = addVec( getRow(end[1] + 1) );
      }

      // first element of row!
      *outi = medianInBox();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         addToColInList( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol(*listi);
         ++listi;

         *outi = medianInBox();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         addToColInList( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInBox();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInBox();
         ++outi; ++pos[0];
      }
      // increase y
      ++pos[1]; ++end[1];
   }

   // loop for y constant box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(end[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      if(end[1] == inarray.length(2)) {
         delete box;
         box = nextrowbox;
         nextrowbox = delRow(start[1]+1);
      }
      else {
         vector<fpixel<T> > * const dummy = box;
         box = nextrowbox;
         nextrowbox = dummy;
         replaceRowInNewBox(*nextrowbox, *box, start[1]+1, getRow(end[1] + 1));
      }

      // first element of row!
      *outi = medianInBox();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         replaceInCol( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol(*listi);
         ++listi;

         *outi = medianInBox();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         replaceInCol( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInBox();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInBox();
         ++outi; ++pos[0];
      }
      ++pos[1]; ++end[1];
   }
   --end[1];

   // loop for y decreasing box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(pos[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      if(pos[1] != inarray.length(2)) {
         nextrowbox = delRow(start[1]+1);
      }

      // first element of row!
      *outi = medianInBox();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         delFromColInList( start[1] );
         addCol(*listi);
         ++listi;

         *outi = medianInBox();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         delFromColInList( start[1] );
         replaceColInBox(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInBox();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];

         *outi = medianInBox();
         ++outi; ++pos[0];
      }
      ++pos[1];
   }

   delete box; box = nextrowbox = NULL;

   done_=true;
   done_box=true;

   return outarray;
}

template<class T>
MArray<T, 2> td_median_stencil<T>::evaluateBox_nan()
{
   initBox_nan();

   // first element!
   *outi = medianInBox_nan();
   ++outi; ++pos[0]; ++end[0];

   // 1st loops to init the colvector list (which actually is a vector)
   initColListBox_nan();

   // loop for y increasing box
   while(end[1] <= maxboxlength) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      // "next" box already at max?
      if(end[1] == maxboxlength) { // yes! => replace
         nextrowbox = new vector<fpixel<T> >(box->size());
         replaceRowInNewBox_nan(*nextrowbox, *box, start[1], getRow_nan(end[1] + 1));
      }
      else { // no! => grow
         nextrowbox = addVec_nan( getRow_nan(end[1] + 1) );
      }

      // first element of row!
      *outi = medianInBox_nan();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         addToColInList_nan( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol_nan(*listi);
         ++listi;

         *outi = medianInBox_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         addToColInList_nan( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox_nan(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInBox_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInBox_nan();
         ++outi; ++pos[0];
      }
      // increase y
      ++pos[1]; ++end[1];
   }

   // loop for y constant box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(end[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      if(end[1] == inarray.length(2)) {
         delete box;
         box = nextrowbox;
         nextrowbox = delRow(start[1]+1);
      }
      else {
         vector<fpixel<T> > * const dummy = box;
         box = nextrowbox;
         nextrowbox = dummy;
         replaceRowInNewBox_nan(*nextrowbox, *box, start[1]+1, getRow_nan(end[1] + 1));
      }

      // first element of row!
      *outi = medianInBox_nan();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         replaceInCol_nan( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol_nan(*listi);
         ++listi;

         *outi = medianInBox_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         replaceInCol_nan( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox_nan(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInBox_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInBox_nan();
         ++outi; ++pos[0];
      }
      ++pos[1]; ++end[1];
   }
   --end[1];

   // loop for y decreasing box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(pos[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      if(pos[1] != inarray.length(2)) {
         nextrowbox = delRow(start[1]+1);
      }

      // first element of row!
      *outi = medianInBox_nan();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         delFromColInList( start[1] );
         addCol_nan(*listi);
         ++listi;

         *outi = medianInBox_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         delFromColInList( start[1] );
         replaceColInBox_nan(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInBox_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];

         *outi = medianInBox_nan();
         ++outi; ++pos[0];
      }
      ++pos[1];
   }

   delete box; box = nextrowbox = NULL;

   done_=true;
   done_box=true;

   return outarray;
}

template<class T>
MArray<T, 2> td_median_stencil<T>::evaluateCircle()
{
   initBox();

   // first element!
   *outi = medianInCircle();
   ++outi; ++pos[0]; ++end[0];

   // 1st loops to init the colvector list (which actually is a vector)
   initColListCircle();

   // loop for y increasing box
   while(end[1] <= maxboxlength) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      // "next" box already at max?
      if(end[1] == maxboxlength) { // yes! => replace
         nextrowbox = new vector<fpixel<T> >(box->size());
         replaceRowInNewBox(*nextrowbox, *box, start[1], getRow(end[1] + 1));
      }
      else { // no! => grow
         nextrowbox = addVec( getRow(end[1] + 1) );
      }

      // first element of row!
      *outi = medianInCircle();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         addToColInList( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol(*listi);
         ++listi;

         *outi = medianInCircle();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         addToColInList( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInCircle();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInCircle();
         ++outi; ++pos[0];
      }
      // increase y
      ++pos[1]; ++end[1];
   }

   // loop for y constant box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(end[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      if(end[1] == inarray.length(2)) {
         delete box;
         box = nextrowbox;
         nextrowbox = delRow(start[1]+1);
      }
      else {
         vector<fpixel<T> > * const dummy = box;
         box = nextrowbox;
         nextrowbox = dummy;
         replaceRowInNewBox(*nextrowbox, *box, start[1]+1, getRow(end[1] + 1));
      }

      // first element of row!
      *outi = medianInCircle();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         replaceInCol( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol(*listi);
         ++listi;

         *outi = medianInCircle();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         replaceInCol( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInCircle();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInCircle();
         ++outi; ++pos[0];
      }
      ++pos[1]; ++end[1];
   }
   --end[1];

   // loop for y decreasing box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(pos[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      if(pos[1] != inarray.length(2)) {
         nextrowbox = delRow(start[1]+1);
      }

      // first element of row!
      *outi = medianInCircle();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         delFromColInList( start[1] );
         addCol(*listi);
         ++listi;

         *outi = medianInCircle();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         delFromColInList( start[1] );
         replaceColInBox(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInCircle();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];

         *outi = medianInCircle();
         ++outi; ++pos[0];
      }
      ++pos[1];
   }

   delete box; box = nextrowbox = NULL;

   done_=true;
   done_box=false;

   return outarray;
}

template<class T>
MArray<T, 2> td_median_stencil<T>::evaluateCircle_nan()
{
   initBox_nan();

   // first element!
   *outi = medianInCircle_nan();
   ++outi; ++pos[0]; ++end[0];

   // 1st loops to init the colvector list (which actually is a vector)
   initColListCircle_nan();

   // loop for y increasing box
   while(end[1] <= maxboxlength) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      // "next" box already at max?
      if(end[1] == maxboxlength) { // yes! => replace
         nextrowbox = new vector<fpixel<T> >(box->size());
         replaceRowInNewBox_nan(*nextrowbox, *box, start[1], getRow_nan(end[1] + 1));
      }
      else { // no! => grow
         nextrowbox = addVec_nan( getRow_nan(end[1] + 1) );
      }

      // first element of row!
      *outi = medianInCircle_nan();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         addToColInList_nan( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol_nan(*listi);
         ++listi;

         *outi = medianInCircle_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         addToColInList_nan( fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox_nan(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInCircle_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInCircle_nan();
         ++outi; ++pos[0];
      }
      // increase y
      ++pos[1]; ++end[1];
   }

   // loop for y constant box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(end[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      if(end[1] == inarray.length(2)) {
         delete box;
         box = nextrowbox;
         nextrowbox = delRow(start[1]+1);
      }
      else {
         vector<fpixel<T> > * const dummy = box;
         box = nextrowbox;
         nextrowbox = dummy;
         replaceRowInNewBox_nan(*nextrowbox, *box, start[1]+1, getRow_nan(end[1] + 1));
      }

      // first element of row!
      *outi = medianInCircle_nan();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         replaceInCol_nan( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         addCol_nan(*listi);
         ++listi;

         *outi = medianInCircle_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         replaceInCol_nan( *listi, start[1], fpixel<T>(inarray(end[0], end[1]), end[0], end[1]) );
         replaceColInBox_nan(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInCircle_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];
    
         *outi = medianInCircle_nan();
         ++outi; ++pos[0];
      }
      ++pos[1]; ++end[1];
   }
   --end[1];

   // loop for y decreasing box
   // start[1] is one behind until delCol loop,
   // i.e. has coordinate of "to be deleted" row
   while(pos[1] <= inarray.length(2)) {
    
      // reset x
      resetX();

      delete box;
      box = nextrowbox;
      if(pos[1] != inarray.length(2)) {
         nextrowbox = delRow(start[1]+1);
      }

      // first element of row!
      *outi = medianInCircle_nan();
      ++outi; ++pos[0]; ++end[0];

      // x increasing box
      while( end[0] <= maxboxlength ) {
         delFromColInList( start[1] );
         addCol_nan(*listi);
         ++listi;

         *outi = medianInCircle_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      // x constant box
      while(end[0] <= inarray.length(1)) {
         delFromColInList( start[1] );
         replaceColInBox_nan(start[0], *listi);
         ++start[0];
         ++listi;

         *outi = medianInCircle_nan();
         ++outi; ++pos[0]; ++end[0];
      }
      --end[0];
      ++start[1]; // now y-start is at actual corner
      // x decreasing box
      while(pos[0] <= inarray.length(1)) {
         delCol(start[0]);
         ++start[0];

         *outi = medianInCircle_nan();
         ++outi; ++pos[0];
      }
      ++pos[1];
   }

   delete box; box = nextrowbox = NULL;

   done_=true;
   done_box=false;

   return outarray;
}


// get col vector for box
template<class T>
vector<fpixel<T> > td_median_stencil<T>::getCol(const int x)
{
   vector<fpixel<T> > colvec(getCol_common(x));
   sort(colvec.begin(), colvec.end());
   return colvec;
}
template<class T>
vector<fpixel<T> > td_median_stencil<T>::getCol_nan(const int x)
{
   vector<fpixel<T> > colvec(getCol_common(x));
   sort(colvec.begin(), colvec.end(), pred);
   return colvec;
}
template<class T>
vector<fpixel<T> > td_median_stencil<T>::getCol_common(const int x)
{
   const MArray<T, 1> col( inarray(x, Range(start[1], end[1])) );
   typename MArray<T, 1>::const_iterator coli = col.begin();
   vector<fpixel<T> > colvec(getActLength(1));
   typename vector<fpixel<T> >::iterator vi = colvec.begin();
   for(int i = start[1]; i <= end[1]; ++i) {
      *vi = fpixel<T>(*coli, x, i);
      ++vi;
      ++coli;
   }
   return colvec;
}

// get row vector for box
template<class T>
vector<fpixel<T> > td_median_stencil<T>::getRow(const int y)
{
   vector<fpixel<T> > rowvec(getRow_common(y));
   sort(rowvec.begin(), rowvec.end());
   return rowvec;
}
template<class T>
vector<fpixel<T> > td_median_stencil<T>::getRow_nan(const int y)
{
   vector<fpixel<T> > rowvec(getRow_common(y));
   sort(rowvec.begin(), rowvec.end(), pred);
   return rowvec;
}
template<class T>
vector<fpixel<T> > td_median_stencil<T>::getRow_common(const int y)
{
   const MArray<T, 1> row( inarray(Range(start[0], end[0]), y) );
   typename MArray<T, 1>::const_iterator rowi = row.begin();
   vector<fpixel<T> > rowvec(getActLength(0));
   typename vector<fpixel<T> >::iterator vi = rowvec.begin();
   for(int i = start[0]; i <= end[0]; ++i) {
      *vi = fpixel<T>(*rowi, i, y);
      ++vi;
      ++rowi;
   }
   return rowvec;
}


// add vector to box
template<class T>
vector<fpixel<T> > * td_median_stencil<T>::addVec(const vector<fpixel<T> > & vec)
{
   const vector<fpixel<T> > * oldbox = box;
   vector<fpixel<T> > * newbox = new vector<fpixel<T> >( oldbox->size() + vec.size() );
   merge(oldbox->begin(), oldbox->end(),
         vec.begin(), vec.end(),
         newbox->begin());
   return newbox;
}
template<class T>
vector<fpixel<T> > * td_median_stencil<T>::addVec_nan( const vector<fpixel<T> > & vec)
{
   const vector<fpixel<T> > * oldbox = box;
   vector<fpixel<T> > * newbox = new vector<fpixel<T> >( oldbox->size() + vec.size() );
   merge(oldbox->begin(), oldbox->end(),
         vec.begin(), vec.end(),
         newbox->begin(),
         pred);
   return newbox;
}

template<class T>
void td_median_stencil<T>::addCol(const vector<fpixel<T> > & vec)
{
   vector<fpixel<T> > * oldbox = box;
   box = addVec( vec );
   delete oldbox;
}
template<class T>
void td_median_stencil<T>::addCol_nan(const vector<fpixel<T> > & vec)
{
   vector<fpixel<T> > * oldbox = box;
   box = addVec_nan( vec );
   delete oldbox;
}

// remove column from box
template<class T>
void td_median_stencil<T>::delCol(const int x)
{
   vector<fpixel<T> > * oldbox = box;
   box = new vector<fpixel<T> >( oldbox->size() - getActLength(1) );
   typename vector<fpixel<T> >::iterator oldboxi = oldbox->begin();
   typename vector<fpixel<T> >::iterator boxi = box->begin();
   // save surviving pixels to stripped box
   while(boxi != box->end()) {
      if( (*oldboxi).x != x) {
         *boxi = *oldboxi;
         ++boxi;
      }
      ++oldboxi;
   }
   delete oldbox;
}

// remove row from box
template<class T>
vector<fpixel<T> > * td_median_stencil<T>::delRow(const int y)
{
   const vector<fpixel<T> > * oldbox = box;
   vector<fpixel<T> > * newbox =
      new vector<fpixel<T> >( oldbox->size() - getActLength(0) );
   typename vector<fpixel<T> >::const_iterator oldboxi = oldbox->begin();
   typename vector<fpixel<T> >::iterator newboxi = newbox->begin();
   // save surviving pixels to stripped box
   while(newboxi != newbox->end()) {
      if( (*oldboxi).y != y) {
         *newboxi = *oldboxi;
         ++newboxi;
      }
      ++oldboxi;
   }
   return newbox;
}

// replace colvector at x with new vector in well known boxes
template<class T>
void td_median_stencil<T>::replaceColInBox(const int x,
					   const vector<fpixel<T> > & newvector)
{
   const size_t length = box->size();
   if(length != dummybox->size()) {
      delete dummybox;
      dummybox = new vector<fpixel<T> >(length);
   }
   replaceColInNewBox(*dummybox, *box, x, newvector);
   vector<fpixel<T> > * const holder = box;
   box = dummybox; dummybox = holder;
}
template<class T>
void td_median_stencil<T>::replaceColInBox_nan(const int x,
					       const vector<fpixel<T> > & newvector)
{
   const size_t length = box->size();
   if(length != dummybox->size()) {
      delete dummybox;
      dummybox = new vector<fpixel<T> >(length);
   }
   replaceColInNewBox_nan(*dummybox, *box, x, newvector);
   vector<fpixel<T> > * const holder = box;
   box = dummybox; dummybox = holder;
}

// replace (distributed) col-vector in box, but to a new box
template<class T>
inline void td_median_stencil<T>::replaceColInNewBox(vector<fpixel<T> > & newbox,
                                                     const vector<fpixel<T> > & oldbox,
                                                     const int x,
                                                     const vector<fpixel<T> > & newvector)
{
   replaceVecCopy<T, 1>(newbox.begin(), newbox.end(),
                        oldbox.begin(), oldbox.end(),
                        x,
                        newvector.begin(), newvector.end());
}
template<class T>
inline void td_median_stencil<T>::replaceColInNewBox_nan(vector<fpixel<T> > & newbox,
                                                         const vector<fpixel<T> > & oldbox,
                                                         const int x,
                                                         const vector<fpixel<T> > & newvector)
{
   replaceVecCopy<T, 1, less_fp_nan<T> >(newbox.begin(), newbox.end(),
                                         oldbox.begin(), oldbox.end(),
                                         x,
                                         newvector.begin(), newvector.end(),
                                         pred);
}

// replace (distributed) row-vector in box, but to a new box
template<class T>
inline void td_median_stencil<T>::replaceRowInNewBox(vector<fpixel<T> > & newbox,
                                                     const vector<fpixel<T> > & oldbox,
                                                     const int y,
                                                     const vector<fpixel<T> > & newvector)
{
   replaceVecCopy<T, 2>(newbox.begin(), newbox.end(),
                        oldbox.begin(), oldbox.end(),
                        y,
                        newvector.begin(), newvector.end());
}
template<class T>
inline void td_median_stencil<T>::replaceRowInNewBox_nan(vector<fpixel<T> > & newbox,
                                                         const vector<fpixel<T> > & oldbox,
                                                         const int y,
                                                         const vector<fpixel<T> > & newvector)
{
   replaceVecCopy<T, 2, less_fp_nan<T> >(newbox.begin(), newbox.end(),
                                         oldbox.begin(), oldbox.end(),
                                         y,
                                         newvector.begin(), newvector.end(),
                                         pred);
}

// add pixel to col vector in list
template<class T>
void td_median_stencil<T>::addToColInList(const fpixel<T> & newpixel)
{
   vector<fpixel<T> > newcol( (*listi).size() + 1);
   typename vector<fpixel<T> >::iterator newi = newcol.begin();
   typename vector<fpixel<T> >::const_iterator oldi = (*listi).begin();
   while( (oldi != (*listi).end()) && (*oldi < newpixel) ) {
      *newi = *oldi;
      ++newi; ++oldi;
   }
   *newi = newpixel;
   ++newi;
   while( newi != newcol.end() ){
      *newi = *oldi;    
      ++newi; ++oldi;
   }
   *listi = newcol;
}
template<class T>
void td_median_stencil<T>::addToColInList_nan(const fpixel<T> & newpixel)
{
   vector<fpixel<T> > newcol( (*listi).size() + 1);
   typename vector<fpixel<T> >::iterator newi = newcol.begin();
   typename vector<fpixel<T> >::const_iterator oldi = (*listi).begin();
   while( (oldi != (*listi).end()) && pred(*oldi, newpixel) ) {
      *newi = *oldi;
      ++newi; ++oldi;
   }
   *newi = newpixel;
   ++newi;
   while( newi != newcol.end() ){
      *newi = *oldi;    
      ++newi; ++oldi;
   }
   *listi = newcol;
}

// remove pixel from col vector in list
template<class T>
void td_median_stencil<T>::delFromColInList(const int y)
{
   vector<fpixel<T> > newcol( (*listi).size() - 1);
   typename vector<fpixel<T> >::iterator newi = newcol.begin();
   typename vector<fpixel<T> >::const_iterator oldi = (*listi).begin();
   while( (*oldi).y != y ) {
      *newi = *oldi;
      ++newi; ++oldi;
   }
   ++oldi;
   while( newi != newcol.end() ) {
      *newi = *oldi;    
      ++newi; ++oldi;
   }
   *listi = newcol;
}

// replace element in col-vector
template<class T>
inline void td_median_stencil<T>::replaceInCol(vector<fpixel<T> > & colvec,
                                               const int y,
                                               const fpixel<T> & newpixel)
{
   replaceInVec<T, 2>(colvec.begin(), colvec.end(), y, newpixel);
}
template<class T>
inline void td_median_stencil<T>::replaceInCol_nan(vector<fpixel<T> > & colvec,
                                                   const int y,
                                                   const fpixel<T> & newpixel)
{
   replaceInVec<T, 2, less_fp_nan<T> >(colvec.begin(), colvec.end(), y, newpixel, pred);
}


template<class T>
inline T td_median_stencil<T>::nthInBox(const int n)
{ return ((*box)[n-1]).value; }
      
template<class T>
inline T td_median_stencil<T>::medianInBox() {
   return medianInBox_common(box->size());
}
template<class T>
T td_median_stencil<T>::medianInBox_nan() {
   const size_t length = box->size() - count_nan();
   if(length == size_t(0)) return pred.nan_;
   return medianInBox_common(length);
}
template<class T>
T td_median_stencil<T>::medianInBox_common(const size_t length)
{
   const size_t hlen   = length / 2;
   const T med = ((*box)[hlen]).value;
   if( length % 2 ) return med;
   return (med + ((*box)[hlen - 1]).value) / T(2);
}

template<class T>
inline bool td_median_stencil<T>::isInCircle(const int x, const int y) {
   const int rx = x-pos[0]; const int ry = y-pos[1];
   return ( (rx*rx + ry*ry) <= horizon2 );
}

template<class T>
T td_median_stencil<T>::nthInCircle(const int n) {
   typename vector<fpixel<T> >::const_iterator boxi = box->begin();
   int i = 1;
   while( i < n ) {
      i += isInCircle( (*boxi).x, (*boxi).y );
      ++boxi;
   }
   while( !isInCircle((*boxi).x, (*boxi).y) ) ++boxi;
   return (*boxi).value;
}
      
template<class T>
T td_median_stencil<T>::medianInCircle() {
   vector<bool> isin(box->size());
   const int length = medianInCircle_init(isin.begin(), isin.end());
   return medianInCircle_common(isin.begin(), length);
}
template<class T>
T td_median_stencil<T>::medianInCircle_nan() {
   vector<bool> isin(box->size() - count_nan());
   const int length = medianInCircle_init(isin.begin(), isin.end());
   if(length == 0) return pred.nan_;
   return medianInCircle_common(isin.begin(), length);
}
template<class T>
int td_median_stencil<T>::medianInCircle_init(typename vector<bool>::iterator isini,
                                              const typename vector<bool>::iterator isinend)
{
   typename vector<fpixel<T> >::const_iterator boxi = box->begin();
   int count = 0;         
   while(isini != isinend)
   {
      *isini = isInCircle( (*boxi).x, (*boxi).y );
      count += *isini;
      ++boxi;
      ++isini;
   }
   return count;
}
template<class T>
T td_median_stencil<T>::medianInCircle_common(typename vector<bool>::const_iterator isini,
                                              const int length)
{
   typename vector<fpixel<T> >::const_iterator boxi = box->begin();
   const int hlen = length / 2;
   const bool isodd = length % 2;
   int count = *isini;
   while(count <= hlen)
   {
      ++isini;
      ++boxi;
      count += *isini;
   }
   const T med = (*boxi).value;
   if(isodd) return med;
   do
   {
      --isini;
      --boxi;
   }
   while( !(*isini) );
   return (med + (*boxi).value) / T(2);
}
      
template<class T>
int td_median_stencil<T>::count_nan() {
   int count = 0;
   typename vector<fpixel<T> >::const_reverse_iterator boxi = box->rbegin();
   const typename vector<fpixel<T> >::const_reverse_iterator boxrend = box->rend();
   while(boxi != boxrend) {
      if( (*boxi).value == pred.nan_ ) ++count;
      else { boxi = boxrend; --boxi;}
      ++boxi;
   }
   return count;
}


// replace (distributed) vector in old box by a new vector, but to a new box
// predicate is a coordinate match i of DIM
// old box = source with obsolete distributed vector
// new box = destination
// old vec = the obsolete distributed vector in old box
// new vec = the replacement for old vec
template<class T, int DIM>
void replaceVecCopy(const typename vector<fpixel<T> >::iterator newbegin,
		    const typename vector<fpixel<T> >::iterator newend,  
		    const typename vector<fpixel<T> >::const_iterator oldbegin,
		    const typename vector<fpixel<T> >::const_iterator oldend,  
		    const int i,
		    const typename vector<fpixel<T> >::const_iterator vecbegin,
		    const typename vector<fpixel<T> >::const_iterator vecend)
{
   typename vector<fpixel<T> >::const_iterator vec_i = vecbegin;
   typename vector<fpixel<T> >::const_iterator old_i = oldbegin;
   const typename vector<fpixel<T> >::const_iterator oldlast = oldend - 1;
   typename vector<fpixel<T> >::iterator new_i = newbegin;
  
   while( eq_dim_fp<T, DIM>()(*old_i, i ) ) ++old_i; // skip old vec values
   // save surviving pixels to new box and insert new vector
   while(new_i != newend) {
      if( *vec_i < *old_i ) { // use new vec or old box?
         *new_i = *vec_i; // use new vec
         ++new_i;
         ++vec_i;
         if(vec_i == vecend) { // no new vew value left?
            while(new_i != newend) { // fill up new box with old box values
               if( neq_dim_fp<T, DIM>()(*old_i, i ) ) { // skip old vec values
                  *new_i = *old_i;
                  ++new_i;
               }
               ++old_i;
            }
         }
      }
      else {
         *new_i = *old_i; // use old box
         ++new_i;
         ++old_i;
         if(old_i != oldend) {
            while( (old_i != oldlast) && eq_dim_fp<T, DIM>()(*old_i, i) ) ++old_i; // skip old vec values
            if(eq_dim_fp<T, DIM>()(*old_i, i)) ++old_i;
         }
         if(old_i == oldend) { // no old box value left?
            while(new_i != newend) { // fill up new box with new vec values
               *new_i = *vec_i;
               ++new_i;
               ++vec_i;
            }
         }
      }
   }
#ifdef __SMOOTH_DEBUG__
   // sanity check
   for(vec_i = newbegin; vec_i != newend-1; ++vec_i) {
      if( (*vec_i > *(vec_i+1)) || eq_dim_fp<T, DIM>()(*vec_i, i) ) {
         cerr << "Old box:" << endl;
         showvec<T>(oldbegin, oldend);
         cerr << "New vector:" << endl;
         showvec<T>(vecbegin, vecend);
         cerr << "New box:" << endl;
         showvec<T>(newbegin, newend);
         exit(1);
      }
   }
#endif
}

template<class T, int DIM, class P>
void replaceVecCopy(const typename vector<fpixel<T> >::iterator newbegin,
		    const typename vector<fpixel<T> >::iterator newend,  
		    const typename vector<fpixel<T> >::const_iterator oldbegin,
		    const typename vector<fpixel<T> >::const_iterator oldend,  
		    const int i,
		    const typename vector<fpixel<T> >::const_iterator vecbegin,
		    const typename vector<fpixel<T> >::const_iterator vecend,
		    const P & pred)
{
   typename vector<fpixel<T> >::const_iterator vec_i = vecbegin;
   typename vector<fpixel<T> >::const_iterator old_i = oldbegin;
   const typename vector<fpixel<T> >::const_iterator oldlast = oldend - 1;
   typename vector<fpixel<T> >::iterator new_i = newbegin;
  
   while( eq_dim_fp<T, DIM>()(*old_i, i ) ) ++old_i; // skip old vec values
   // save surviving pixels to new box and insert new vector
   while(new_i != newend) {
      if( pred(*vec_i, *old_i) ) { // use new vec or old box?
         *new_i = *vec_i; // use new vec
         ++new_i;
         ++vec_i;
         if(vec_i == vecend) { // no new vew value left?
            while(new_i != newend) { // fill up new box with old box values
               if( neq_dim_fp<T, DIM>()(*old_i, i ) ) { // skip old vec values
                  *new_i = *old_i;
                  ++new_i;
               }
               ++old_i;
            }
         }
      }
      else {
         *new_i = *old_i; // use old box
         ++new_i;
         ++old_i;
         if(old_i != oldend) {
            while( (old_i != oldlast) && eq_dim_fp<T, DIM>()(*old_i, i) ) ++old_i; // skip old vec values
            if(eq_dim_fp<T, DIM>()(*old_i, i)) ++old_i;
         }
         if(old_i == oldend) { // no old box value left?
            while(new_i != newend) { // fill up new box with new vec values
               *new_i = *vec_i;
               ++new_i;
               ++vec_i;
            }
         }
      }
   }
#ifdef __SMOOTH_DEBUG__
   // sanity check
   for(vec_i = newbegin; vec_i != newend-1; ++vec_i) {
      if( pred(*(vec_i+1), *vec_i) || eq_dim_fp<T, DIM>()(*vec_i, i) ) {
         cerr << "Old box:" << endl;
         showvec<T>(oldbegin, oldend);
         cerr << "New vector:" << endl;
         showvec<T>(vecbegin, vecend);
         cerr << "New box:" << endl;
         showvec<T>(newbegin, newend);
         exit(1);
      }
   }
#endif
}

// replace 1st fpixel with coordinate match in vector
template<class T, int DIM>
void replaceInVec(const typename vector<fpixel<T> >::iterator begin,
		  const typename vector<fpixel<T> >::iterator end,
		  const int i,
		  const fpixel<T> & newpixel)
{
   typename vector<fpixel<T> >::iterator vec_i = begin;

   // find element to drop
   while( neq_dim_fp<T, DIM>()(*vec_i, i) ) ++vec_i;

   if( *vec_i < newpixel ) {
      // insert new pixel here or above
      const typename vector<fpixel<T> >::const_iterator last = end - 1;
      if(vec_i != last) {
         typename vector<fpixel<T> >::const_iterator vecnext = vec_i;
         ++vecnext;
         while( ( *vecnext < newpixel ) &&
                (vecnext != last) ) {
            *vec_i = *vecnext;
            ++vec_i;
            ++vecnext;
         }
         if((*vecnext < newpixel) && (vecnext == last)) {
            *vec_i = *vecnext; ++vec_i;
         }
      }
   }
   else {
      // insert new pixel here or below
      if(vec_i != begin) {
         typename vector<fpixel<T> >::const_iterator vecprev = vec_i - 1;
         while( ( newpixel < *vecprev ) &&
                (vecprev != begin) ) {
            *vec_i = *vecprev;
            --vec_i;
            --vecprev;
         }
         if((newpixel < *vecprev) && (vecprev == begin)) {
            *vec_i = *vecprev; --vec_i;
         }
      }
   }
   *vec_i = newpixel;

#ifdef __SMOOTH_DEBUG__
   // sanity check
   for(vec_i = begin; vec_i != end-1; ++vec_i) {
      if(*vec_i > *(vec_i+1)) {
         cerr << "new pixel: " << newpixel.value << " " 
              << newpixel.x << " " << newpixel.y << endl;
         showvec<T>(begin, end);
         exit(1);
      }
   }
#endif
}

// replace 1st fpixel with coordinate match in vector
template<class T, int DIM, class P>
void replaceInVec(const typename vector<fpixel<T> >::iterator begin,
		  const typename vector<fpixel<T> >::iterator end,
		  const int i,
		  const fpixel<T> & newpixel,
		  const P & pred)
{
   typename vector<fpixel<T> >::iterator vec_i = begin;

   // find element to drop
   while( neq_dim_fp<T, DIM>()(*vec_i, i) ) ++vec_i;

   if( pred(*vec_i, newpixel) ) {
      // insert new pixel here or above
      const typename vector<fpixel<T> >::const_iterator last = end - 1;
      if(vec_i != last) {
         typename vector<fpixel<T> >::const_iterator vecnext = vec_i;
         ++vecnext;
         while( pred(*vecnext, newpixel) &&
                (vecnext != last) ) {
            *vec_i = *vecnext;
            ++vec_i;
            ++vecnext;
         }
         if(pred(*vecnext, newpixel) && (vecnext == last)) {
            *vec_i = *vecnext; ++vec_i;
         }
      }
   }
   else {
      // insert new pixel here or below
      if(vec_i != begin) {
         typename vector<fpixel<T> >::const_iterator vecprev = vec_i - 1;
         while( pred( newpixel, *vecprev ) &&
                (vecprev != begin) ) {
            *vec_i = *vecprev;
            --vec_i;
            --vecprev;
         }
         if(pred(newpixel, *vecprev) && (vecprev == begin)) {
            *vec_i = *vecprev; --vec_i;
         }
      }
   }
   *vec_i = newpixel;

#ifdef __SMOOTH_DEBUG__
   // sanity check
   for(vec_i = begin; vec_i != end-1; ++vec_i) {
      if(pred( *(vec_i+1), *vec_i) ) {
         cerr << "new pixel: " << newpixel.value << " " 
              << newpixel.x << " " << newpixel.y << endl;
         showvec<T>(begin, end);
         exit(1);
      }
   }
#endif
}


template<class T>
MArray<T, 2> stencil<T>::getCircle(const int radius) const
{
   const int horizon2 = radius * radius + 1;
   MArray<T, 2> circle(Range(-radius, radius), Range(-radius, radius));
   circle = cast<T>()( (pow2(indexPosInt(circle, 1)) + pow2(indexPosInt(circle, 2))) <= horizon2 );
   return circle;
}

template<class Expr, int N>
double getSEM(const typename ExprNode<Expr,N>::value_type & mean, ExprNode<Expr,N> e)
{
   const int n = e.shape()->nelements();
   variance_reduction<typename ExprNode<Expr,N>::value_type> R( n, mean);
   eval_full_reduction( e, R );
   return  sqrt( R.result() / double( n ) );   
}

template<class T, int N>
double getSEM(const T & mean, const MArray<T, N> & inarray)
{
   ExprNode<typename MArray<T, 2>::const_iterator, 2> e( inarray.begin() );
   return getSEM(mean, e);
}

template<class Expr, int N>
double getSEM(const typename ExprNode<Expr,N>::value_type & mean,
              ExprNode<Expr,N> e,
              const typename ExprNode<Expr,N>::value_type & nan)
{
   const double n = double(e.shape()->nelements());
   variance_reduction_nan<typename ExprNode<Expr,N>::value_type> R( mean, nan );
   eval_full_reduction( e, R );
   return sqrt( R.result() / n );
}

template<class T, int N>
double getSEM(const T & mean, const MArray<T, N> & inarray, const T & nan)
{
   ExprNode<typename MArray<T, 2>::const_iterator, 2> e( inarray.begin() );
   return getSEM(mean, e, nan);
}

template<class T>
MArray<T, 2> calculateSEM(const MArray<T, 2> & meanarray, const MArray<T, 2> & inarray,
                          const int radius, const bool circular, const bool hasnan,
                          const T nan = T(0))
{
   MArray<float, 2> error(inarray.shape());
   MArray<float, 2>::IndexIterator ii = error.indexBegin();

   if(!circular) { // boxy
      if(!hasnan)
         while( ! ii.done() ) {
            error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                          inarray( Range( max((ii(1) - radius), inarray.minIndex(1)),
                                                          min((ii(1) + radius), inarray.maxIndex(1)) ),
                                                   Range( max((ii(2) - radius), inarray.minIndex(2)),
                                                          min((ii(2) + radius), inarray.maxIndex(2)) )) );
            ++ii;
         }
      else {
         while( ! ii.done() ) {
            error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                          inarray( Range( max((ii(1) - radius), inarray.minIndex(1)),
                                                          min((ii(1) + radius), inarray.maxIndex(1)) ),
                                                   Range( max((ii(2) - radius), inarray.minIndex(2)),
                                                          min((ii(2) + radius), inarray.maxIndex(2)) )),
                                          nan);
            ++ii;
         }         
      }
   }
   else {
      if(!hasnan) {
         stencil<T> sten; // provide an instance for stencil methods
         const MArray<T, 2> cstencil( sten.getCircle(radius) );
         while( ! ii.done() ) {
            MArray<T, 2> window_( inarray( Range( max((ii(1) - radius), inarray.minIndex(1)),
                                                  min((ii(1) + radius), inarray.maxIndex(1)) ),
                                           Range( max((ii(2) - radius), inarray.minIndex(2)),
                                                  min((ii(2) + radius), inarray.maxIndex(2)) )) );
	    if(window_.nelements() < cstencil.nelements()) {
               // which part?
               if(window_.length(1) < cstencil.length(1)) {
                  // subarray of x
                  if(window_.length(2) < cstencil.length(2)) {
                     // and of y ... but which parts ...
                     if( ii(1) <= radius ) {
                        // ah ... this part of x ... and of y?
                        if( ii(2) <= radius ) {
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         window_ * cstencil(Range(0 - ii(1), cstencil.maxIndex(1)),
                                                                            Range(0 - ii(2), cstencil.maxIndex(2))),
                                                         nan);
                        }
                        else {
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         window_ * cstencil(Range(0 - ii(1), cstencil.maxIndex(1)),
                                                                            Range(cstencil.minIndex(2), 
                                                                                  window_.length(2) - (radius + 1))),
                                                         nan);
                        }
                     }
                     else {
                        // other part of x ... and y?
                        if( ii(2) <= radius )
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         window_ * cstencil(Range(cstencil.minIndex(1),
                                                                                  window_.length(1) - (radius + 1)),
                                                                            Range(0 - ii(2), cstencil.maxIndex(2))),
                                                         nan);
                        else {
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         window_ * cstencil(Range(cstencil.minIndex(1),
                                                                                  window_.length(1) - (radius + 1)),
                                                                            Range(cstencil.minIndex(2), 
                                                                                  window_.length(2) - (radius + 1))),
                                                         nan);
                        }
                     }
                  }
                  // but no subarray of y
                  else {
                     // so which part of x?
                     if( ii(1) <= radius)
                        error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                      window_ * cstencil(Range(0 - ii(1), cstencil.maxIndex(1)),
                                                                         Range::all()),
                                                      nan);
                     else // ok other part
                        error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                      window_ * cstencil(Range(cstencil.minIndex(1),
                                                                               window_.length(1) - (radius + 1)),
                                                                         Range::all()),
                                                      nan);
                  }
               }
               // hmmm ... need complete x, so only part of y
               else {
                  // but which part of y?
                  if( ii(2) <= radius)
                     error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                   window_ * cstencil(Range::all(),
                                                                      Range(0 - ii(2), cstencil.maxIndex(2))),
                                                   nan);
                  else // ok other part
                     error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                   window_ * cstencil(Range::all(),
                                                                      Range(cstencil.minIndex(2),
                                                                            window_.length(2) - (radius + 1))),
                                                   nan );
               }
	    }
	    // no ... need complete stencil
	    // finally the "usual" case no border issues ... so
	    else
               error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                             window_ * cstencil, nan);
	    ++ii;
         }
      }
      else {
         stencil<bool> sten; // provide an instance for stencil methods
         const MArray<bool, 2> cstencil( sten.getCircle(radius) );
         MArray<T, 2> porthole_;
         while( ! ii.done() ) {
            MArray<T, 2> window_( inarray( Range( max((ii(1) - radius), inarray.minIndex(1)),
                                                  min((ii(1) + radius), inarray.maxIndex(1)) ),
                                           Range( max((ii(2) - radius), inarray.minIndex(2)),
                                                  min((ii(2) + radius), inarray.maxIndex(2)) )) );
	    if(window_.nelements() < cstencil.nelements()) {
               // which part?
               if(window_.length(1) < cstencil.length(1)) {
                  // subarray of x
                  if(window_.length(2) < cstencil.length(2)) {
                     // and of y ... but which parts ...
                     if( ii(1) <= radius ) {
                        // ah ... this part of x ... and of y?
                        if( ii(2) <= radius ) {
                           porthole_ = merge(cstencil(Range(0 - ii(1), cstencil.maxIndex(1)),
                                                      Range(0 - ii(2), cstencil.maxIndex(2))),
                                             window_, nan);
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         porthole_ ,
                                                         nan);
                        }
                        else {
                           porthole_ = merge(cstencil(Range(0 - ii(1), cstencil.maxIndex(1)),
                                                      Range(cstencil.minIndex(2), 
                                                            window_.length(2) - (radius + 1))),
                                             window_, nan);
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         porthole_ ,
                                                         nan);
                        }
                     }
                     else {
                        // other part of x ... and y?
                        if( ii(2) <= radius ) {
                           porthole_ = merge(cstencil(Range(cstencil.minIndex(1),
                                                            window_.length(1) - (radius + 1)),
                                                      Range(0 - ii(2), cstencil.maxIndex(2))),
                                             window_, nan);
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         porthole_ ,
                                                         nan);
                        }
                        else {
                           porthole_ = merge(cstencil(Range(cstencil.minIndex(1),
                                                            window_.length(1) - (radius + 1)),
                                                      Range(cstencil.minIndex(2), 
                                                            window_.length(2) - (radius + 1))),
                                             window_, nan);
                           error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                         porthole_ ,
                                                         nan);
                        }
                     }
                  }
                  // but no subarray of y
                  else {
                     // so which part of x?
                     if( ii(1) <= radius) {
                        porthole_ = merge(cstencil(Range(0 - ii(1), cstencil.maxIndex(1)),
                                                   Range::all()),
                                          window_, nan);
                        error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                      porthole_ ,
                                                      nan);
                     }
                     else {// ok other part
                        porthole_ = merge(cstencil(Range(cstencil.minIndex(1),
                                                         window_.length(1) - (radius + 1)),
                                                   Range::all()),
                                          window_, nan);
                        error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                      porthole_ ,
                                                      nan);
                        }
                  }
               }
               // hmmm ... need complete x, so only part of y
               else {
                  // but which part of y?
                  if( ii(2) <= radius)
                  {
                        porthole_ = merge(cstencil(Range::all(),
                                                   Range(0 - ii(2), cstencil.maxIndex(2))),
                                          window_, nan);
                        error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                      porthole_ ,
                                                      nan);
                  }
                  else { // ok other part
                        porthole_ = merge(cstencil(Range::all(),
                                                   Range(cstencil.minIndex(2),
                                                         window_.length(2) - (radius + 1))),
                                          window_, nan);
                     error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                                   porthole_ ,
                                                   nan );
                  }
               }
	    }
	    // no ... need complete stencil
	    // finally the "usual" case no border issues ... so
	    else {
               porthole_ = merge( cstencil,
                                  window_, nan);
               error(ii(1), ii(2)) = getSEM( meanarray(ii(1), ii(2)),
                                             porthole_ , nan);
            }
	    ++ii;
         }      
      }
   }
   
   return error;
}

#endif
