/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: statfits.cpp 157 2014-04-29 15:43:00Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT float

#include "fitstools.h"


// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) :
         region(dimension), verbose(1) { };
      bool print_header;
      bool print_filename;
      FT nan_value;
      bool nan;
      bool count;
      bool minimum;
      bool maximum;
      bool median;
      bool average;
      bool sigma;
      bool twist;
      double kappa;
      bool covariance;
      Region region; // region class defined in util/option.h
      int donut[4];
      string usekey;
      int precision;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

// container for pixel coordinates
class psf
{
   public:
      psf(const int inx, const int iny) :
         x(inx), y(iny),
         background(0.0), background_rms(0.0), flux(0.0), 
         cx(0.0), cy(0.0), mx(0.0), my(0.0), mxy(0.0),
         a4(0.0), a5(0.0), a6(0.0), ok(false), invalid(false) { };
      //  fpixel(const fpixel<T> & in) :
      //    value(in.value), x(in.x), y(in.y) { };
      psf() { };
      int x;
      int y;
      double background, background_rms;
      double flux;
      double cx, cy;
      double mx, my, mxy;
      double a4, a5, a6;
      bool ok;
      bool invalid;
      void tocout(const bool header = false);
};   

// help functions to evaluate PSF -> start
template<class T>
class stencil{
   public:
      ltl::MArray<T, 2> getCircle(const int radius) const;
      ltl::MArray<T, 2> getRing(const int outerradius, const int innerradius) const;
};

template<class T>
ltl::MArray<T, 2> stencil<T>::getCircle(const int radius) const
{
   const int horizon2 = radius * radius + 1;
   ltl::MArray<T, 2> circle(ltl::Range(-radius, radius), ltl::Range(-radius, radius));
   circle = ( (pow2(indexPosInt(circle, 1)) + pow2(indexPosInt(circle, 2))) <= horizon2 );
   return circle;
}

template<class T>
ltl::MArray<T, 2> stencil<T>::getRing(const int outerradius, const int innerradius) const
{
   const int outerhorizon2 = outerradius * outerradius + 1;
   const int innerhorizon2 = innerradius * innerradius + 1;
   ltl::MArray<T, 2> ring(ltl::Range(-outerradius, outerradius), ltl::Range(-outerradius, outerradius));
   ring = ( (pow2(indexPosInt(ring, 1)) + pow2(indexPosInt(ring, 2))) <= outerhorizon2 ) &&
      ( (pow2(indexPosInt(ring, 1)) + pow2(indexPosInt(ring, 2))) >= innerhorizon2 );
   return ring;
}

psf eval_psf(FitsIn& fitsfile,
             const int x_in, const int y_in,
             const int sky_aperture, const int star_aperture,
             const double kappa = 2.0, const int verbose = 0)
{
   psf star(x_in, y_in);
   // check if aperture is within array
   if( (x_in - sky_aperture < 1) || (y_in - sky_aperture < 1) || 
       (x_in + sky_aperture > fitsfile.getNaxis(1)) || (y_in + sky_aperture > fitsfile.getNaxis(2)) ){
      star.invalid = true;
      return star;
   }
   stencil<bool> sten;
   util::Region region(2);
   const int aperture = sky_aperture ? sky_aperture : star_aperture;
   region.setRange(1, x_in - aperture, x_in + aperture);
   region.setRange(2, y_in - aperture, y_in + aperture);
   fitsfile.setRegion(region);
   ltl::MArray<float, 2> a_in;
   fitsfile >> a_in;
   // get local background
   double mean, sigma;
   if(sky_aperture)
   {
      //ltl::Range xrange(x_in - sky_aperture, x_in + sky_aperture);
      //ltl::Range yrange(y_in - sky_aperture, y_in + sky_aperture);
      //ltl::MArray<float, 2> a = ltl::cast<float>()( a_in(xrange, yrange) );
      const int remaining = ltl::kappa_median_average(merge(sten.getRing(sky_aperture, star_aperture), a_in, 0.0f), kappa, 0.0f,
                                                      &mean, &sigma);
      //cerr << "clipping for background left: " << remaining << endl;
   }
   else{ mean = 0.0; sigma = 1.0; }
   //cerr << mean << '\t' << sigma << endl;
   star.background = mean;
   star.background_rms = sigma;
   // subtract background and clip "noise"
   {
      //ltl::Range xrange(x_in - star_aperture, x_in + star_aperture);
      //ltl::Range yrange(y_in - star_aperture, y_in + star_aperture);
      //ltl::MArray<float, 2> a = ltl::cast<float>()( a_in(xrange, yrange) );
      const int apdiff = sky_aperture - star_aperture;
      ltl::MArray<float, 2> a( (sky_aperture ?
                                a_in( Range(apdiff+1, a_in.length(1)-apdiff), Range(apdiff+1, a_in.length(2)-apdiff) ) :
                                a_in) );
      a -= mean;
      // only take "real" counts into account
      const float limit = float(kappa * sigma);
      a = merge( ((limit == 0.0f)||(a > limit)) && sten.getCircle(star_aperture), a, 0.0f);
      const double i0 = sum(a);
      star.flux = i0;
      if( i0 < 4.0 * kappa * (sigma * sigma) ){
         star.ok = false;
         cerr << "PSFeval: star too dim, flux < " << (4.0 * kappa * (sigma * sigma));
      }
      else{
         star.ok = true;
         const double x0 = sum(indexPosFlt(a,1) * a) / i0;
         star.cx = x0 + double(x_in - star_aperture - 1);
         const double y0 = sum(indexPosFlt(a,2) * a) / i0;
         star.cy = y0 + double(y_in - star_aperture - 1);
         star.mx = sum( pow2(indexPosFlt(a,1)-x0) * a) / i0;
         star.my = sum( pow2(indexPosFlt(a,2)-y0) * a) / i0;
         star.mxy = sum( (indexPosFlt(a,1)-x0) * (indexPosFlt(a,2)-y0) * a) / i0;
         star.a4 = sqrt((star.mx + star.my)/2.0);
         const double mxmy_14 = sqrt(sqrt(star.mx * star.my));
         star.a5 = star.mxy / mxmy_14 ;
         star.a6 = (star.mx - star.my) / (mxmy_14 * 2.0);
      }
   }
   return star;
};
// help functions to evaluate PSF -> end

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define(s) to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if( flagVal.verbose > 1 )
      cerr << flags << endl;
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if( (files.size() < 1) ||
       !(flagVal.print_header || flagVal.print_filename ||
         (flagVal.usekey.length() > 0) ||
         flagVal.median || flagVal.average || flagVal.count) )
      print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   cout.precision(flagVal.precision);

   if(flagVal.print_header)
   {
      cout << "#";
      if(flagVal.print_filename)
         cout << "filename\t";
      if(flagVal.usekey.length() > 0)
         cout << flagVal.usekey << '\t';
      if(flagVal.region.getStart(1))
         cout << "region\t";
      if(flagVal.count)
      {
         cout << "N";
         if(flagVal.nan)
            cout << "!=NAN";
         cout << '\t';
      }
      if(flagVal.minimum) // minimum
         cout << "min." << '\t';
      if(flagVal.maximum) // maximum
         cout << "max." << '\t';
      if(flagVal.median)
      {
         if(flagVal.kappa)
         {
            cout << "kappa-";
            if(flagVal.twist) cout << "average-";
            else cout << "sigma-";
            cout << "clipped_";
         }
         cout << "median\t";
         if(flagVal.sigma) cout << "sigma+SEM\t";
         else cout << "SEM\t";
      }
      if(flagVal.average)
      {
         if(flagVal.kappa)
         {
            cout << "kappa-";
            if(flagVal.twist) cout << "median-";
            else cout << "sigma-";
            cout << "clipped_";
         }
         cout << "average\t";
         if(flagVal.sigma) cout << "sigma+SEM\t";
         else cout << "SEM\t";
      }
      if(flagVal.covariance)
      {
         cout << "variance\tcovariance\t";
      }
      if(flagVal.donut[2])
      {
         psf startest; startest.tocout(true);
      }   
      cout << endl;
   }
      
   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{

         string filename = *file_i;

         // array for input
         MArray<FT, dimension> a;

         // open IO-file
         FitsIn infile(filename,
                       flagVal.quiet_please, flagVal.ignore_header);

         // print filename
         if(flagVal.print_filename)
            cout << filename << '\t';
         
         // get value of FITS key
         if( flagVal.usekey.length() > 0 )
         {
            try
            {               
               cout << infile.getFloat( infile.adjustKeyword(flagVal.usekey) )
                    << '\t';
            }
            catch(FitsException e)
            {
               try
               {               
                  cout << infile.getString( infile.adjustKeyword(flagVal.usekey) )
                       << '\t';
               }
               catch(FitsException e)
               {
                  cout << 0 << '\t';
               }
            }
         }
         
         // print region
         if(flagVal.region.getStart(1))
            cout << flagVal.region.toString() << '\t';

         if(flagVal.count && (!flagVal.nan))
            cout << (infile.getDataLength() / infile.getBytpix()) << '\t';

         // get median or/and average of region
         if( flagVal.minimum || flagVal.maximum ||
             flagVal.median || flagVal.average ||
             (flagVal.count && flagVal.nan) )
         {
            // read data from file
            ft.selectRegionFile(infile, flagVal.region);
            if((flagVal.region.getLength(1) == 1 || flagVal.region.getLength(2) == 1) && flagVal.region.getLength() > 1)
            { // region is a slice of file
               MArray<FT, 1> b;
               if(flagVal.verbose > 0)
                  cerr << "Reading " << infile.getFilename() << endl;
               // build MArray
               infile >> b;
               infile.freeData();
               a.makeReference(MArray<FT, DIMENSION>( flagVal.region.getLength(1) > flagVal.region.getLength(2) ?
                                                      flagVal.region.getLength(1) : flagVal.region.getLength(2), 1));
               a(Range::all(), 1) = b;
            }
            else // region has same dimension as file
               ft.loadImage(infile, a);

            if(flagVal.count && flagVal.nan)
               cout << count(a != flagVal.nan_value) << '\t';

            if(flagVal.nan)
            {
               if(flagVal.minimum) // minimum
                  cout << min(a, flagVal.nan_value) << '\t';
               if(flagVal.maximum) // minimum
                  cout << max(a, flagVal.nan_value) << '\t';
            }
            else
            {
               if(flagVal.minimum) // minimum
                  cout << min(a) << '\t';
               if(flagVal.maximum) // minimum
                  cout << max(a) << '\t';
            }

            if(flagVal.median) // median
            {
               double median, sigma;
               ft.medianArray(median, sigma, a, flagVal.kappa,
                              flagVal.sigma, flagVal.twist,
                              flagVal.nan, flagVal.nan_value);
               cout << median << '\t' << sigma << '\t';
            }
            
            if(flagVal.average) // average
            {
               double average, sigma;
               ft.averageArray(average, sigma, a, flagVal.kappa,
                               flagVal.sigma, flagVal.twist,
                               flagVal.nan, flagVal.nan_value);
               cout << average << '\t' << sigma << '\t';
            }

            if(flagVal.covariance) // covariance
            {
               double davg;
               const double var = variance(a, &davg);
               const FT avg = FT(davg);
               //a -= avg;
               //cerr << "test: " << average(a) << '\t';
               Range xrange(2, a.length(1)-1);
               Range yrange(2, a.length(2)-1);
               double varac = 0.0;
               for(int m = -1; m <= 1; ++m)
                  for(int n = -1; n <= 1; ++n)
                     varac += sum( a(xrange, yrange) * a(xrange + m, yrange + n) );
               varac /= double(xrange.length() * yrange.length() - 1);
               cout << var << '\t' << varac << '\t' ;
            }

         }
         if(flagVal.donut[2])
         {
            psf startest = eval_psf(infile, flagVal.donut[0], flagVal.donut[1], flagVal.donut[3], flagVal.donut[2],
                                    flagVal.kappa, flagVal.verbose);
            startest.tocout();
         }

         cout << endl;

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: statfits [-FLAGS [option]] FILES\n") +
      string("print statistics of FILES on stdout\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2013 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "ignore nan for statistics",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new BoolOption("print_header", "FALSE", "print headline",
                                     'p', &flagVal.print_header) );
      flags.addOption(new BoolOption("count", "FALSE", "count value NOT nan",
                                     'c', &flagVal.count) );
      flags.addOption(new BoolOption("print_filename", "FALSE", "print filename for 1st column",
                                     'f', &flagVal.print_filename) );
      flags.addOption(new BoolOption("average", "FALSE",
                                     "print (clipped) average (of region)",
                                     'a', &flagVal.average ) );
      flags.addOption(new BoolOption("median", "FALSE",
                                     "print (clipped) median (of region)",
                                     'm', &flagVal.median ) );
      flags.addOption(new BoolOption("minimum", "FALSE",
                                     "print minimum (of region)",
                                     0, &flagVal.minimum ) );
      flags.addOption(new BoolOption("maximum", "FALSE",
                                     "print maximum (of region)",
                                     0, &flagVal.maximum ) );
      flags.addOption(new DoubleOption("kappa", "0.0",
                                      "for kappa sigma clipping, default: no clipping",
                                      'k', &flagVal.kappa) );
      flags.addOption(new BoolOption("sigma", "FALSE",
                                     "print (kappa corrected) sigma; default is printing just (kappa corrected) standard error of average/median",
                                     's', &flagVal.sigma ) );
      flags.addOption(new BoolOption("twist", "FALSE",
                                     "if average clip respective to median, if median clip respective to average",
                                     't', &flagVal.twist ) );
      flags.addOption(new BoolOption("covariance", "FALSE",
                                     "print variance and cross covariance of closest neighbors (of region)",
                                     0, &flagVal.covariance ) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "use x_start:x_end,y_start:y_end region to determine statistics",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new IntArrayOption("moments", "0,0,0,0",
                                         "use x_center,y_center,fit_aperture_radius,sky_aperture_radius to evaluate up to 2nd order moments within aperture",
                                         0, 4, flagVal.donut ) );
      flags.addOption(new StringOption("use_key", "", "print value of FITS keyword",
                                       'u', &flagVal.usekey) );
      flags.addOption(new IntOption("precision", "6",
                                    "set precision for numbers",
                                    0, &flagVal.precision ) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "0",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}
  
void psf::tocout(const bool header)
{
   if(header){
      cout << "x \ty \tflux \tbg \tbg_rms \tmx \tmy \tmxy \ta4 \ta5 \ta6 \t";
   } else{
      if(invalid)
         for(int i = 0; i<11; ++i) cout << "invalid \t";
      else if(ok)
         cout << cx << " \t" << cy << " \t"
              << flux << " \t" << background << " \t" << background_rms << " \t"
              << mx << " \t" << my << " \t" << mxy << " \t"
              << a4 << " \t" << a5 << " \t" << a6 << " \t";
      else{
         cout << "dim \tdim \t";
         cout << flux << " \t" << background << " \t" << background_rms << " \t";
         for(int i = 0; i<6; ++i) cout << "dim \t";
      }
   }
}      


