/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: subtractfits.cpp 174 2018-04-03 11:59:45Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT float

#include "fitstools.h"

// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) :
         region(dimension), of_region(dimension), perrow(1), per2row(2), percol(1), per2col(2), verbose(1) { };
      FT subtractor[2]; // [0] for subtractor, [1] for error
      string subtractor_file;
      FT nan_value;
      bool nan;
      bool average;
      bool median;
      bool sigma;
      bool twist;
      double kappa;
      Region region; // region class defined in util/option.h
      Region of_region; // region class defined in util/option.h
      Region perrow;
      Region per2row;
      Region percol;
      Region per2col;
      bool inverse;
      string prefix;
      string usekey;
      string useerror;
      string fitskey;
      string fitsfilekey;
      bool zero;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define(s) to get other versions

   // class for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);
   const bool is_perrow = flagVal.perrow.getStart(1) || flagVal.per2row.getStart(1);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instances for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);
   FitsTools<FT, 1> ftone(flagVal.verbose, flagVal.ignore_header);

   // for file subtraction: readfile
   MArray<FT, dimension> b, e_b, rb, e_rb;
   if( flagVal.subtractor_file.length() > 0 )
   {
      const bool have_error =
         ft.loadImageError2( flagVal.subtractor_file, b, e_b );
      // get region to operate on, default is whole array
      ft.referenceRegion(rb, b, flagVal.region);
      if(have_error)
         ft.referenceRegion(e_rb, e_b, flagVal.region);
   }
   
   flagVal.subtractor[1] *= flagVal.subtractor[1]; // square constant error

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<FT, dimension> a, e_a, ra, e_ra;
	 MArray<FT, 1> e_line;

         // subtractor for this run
         FT s = flagVal.subtractor[0];
         FT e_s = flagVal.subtractor[1];

         // open IO-file(s)
         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         ft.loadImage(infile, a);
         // get region to operate on, default is whole array
         ft.referenceRegion(ra, a, flagVal.region);
         
         const string outfilename =
            ft.prependFilename(*file_i, flagVal.prefix);
         FitsOut outfile(outfilename,
                         infile, flagVal.quiet_please);
         {
            string to_history = string("subtracted");
            if(flagVal.region.getStart(1))
               to_history += string(" of region ") + flagVal.region.toString();
            to_history +=string(":");
            if(flagVal.verbose > 0)
               cerr << to_history << endl;
            outfile.addHistory(to_history);
         }


         // get subtractors apart from file

         if( s || e_s ) // get constant
         {
            char buf[80];
            sprintf(buf, "constant: %g with error %g", s, sqrt(e_s));
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }

         // get value of FITS key
         if( ( flagVal.usekey.length() > 0 ) || ( flagVal.useerror.length() > 0 ) )
         {
            double holder = 0.0, errorholder = 0.0;
            char buf[128];
            strcpy(buf,"values of FITS keys: ");
            if( flagVal.usekey.length() > 0 )
            {
               holder = infile.getFloat( infile.adjustKeyword(flagVal.usekey) );
               s += FT(holder);
               sprintf(buf + strlen(buf), "%s = %g",
                       flagVal.usekey.c_str(), holder);
               if( flagVal.useerror.length() > 0 ) strcat(buf, ", with ");
            }
            if( flagVal.useerror.length() > 0 )
            {
               errorholder =
                  infile.getFloat( infile.adjustKeyword(flagVal.useerror) );
               e_s += FT(errorholder * errorholder);
               sprintf(buf + strlen(buf), "error %s = %g",
                       flagVal.useerror.c_str(), errorholder);
            }
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }
         
         // get median or/and average of region
         if(flagVal.median || flagVal.average)
         {
	    string result = "";
	    // average / median per line / col cases
	    if(is_perrow || flagVal.percol.getStart(1) || flagVal.per2col.getStart(1))
	    { // is_perrow defined after reading flags
	       if( !is_perrow ) // turn col case into row case
		  a.transposeSelf();
	       const bool is_prepost = is_perrow ? flagVal.per2row.getStart(1) : flagVal.per2col.getStart(1);
	       // unify all cases 
	       const int x1 = is_perrow ?
		  (is_prepost ? flagVal.per2row.getStart(1) : flagVal.perrow.getStart(1)) :
		  (is_prepost ? flagVal.per2col.getStart(1) : flagVal.percol.getStart(1));
	       const int x2 = is_perrow ?
		  (is_prepost ? flagVal.per2row.getEnd(1) : flagVal.perrow.getEnd(1)) :
		  (is_prepost ? flagVal.per2col.getEnd(1) : flagVal.percol.getEnd(1));
	       const int x3 = is_perrow ?
		  (is_prepost ? flagVal.per2row.getStart(2) : 0) :
		  (is_prepost ? flagVal.per2col.getStart(2) : 0);
	       const int x4 = is_perrow ?
		  (is_prepost ? flagVal.per2row.getEnd(2) : 0) :
		  (is_prepost ? flagVal.per2col.getEnd(2) : 0);
	       
	       MArray<FT, 1>line;
	       MArray<FT, 1> m;
	       e_line.makeReference(MArray<FT, 1>(a.length(2)));
	       e_line = FT(0.0);
	       for(int y = 1; y <= a.length(2); ++y)
	       { // for every row
		  line.makeReference(a(Range::all(), y));
		  if(is_prepost){
		     m.makeReference(MArray<FT, 1>(x2 - x1 + 1 + x4 - x3 +1));
		     m(Range(x1, x2)) = line(Range(x1, x2));
		     m(Range(x2+1, m.nelements())) = line(Range(x3, x4));
		  }
		  else m.makeReference( line(Range(x1, x2)) );
		  double med = 0.0, sigma = 0.0;
		  if(flagVal.average) // average
		     ftone.averageArray(med, sigma, m, flagVal.kappa,
					flagVal.sigma, flagVal.twist,
					flagVal.nan, flagVal.nan_value);
		  else if(flagVal.median) // median
		     ftone.medianArray(med, sigma, m, flagVal.kappa,
				       flagVal.sigma, flagVal.twist,
				       flagVal.nan, flagVal.nan_value);
		  line   -= FT(med); // subtract from row / col
		  e_line(y) = FT(sigma * sigma); // add to error
		  
	       }
	       if( !is_perrow )
		  a.transposeSelf(); // if col case reverse back
	       result += string("per ") +
		  (is_perrow ? string("row ") : string("column ")) +
		  (flagVal.median ? string("median") : string("average")) + " of region" +
		  (is_prepost ?
		   (string("s ") + (is_perrow ? flagVal.per2row.toString() : flagVal.per2col.toString())) :
		   (string(" ") + (is_perrow ? flagVal.perrow.toString() : flagVal.percol.toString()) ));
	    }
	    else{ // deal with simple average / median of region
	       MArray<FT, dimension> m;
	       ft.referenceRegion(m, a, flagVal.of_region);
	       
	       if(flagVal.median) // median
	       {
		  double median, sigma;
		  result += ft.medianArray(median, sigma, m, flagVal.kappa,
					   flagVal.sigma, flagVal.twist,
					   flagVal.nan, flagVal.nan_value);
		  s   += FT(median); // add to constant
		  e_s += FT(sigma * sigma); // add to error
	       }
	       
	       if(flagVal.average) // average
	       {
		  double average, sigma;
		  result += ft.averageArray(average, sigma, m, flagVal.kappa,
					    flagVal.sigma, flagVal.twist,
					    flagVal.nan, flagVal.nan_value);
		  s   += FT(average); // add to constant
		  e_s += FT(sigma * sigma); // add to error
	       }
	       
	       if(flagVal.of_region.getStart(1))
		  result += string(" of region ") + flagVal.of_region.toString();
	    }
	    if(flagVal.verbose > 0)
	       cerr << result << endl;
	    outfile.addHistory(result);
	 }
         
         // subtract file?
         if( b.isAllocated() )
         {
            const string buf = string("file: ") + flagVal.subtractor_file;
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }

         // calculate difference file
         ft.subtractArrayConstant(ra, rb, s, flagVal.nan, flagVal.nan_value);

         // invert result?
         if(flagVal.inverse)
         {
            if( flagVal.nan ) // with nan
               ra *= ( ra == flagVal.nan_value ) * (-1.0f) + 
                  ( ra != flagVal.nan_value ) * (1.0f);
            else ra *= -1.0f; // without nan
            const string buf = string("multiplied data with -1.0");
            if(flagVal.verbose > 0)
               cerr << buf << endl;
            outfile.addHistory(buf);
         }

         // if requested write subtractor constant to key
         if( flagVal.fitskey.length() > 0 )
            outfile.addValueCard(flagVal.fitskey, s,
                                 "used for subtractor constant");
         // if requested write subtractor file name to key
         if( flagVal.fitsfilekey.length() > 0 )
            outfile.addValueCard(flagVal.fitsfilekey, flagVal.subtractor_file,
                                 "used for file subtraction");
         
         // save file
         ft.saveImage(a, outfile);

         // load error
         const bool have_error =
            ft.loadError2(*file_i, e_a);

         // if no input error start with zero
         if( !have_error )
         {
            e_a.makeReference( a ); // recycle abondoned a
            e_a = 0.0f;
         }
         else { ra.free(); a.free(); } // not needed any more
         
         ft.referenceRegion( e_ra, e_a, flagVal.region );
         
         // calculate error
         if(flagVal.verbose > 0)
            cerr << "Calculating error ..." << endl;
         if( (!have_error) || flagVal.zero ) // no initial error in file -> zero not dominant
         {
	    if( e_line.isAllocated() ){
	       if(is_perrow){
		  for(int x = 1; x <= e_a.length(1); ++x)
		  {
		     MArray<FT, 1>a_line(e_a(x, Range::all()));
		     ftone.propagateZeroError2(a_line, e_line, FT(0));
		  }
	       }
	       else
	       {
		  for(int y = 1; y <= e_a.length(2); ++y)
		  {
		     MArray<FT, 1>a_line(e_a(Range::all(), y));
		     ftone.propagateZeroError2(a_line, e_line, FT(0));
		  }
	       }
	    }
	    if( e_b.isAllocated() || e_s )// must I really?
               ft.propagateZeroError2(e_ra, e_rb, e_s);
         }
         // if there was a error given -> propagate with
         // negative or zero in file(s) dominant
         else
         {
	    if( e_line.isAllocated() ){
	       if(is_perrow){
		  for(int x = 1; x <= e_a.length(1); ++x)
		  {
		     MArray<FT, 1>a_line(e_a(x, Range::all()));
		     ftone.propagateError2(a_line, e_line, FT(0));
		  }
	       }
	       else
	       {
		  for(int y = 1; y <= e_a.length(2); ++y)
		  {
		     MArray<FT, 1>a_line(e_a(Range::all(), y));
		     ftone.propagateError2(a_line, e_line, FT(0));
		  }
	       }
	    }
            if( e_b.isAllocated() || e_s ) // must I really?
               ft.propagateError2(e_ra, e_rb, e_s);
         }
         if( have_error || e_b.isAllocated() || e_s || e_line.isAllocated())
            ft.saveError2(e_a, outfile);
         else if(flagVal.verbose > 0)
            cerr << "but dropping it because it is zero." << endl;

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: subtractfits [-FLAGS [option]] FILES\n") +
      string("subtract flag dependent value or file from FILES\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("file", "", "subtract file",
                                       'f', &flagVal.subtractor_file) );
      flags.addOption(new FloatArrayOption("constant", "0.0,0.0", "subtractor[, error]",
                                           'c', 2, flagVal.subtractor ) );
      flags.addOption(new BoolOption("average", "FALSE",
                                     "subtract (clipped) average of region",
                                     'a', &flagVal.average ) );
      flags.addOption(new BoolOption("median", "FALSE",
                                     "subtract (clipped) median of region",
                                     'm', &flagVal.median ) );
      flags.addOption(new DoubleOption("kappa", "0.0",
                                      "for kappa sigma clipping, default: no clipping",
                                      'k', &flagVal.kappa) );
      flags.addOption(new BoolOption("sigma", "FALSE",
                                     "add (kappa corrected) sigma to error; default is using just (kappa corrected) standard error of average/median",
                                     's', &flagVal.sigma ) );
      flags.addOption(new BoolOption("twist", "FALSE",
                                     "if average clip respective to median, if median clip respective to average",
                                     't', &flagVal.twist ) );
      flags.addOption(new RegionArrayOption("of_region", "0:0,0:0",
                                            "use x_start:x_end,y_start:y_end region to determine average/median",
                                            'o', 2, &flagVal.of_region) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform only on region, leave rest of frame(s) untouched",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new RegionArrayOption("perrow", "0:0",
                                            "use x_start:x_end region to determine per row average/median",
                                            0, 1, &flagVal.perrow) );
      flags.addOption(new RegionArrayOption("per2row", "0:0,0:0",
                                            "use x1_start:x1_end,x2_start:x2_end regions to determine per row average/median",
                                            0, 2, &flagVal.per2row) );
      flags.addOption(new RegionArrayOption("percol", "0:0",
                                            "use y_start:y_end region to determine per column average/median",
                                            0, 1, &flagVal.percol) );
      flags.addOption(new RegionArrayOption("per2col", "0:0,0:0",
                                            "use y1_start:y1_end,y2_start:y2_end regions to determine per column average/median",
                                            0, 2, &flagVal.per2col) );
      flags.addOption(new StringOption("use_key", "", "subtract value of FITS keyword",
                                       0, &flagVal.usekey) );
      flags.addOption(new StringOption("use_error", "", "add to error value of FITS keyword",
                                       0, &flagVal.useerror) );
      flags.addOption(new StringOption("write_key", "", "write subtractor constant to FITS key",
                                       0, &flagVal.fitskey) );
      flags.addOption(new StringOption("write_file_key", "", "write subtractor file name to FITS key",
                                       0, &flagVal.fitsfilekey) );
      flags.addOption(new BoolOption("invert", "FALSE",
                                     "inverse arguments: output = (-1) * (FILES - subtractor)",
                                     'i', &flagVal.inverse ) );
      flags.addOption(new StringOption("prefix", "s", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("zero_error_not_dominant", "FALSE",
                                     "do NOT keep zero error dominant (only sometimes makes sense)",
                                     'z', &flagVal.zero ) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}

