/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: sumfits.cpp 94 2005-02-21 12:18:38Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT double

#include "fitstools.h"

// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) : region(dimension) { };
      bool nan;
      FT nan_value;
      Region region; // region class defined in util/option.h
      string exptime;
      string juldate;
      string weight;
      string prefix;
      bool renorm;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);

   MArray<double, 1> exptimes(files.size());
   MArray<double, 1>::iterator exptime_i(exptimes.begin());
   MArray<double, 1> juldates(files.size());
   MArray<double, 1>::iterator juldate_i(juldates.begin());
   MArray<FT, 1> weights(files.size());
   MArray<FT, 1>::iterator weight_i(weights.begin());

   try{
      const bool isExptime = flagVal.exptime.length() > 0;
      const bool isJuldate = flagVal.juldate.length() > 0;
      const bool isWeight  = flagVal.weight.length() > 0;

      // read 1st file
      list<string>::const_iterator file_i(files.begin());
      MArray<FT, dimension> a, ra, wa;
      FitsIn infile(*file_i, flagVal.quiet_please, flagVal.ignore_header);
      FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                      infile, flagVal.quiet_please);

      // read data from file
      if(isJuldate)
      {
         *juldate_i = infile.getFloat(infile.adjustKeyword(flagVal.juldate));
         ++juldate_i;
      }
      if(isExptime)
      {
         *exptime_i = infile.getFloat(infile.adjustKeyword(flagVal.exptime));
         ++exptime_i;
      }
      ft.loadImage(infile, a);
      // get region to operate on, default is whole array
      ft.referenceRegion(ra, a, flagVal.region);
      if(isWeight)
      {
         *weight_i = FT(infile.getFloat(infile.adjustKeyword(flagVal.weight)));
         ft.multiplyConstant(ra, *weight_i, flagVal.nan, flagVal.nan_value);
      }
      else weights = 1.0;
      if(!flagVal.nan)
      {
         wa.makeReference(MArray<FT, dimension>(ra.shape()));
         wa = merge(ra != flagVal.nan_value, *weight_i, 0.0);
      }
      ++weight_i;

      string history = string("added to ") + *file_i + string(": ");

      ++file_i;
      while( file_i != files.end() )
      {
         MArray<FT, dimension> b;
         FitsIn addfile(*file_i,
                        flagVal.quiet_please, flagVal.ignore_header);
         if(isJuldate)
         {
            *juldate_i = addfile.getFloat(addfile.adjustKeyword(flagVal.juldate));
            ++juldate_i;
         }
         if(isExptime)
         {
            *exptime_i = addfile.getFloat(addfile.adjustKeyword(flagVal.exptime));
            ++exptime_i;
         }
         if(flagVal.region.getStart(1))
            addfile.setRegion( flagVal.region );
         ft.loadImage(addfile, b);
         if(isWeight)
         {
            *weight_i = FT(addfile.getFloat(addfile.adjustKeyword(flagVal.weight)));
            ft.multiplyConstant(b, *weight_i, flagVal.nan, flagVal.nan_value);
         }
         if(!flagVal.nan)
         {
            wa += merge(b != flagVal.nan_value, *weight_i, 0.0);
         }
         ++weight_i;

         history += (*file_i) + string(" ");
         ft.addArray(ra, b, flagVal.nan, flagVal.nan_value);

         ++file_i;
      }

      outfile.addHistory(history);

      double expsum(0.0);
      const double weightsum( sum(weights) );
      const double n_files( double(files.size()) );
      const double weightnorm( n_files / weightsum );
      if(isExptime)
      {
         expsum = sum(exptimes);
         outfile.addValueCard(outfile.adjustKeyword(flagVal.exptime), expsum,
                              string("summed exposuretime"));
      }
      if(flagVal.renorm)
      {         
         outfile.addValueCard(outfile.adjustKeyword(flagVal.weight), 1.0 / weightnorm, "renormed weightsum");
         if(flagVal.nan)
            ra *= weightnorm;
         else
            ra = merge(wa != 0.0, n_files * ra / wa, flagVal.nan_value);
      }
      if(isJuldate)
      {
         double meanjuldate;
         if(isWeight)
            meanjuldate = sum(juldates * weights) / weightsum;
         else
         {
            if(isExptime) meanjuldate = sum(juldates * exptimes) / expsum;
            else meanjuldate = sum(juldates) / double(juldates.length(1));
         }
         outfile.addValueCard(outfile.adjustKeyword(flagVal.juldate), meanjuldate,
                              string("mean observation date"));
      }
      
      outfile.setBitpixOut(-32);
      ft.saveImage(a, outfile);
      a.free(); ra.free();
      
      // now propagate error
      file_i = files.begin();
      weight_i.reset();
      
      // open IO-file(s)
      ft.loadError2(*file_i, a);

      // get region to operate on, default is whole array
      if(a.isAllocated())
      {
         ft.referenceRegion(ra, a, flagVal.region);
         if(isWeight)
         {
            ra = merge(ra > 0.0, ra * (*weight_i), ra);
            ++weight_i;
         }
         ++file_i;
         while( file_i != files.end() )
         {
            MArray<FT, dimension> b;
            if(flagVal.region.getLength() > 1)
            {
               if(!ft.loadError2(*file_i, b, flagVal.region))
                  throw FitsException("file missing");
            }
            else
            {
               if(!ft.loadError2(*file_i, b))
                  throw FitsException("file missing");
            }
            if(isWeight)
            {
               b = merge(b > 0.0, b * (*weight_i), b);
               ++weight_i;
            }            
            if( flagVal.nan )
               ft.propagateError2(ra, b, 0.0);
            else
               ft.propagateZeroError2(ra, b, 0.0);
            ++file_i;
         }

         if(flagVal.renorm)
         {
            if(flagVal.nan)
               ra = merge(ra > 0.0, ra * weightnorm, ra);
            else
               ra = merge(ra > 0.0, n_files * ra / wa, ra);
         }

         // beautifier for translation misstreated error images
         ra = merge( (ra < 0.0) && (ra != -1.0), fabs(ra), ra);
         ft.saveError2(a, outfile);
      }
   }
   catch (exception &e){
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
   }
   
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: sumfits [-FLAGS [option]] FILES\n") +
      string("sum up FILES, keywords and filename taken from 1st file\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new DoubleOption("nan_value", "0.0",
                                       "use value as nan (see nan option)",
                                       0, &flagVal.nan_value) );
      flags.addOption(new BoolOption("divide_norm", "FALSE",
                                     "divide summation image by sum of weights",
                                     'd', &flagVal.renorm ) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform only on region, keep 1st file for rest of frame",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new StringOption("weight", "", "use value of keyword weight for summand (and if applicable also Juldate) weighing",
                                       'w', &flagVal.weight) );
      flags.addOption(new StringOption("exposuretime", "", "also sum up value of keyword exposuretime",
                                       'e', &flagVal.exptime) );
      flags.addOption(new StringOption("juldate", "", "calculate mean observation date with values of keyword juldate",
                                       'j', &flagVal.juldate) );
      flags.addOption(new StringOption("prefix", "s.", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}
