/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2007-2008 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

// dimension of FITS files
#define DIMENSION 2
// preferred type for images
// input maybe any but output will be of type FT
#define FT int

#include "fitstools.h"

// structure for flag values
class flagValues
{
   public:
      flagValues(int dimension) : region(dimension) { };
      bool nan;
      FT nan_value;
      Region region; // region class defined in util/option.h
      double exptime;
      string prefix;
      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);


int main( int argc, char *argv[] ) 
{

   // constants
   const int dimension = DIMENSION; // only do this with 2D images
   // alter #define to get other versions

   // structure for flag values
   flagValues flagVal(dimension);

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser     flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to class
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl;   
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);
   
   // instance for FitsTools class methods
   FitsTools<FT, DIMENSION> ft(flagVal.verbose, flagVal.ignore_header);
   const double exptime = flagVal.exptime;

   try{
      // read 1st file
      list<string>::const_iterator file_i(files.begin());
      MArray<FT, DIMENSION> a;
      FitsIn infile(*file_i, flagVal.quiet_please, flagVal.ignore_header);
      FitsOut outfile(ft.prependFilename(*file_i, flagVal.prefix),
                      infile, flagVal.quiet_please);

      // read data from file
      if(flagVal.region.getStart(1))
         infile.setRegion( flagVal.region );
      ft.loadImage(infile, a);

      string history = string("added to ") + *file_i + string(": ");

      ++file_i;
      while( file_i != files.end() )
      {
         MArray<FT, DIMENSION> b;
         FitsIn addfile(*file_i,
                        flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.region.getStart(1))
            addfile.setRegion( flagVal.region );
         ft.loadImage(addfile, b);

         history += (*file_i) + string(" ");
         ft.addArray(a, b, flagVal.nan, flagVal.nan_value);

         ++file_i;
      }
      outfile.addHistory(history);
      const double n_files( double(files.size()) );
      outfile.addValueCard(string("EXPTIME "), n_files * exptime,
                           string("integrated exposure time"));
      outfile.addValueCard(string("BZERO   "), 32768,
                           string("value = bzero + bscale * data"));
      outfile.addValueCard(string("BSCALE  "), 1,
                           string("value = bzero + bscale * data"));
      outfile.setBitpixOut(16);
      a-=32768;
      ft.saveImage(a, outfile);
      a.free();
   }
   catch (exception &e){
      if(flagVal.verbose > 0)
         cerr << e.what() << endl;
   }
   
   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: sumintfits [-FLAGS [option]] FILES\n") +
      string("sum up as integer FILES, keywords and filename taken from 1st file\n\n");
   const string copyright = 
      string("Copyright (C) 2007-2008 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new IntOption("nan_value", "0",
                                    "use value as nan (see nan option)",
                                       0, &flagVal.nan_value) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end perform and save only region",
                                            'r', DIMENSION, &flagVal.region) );
      flags.addOption(new DoubleOption("exposuretime", "1.0", "exptime of single image, sum goes to EXPTIME",
                                       'e', &flagVal.exptime) );
      flags.addOption(new StringOption("prefix", "s.", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
}
