/* -*- C++ -*-
 *
 * ---------------------------------------------------------------------
 * $Id: templatefits.cpp 177 2018-08-01 23:15:21Z cag $
 * ---------------------------------------------------------------------
 *
 * Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA 
 *
 * ---------------------------------------------------------------------
 *
 */

#include <ltl/util/command_line_reader.h>
#include <ltl/util/config_file_reader.h>
#include <ltl/util/option_parser.h>

#include <ltl/fitsio.h>

using namespace ltl;
using namespace util;
using namespace std;

// structure for flag values
class flagValues
{
   public:
      flagValues() : region(2) { };
      bool boolean;
      bool inv_boolean;
      int integer;
      float floatprec;
      double doubleprec;
      string str;
      
      int int_array[5];
      float float_array[2];
      double double_array[3];
      Region region; // region class defined in util/region.h

      float nan_value;
      bool nan;
      string prefix;

      int verbose;
      bool quiet_please;
      bool ignore_header;
};

void print_usage(OptionParser& flags);

void add_options(OptionParser& flags, flagValues& flagVal);

// fill in prefix to filename for given path
inline string prepend_filename(string path, const string& prefix)
{
   path.insert(path.find_last_of('/') + 1, prefix);
   return path;
}

int main(int argc, char* argv[])
{

   // structure for flag values
   flagValues flagVal;

   // set options with configfile
   const string cfgfilename = string(argv[0]) + string(".cfg");
   ConfigFileReader cfgfile(cfgfilename);
   OptionParser      flags(&cfgfile);
   try {
      add_options(flags, flagVal);           // save values to struct
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   // set options with commandline
   CommandLineReader comline( argc, argv ); // read from command line
   flags.changeReader(&comline);
   try {
      flags.parseOptions();
   }
   catch( UException e ) {
      if(flagVal.verbose > 0)
      {
         cerr << e.what() << endl;
         print_usage(flags);
      }
      exit(1);
   }

   if(flagVal.verbose > 1)
      cerr << flags << endl; // show flags
   flagVal.quiet_please = (flagVal.verbose == 0);

   const list<string> files = comline.fileArguments(); // get the filenames
   if(files.size() < 1) print_usage(flags);

   // for all FILES
   for( list<string>::const_iterator file_i=files.begin();
        file_i != files.end();
        ++file_i ) {
      try{
         // arrays for input and output
         MArray<float, 2> a, e_a;

         FitsIn infile(*file_i,
                       flagVal.quiet_please, flagVal.ignore_header);
         if(flagVal.verbose > 0)
            cerr << "Reading " << *file_i << endl;
         infile >> a; infile.freeData();

         try{
            const string errorname = prepend_filename(*file_i, "e.");
            FitsIn inerror(errorname,
                           flagVal.quiet_please, true);
            if(flagVal.verbose > 0)
               cerr << "Reading " << errorname << endl;
            inerror >> e_a; inerror.freeData();
         }
         catch (exception &e){
            if(flagVal.verbose > 0)
               cerr << e.what() << endl;
         }

         // do something with file

         // ...

         const string outname = prepend_filename(*file_i, flagVal.prefix);
         FitsOut outfile(outname,
                         infile, flagVal.quiet_please);
         outfile.addHistory(string("Have done something with ") + *file_i);
         if(flagVal.verbose > 0)
            cerr << "Saving " << outname << endl;
         outfile << a; a.free();

         if(e_a.isAllocated())
         {
            const string errorname = prepend_filename(outname, "e.");
            outfile.setFilename(errorname);
            if(flagVal.verbose > 0)
               cerr << "Saving " << errorname << endl;
            outfile << e_a;
         }
         

      } // this file should have worked, otherwise ->
      catch (exception &e){
         if(flagVal.verbose > 0)
            cerr << e.what() << endl;
      }
   }

   if(flagVal.verbose > 0)
      cerr << endl;
   return 0;
}

void print_usage(OptionParser& flags)
{
   const string usage =
      string("Usage: templatefits [-FLAGS [option]] FILES\n") +
      string("do something with FILES\n\n");
   const string copyright = 
      string("Copyright (C) 2000-2005 Claus A. Goessl <cag@usm.uni-muenchen.de>\n");
   cerr << usage;
   flags.printUsage( cerr );
   cerr << copyright;
   exit(1);   
}

void add_options(OptionParser& flags, flagValues& flagVal)
{
   try {
      
      flags.addOption(new BoolOption("bool", "FALSE",
                                     "example for boolean option",
                                     'b', &flagVal.boolean ) );
      flags.addOption(new BoolOption("inv_bool", "TRUE",
                                     "example for boolean option",
                                     0, &flagVal.inv_boolean ) );
      flags.addOption(new IntOption("int", "0",
                                    "example for integer option",
                                    'i', &flagVal.integer ) );
      flags.addOption(new FloatOption("float", "0.0",
                                      "example for floating precision option",
                                      'f', &flagVal.floatprec) );
      flags.addOption(new DoubleOption("double", "0.0",
                                       "example for double precision option",
                                       'd', &flagVal.doubleprec) );
      flags.addOption(new StringOption("string", "",
                                       "example for string option",
                                       's', &flagVal.str) );
      flags.addOption(new IntArrayOption("int_array", "1,2,3,4,5",
                                         "example for 5-elem integer array option",
                                         'j', 5, flagVal.int_array ) );
      flags.addOption(new FloatArrayOption("float_array", "1.0,2.0",
                                           "example for 2-elem float array option",
                                           'g', 2, flagVal.float_array ) );
      flags.addOption(new DoubleArrayOption("double_array", "1.0,2.0,3.0",
                                            "example for 3-elem double array option",
                                            'e', 3, flagVal.double_array ) );
      flags.addOption(new RegionArrayOption("region", "0:0,0:0",
                                            "x_start:x_end,y_start:y_end example for region option",
                                            'r', 2, &flagVal.region) );
      flags.addOption(new BoolOption("nan", "FALSE",
                                     "keep nan_value in either file dominant",
                                     'n', &flagVal.nan ) );
      flags.addOption(new FloatOption("nan_value", "0.0",
                                      "use value as nan (see nan option)",
                                      0, &flagVal.nan_value) );
      flags.addOption(new StringOption("prefix", "tmp.", "output file prefix",
                                       'p', &flagVal.prefix) );
      flags.addOption(new BoolOption("ignore_header", "FALSE",
                                     "do not parse FITS header (even BSCALE / BZERO are ignored)",
                                     0, &flagVal.ignore_header ) );
      flags.addOption(new IntOption("verbose", "1",
                                    "verbosity level 0 = shut up, 1 standard, etc.",
                                    'v', &flagVal.verbose ) );
   }
   catch( UException e ) {
      cerr << e.what() << endl;
      exit(1);
   }
  
}

